// Geoffrey Zweig (gzweig@microsoft.com)
// Copyright (c) 2009, 2010 Microsoft Corporation. All rights reserved.

#include <iostream>
#include <set>
#include <stdlib.h>
#include "commandline.h"
#include <vector>
#include "ngutil.h"

using namespace std;

const int slen = 1000;
char buff[slen];

// return values
static string string_rv;
static int int_rv;
static float float_rv;
static bool bool_rv;
static size_t size_rv;
static vector<bool> boolvec_rv;
static vector<string> stringvec_rv;
static vector<size_t> sizevec_rv;

void clparser::register_arg(string name, string type, string comment, string default_val, bool required) {
	arginfo4[name] = arg(type, default_val, comment, required);
	if( required && default_val.length() ) {
		cout << "No point having a default when argument " << name << " is required" << endl;
		exit(1);
	}
}

void *clparser::get(string name) const {

	if (arginfo4.find(name) == arginfo4.end()) {
		cout << name << " must be registered as a parameter before it can be accessed\n";
		exit(1);
	}

	arg a = arginfo4[name];
	size_t k;
	if (a.type == "string") {
		string_rv = a.val;
		cout << name << ": " << string_rv << endl;
		return (void *) &string_rv;
	} else if (a.type == "int") {
		int_rv = atoi(a.val.c_str());
		cout << name << ": " << int_rv << endl;
		return (void *) &int_rv;
	} else if (a.type == "size_t") {
		size_rv = (size_t)atoi(a.val.c_str());
		cout << name << ": " << size_rv << endl;
		return (void *) &size_rv;
	} else if (a.type == "float") {
		float_rv = (float) atof(a.val.c_str());
		cout << name << ": " << float_rv << endl;
		return (void *) &float_rv;
	} else if (a.type == "bool") {
		bool_rv = a.val=="true";
		cout << name << ": " << (bool_rv?"true":"false") << endl;
		return (void *) &bool_rv;
	} else if (a.type == "bool vector") {
		boolvec_rv.clear();
		const vector<string> vals = IO::GetList(a.val.c_str(), ',');
		for(k=0; k < vals.size(); ++k)
			boolvec_rv.push_back(vals[k] == "true");
		cout << name << ": " << a.val << endl;
		return (void *)&boolvec_rv;
	} else if (a.type == "string vector") {
		stringvec_rv = IO::GetList(a.val.c_str(), ',');
		cout << name << ": " << a.val << endl;
		return (void *)&stringvec_rv;
	} else if (a.type == "size_t vector") {
		sizevec_rv.clear();
		const vector<string> vals = IO::GetList(a.val.c_str(), ',');
		for(k=0; k < vals.size(); ++k)
			sizevec_rv.push_back(atoi(vals[k].c_str()));
		cout << name << ": " << a.val << endl;
		return (void *)&sizevec_rv;
	}

	cout << "Unknown type encountered: " << a.type << endl;
	exit(1);
	return NULL;
}

void clparser::parse(char *list[], int llen) {

	if (llen == 1) { // show the arguments
		cout << "Usage: " << list[0]  << "\n";
		for (map<string, arg>::iterator mi= arginfo4.begin(); mi!=arginfo4.end(); mi++) {
			if (!mi->second.required) 
				cout << "[-" << mi->first << " (" << mi->second.type << ": " 
				     << mi->second.val << ")]"; 
			else 
				cout << "-" << mi->first << " (" << mi->second.type << ")";
			if( mi->second.comment.length() )
				cout << "  " << mi->second.comment;
			cout << endl;
		}
		exit(1);
	}

	if ((llen % 2) == 0) {
		cout << "Command line must have an even number of elements\n";
		cout << "Check argument structure\n";
		exit(1);
	}

	set<string> seen;
	map< string, vector<pair<string, int> > > name2arg;

	for (int i=1; i<llen; i+=2) {
		if (list[i][0] != '-') {
			cout << "Argument names must begin with -\n";
			cout << "Saw: " << list[i] << endl;
			exit(1);
		}
		string aname(&list[i][1]);
		if (arginfo4.find(aname) == arginfo4.end()) {
			// try as a vector
			const char &last = *(aname.rbegin());
			if( last >= '0' && last <= '9' ) {
				char const * p = &last-1;
				for(; *p >= '0' && *p <= '9'; --p)
					;
				p++;
				string base(aname.c_str(), p);
				if( arginfo4.find(base) != arginfo4.end() ) {
					name2arg[base].push_back( pair<string,int>(list[i+1], atoi(p)) );
					continue;
				}
			}
			cout << "Unknown parameter on command line: " << aname << endl;
			exit(1);
		}
		arginfo4[aname].val = string(list[i+1]);
		seen.insert(aname);
	}

	// paste the arguments back into comma-separated list
	map< string, vector<pair<string, int> > >::const_iterator vargs = name2arg.begin();
	for(; vargs != name2arg.end(); ++vargs) {
		arg &dest = arginfo4[vargs->first];
		if( dest.val.length() ) {
			cout << "Cannot specify both -" << vargs->first << " with comma-separated and indexed arguments" << endl;
			exit(1);
		}
		vector<string> largs;
		size_t k;
		for(k=0; k < vargs->second.size(); ++k) {
			const size_t idx = (size_t)vargs->second[k].second;
			if( idx >= largs.size() )
				largs.resize(idx+1, "");
			if( largs[idx].length() != 0 ) {
				cout << "-" << vargs->first << idx << " multiply defined: " << largs[idx] << " vs. " << vargs->second[k].first << endl;
				exit(1);
			}
			largs[idx] = vargs->second[k].first;
		}
		dest.val = PRS::xjoin(",", largs);
	}

	// check that the required arguments have been seen
	map<string, arg>::iterator mi;
	for (mi=arginfo4.begin(); mi!=arginfo4.end(); mi++) 
		if (mi->second.required && !seen.count(mi->first)) {
			cout << "Required argument " << mi->first << " not set on command line\n";
			exit(1);
		}
}
