// Geoffrey Zweig (gzweig@microsoft.com)
// Patrick Nguyen (panguyen@microsoft.com)
// Copyright (c) 2009 Microsoft Corporation. All rights reserved.
//

#ifndef _GCC_
# define _WIN32_WINNT 0x0503
# define WINVER 0x0503
# include <windows.h>
#endif 

#include "ngutil.h"
#include "scarf_config.h"
#include "main_template.h"

#ifdef __USE_MPI
# include "mpiwrap.h"
#endif // __USE_MPI

#include <iostream>
#include <assert.h>
#include "model.h"

static void clamp_features(model &scarf, const std::string &clampsec) {
	char buf[MAX_LINE_LENGTH];
	strcpy_s(buf, clampsec.c_str());
	vector<string> fvspec = PRS::xtokenize(buf, ",");
	size_t k;
	string fname;
	for(k=0; k < fvspec.size(); ++k) {
		const char *p = strchr(fvspec[k].c_str(), '@');
		ERR_check(ERR::ERR_USERINPUT, p, "clamp must be name@value");
		fname.assign(fvspec[k].c_str(), p);
		const float fval = (float)atof(p+1);
		if( fname == "ALL" ) {
			ERR_check(ERR::ERR_USERINPUT, k == 0, "ALL must be specified first");
			(void)printf("Clamping all lambdas to %.4f\n", fval);
			std::fill(scarf.lambda.begin(), scarf.lambda.end(), fval);
			continue;
		}
		// look for hard-coded names
		if( fname == "global_lm" ) {
			(void)printf("Clamping to %s=%.4f\n", "global_lm", fval);
			scarf.lambda[scarf.global_lm_feature_idx] = fval;
			continue;
		}
		if( fname == "baseline" ) {
			(void)printf("Clamping to %s=%.4f\n", "baseline", fval);
			scarf.lambda[scarf.baseline_feature_idx] = fval;
			continue;
		}
		// look for external features
		size_t j;
		bool found = false;
		for(j=0; j < scarf.fexConfig.external_format.size(); ++j) {
			if( scarf.fexConfig.external_format[j] == fname ) {
				(void)printf("Clamping%sexternal %s=%.4f\n", scarf.fexConfig.lexternal[j] ? " all ": " ", fname.c_str(), fval);
				if( scarf.fexConfig.lexternal[j] ) {
					int w;
					for(w=0; w <= scarf.external_lexwords; w++)  // words go from 1 all the way up to external_lexwords
						scarf.lambda[scarf.idx4external[j]+w] = fval;
				} else
					scarf.lambda[scarf.idx4external[j]] = fval;
				found = true;
				break;
			}
		}
		if( !found ) {
			(void)fprintf(stderr, "Could not locate feature %s in model\n", fname.c_str());
			ERR_check(ERR::ERR_USERINPUT, found, "Feature not found");
		}
	}
}

int main(int argc, char *argv[]) {
	cout << "\nSCARF Decoder v1.2\n";
	cout << "gzweig@microsoft.com\n\n";

#ifdef __USE_MPI
	MPI::Initialize(&argc, &argv);
#endif // __USE_MPI

	clparser cl;

	CFG::GlobalConfig globConfig;
	CFG::AppConfig appConfig;
	CFG::DecodeConfig decConfig;

	decConfig.register_args(cl);
	appConfig.register_args(cl);
	globConfig.register_args(cl);

	cl.parse(argv, argc);

	globConfig.load_args(cl);
	initGlobal(cl, globConfig);
	appConfig.load_args(cl);
	decConfig.load_args(cl);

	ERR_check(ERR::ERR_USERINPUT, appConfig.lm.length() != 0, "Please specify a language model");

	WARN_check(ERR::ERR_USERINPUT, appConfig.fex.stream.size() != 0, "No detector streams specified");
	ERR_check(ERR::ERR_USERINPUT, decConfig.den_constraints.length() != 0, "Please specify a list of denominator constraints");

	model scarf(appConfig.fex);
	scarf.read_model(appConfig.model.c_str());

	scarf.max_alphas = appConfig.max_alphas;

	scarf.uttIt->init_raw_detectors();
	language_model LM(appConfig.lm.c_str(), scarf.idx4word, scarf.word4idx);

	scarf.LM = &LM;

	scarf.uttIt->init_constraint_readers("", decConfig.den_constraints.c_str());

	scarf.dictionaries = appConfig.dict;
	int unitnum = -1;
	scarf.read_dicts(scarf.dictionaries, unitnum, true);

	if( decConfig.clamp.length() )
		clamp_features(scarf, decConfig.clamp);

	if(decConfig.printModel)
		scarf.display();

	scarf.decode(decConfig.asConstraint);

#ifdef __USE_MPI
	MPI::UnInitHeartBeat();
	MPI::Uninitialize();
#endif

	return 0;
}


