// Geoffrey Zweig (gzweig@microsoft.com)
// Copyright (c) 2009 Microsoft Corporation. All rights reserved.

#include <iostream>
#include <stdio.h>
#include <assert.h>
#include <time.h>
#include "match_funcs.h"
#include "commandline.h"

#ifndef _GCC_
# pragma warning( disable:4996 )
#endif

// A note on times: this works properly with scarf when the times associated with
// the input units are the *end-times* of their spans as output by HapiVite.

/* e.g.
0 100 !sent_start
100 200 ae
200 300 n
300 400 !sent_end
-->
100 !sent_start
200 ae
300 n
400 !sent_end
*/

void mem_err() {cerr << "Out of memory\n"; exit(1);}

void get_units(vector<int> &units, vector<int> &times, fastMatch &fm, FILE *fin) {
	const int slen = 1000;
	char buff[slen], dname[slen];
	int dtime;

	units.clear();
	times.clear();
	fgets(buff, slen, fin); // stream name
	
	while (fgets(buff, slen, fin)) {
		int nf = sscanf(buff, "%s%i", dname, &dtime);
		if (nf == 1) {
			assert(!strcmp(dname, "."));
			break;
		}
		if (fm.idx4unit.find(dname) == fm.idx4unit.end()) {
			cout << "Unit in input not seen in index: " << dname << endl;
			exit(1);
		}
		units.push_back(fm.idx4unit[dname]);
		times.push_back(dtime);
	}

	for (vector<int>::iterator vi=times.begin()+1; vi<times.end(); vi++)
		if (*vi < *(vi-1)) {
			cout << "detection times must be increasing\n";
			exit(1);
		}
}

struct latrec {
	int st, et, from, to, word;
	string annotations;
	string wordval;  // in case not in index
	latrec() {from=to=-1; annotations="";}  // -1 is a flag - not used
};

void get_lat(vector<latrec> &lattin, fastMatch &fm, FILE *fin) {
	lattin.clear();
	const int slen = 1000;
	char buff[slen], word[slen], annots[slen], from[slen];
	int st, et, to;
	while (fgets(buff, slen, fin)) {
		if (buff[0] == '.')  // reached the end
			return;
		latrec lr;
		int nf = sscanf(buff, "%i%i%s%s%i%s", &st, &et, word, from, &to, annots);
		if (nf < 2) {
			cout << "expecting 3 fields reading lattices, but saw: ";
			cout << buff << endl;
			exit(1);
		}
		lr.st = st;
		lr.et = et;
		string ws(word);
		if (fm.idx4word.count(ws) == 0) {
			cout << "Word " << ws << " found in lattices being annotated, but not in TFIDF index\n";
			cout << "Feature will be 0\n";
			lr.word = INT_MAX;
			lr.wordval = ws;
		} else
			lr.word = -fm.idx4word[ws];  // - due to negative numbering scheme
		if (nf == 4) { 
			lr.annotations = string(from);
		} else if (nf == 5) {
			lr.from = atoi(from);
			lr.to = to;
		} else if (nf == 6) {
			lr.from = atoi(from);
			lr.to = to;
			lr.annotations = annots;
		}
			
		lattin.push_back(lr);
	}
}


int main(int argc, char * argv[])
{
	cout << "\nSCARF Fast Match v1.2\n";
	cout << "gzweig@microsoft.com\n\n";

	set_new_handler(mem_err);

	clparser cl;
	cl.register_arg("dictionary", "string", "the lexical count file from which to create the index", "");
	cl.register_arg("unit-infile", "string", "file containing detection events from which to decode", "");
	cl.register_arg("lattice-outfile", "string", "file to store output lattices in", "");
	cl.register_arg("lattice-infile", "string", "file to read lattices from in order to annotate them", "");
	cl.register_arg("model-in", "string", "name of index file to use for decoding or lattice annotation", "");
	cl.register_arg("model-out", "string", "name of file to write index into after construction", "");
	cl.register_arg("ngram-level", "int", "length of unit ngrams used in index", "2");
	cl.register_arg("max-matches", "int", "number of word candidates to consider for a segment in first pass", "50");
	cl.register_arg("max-word-len", "int", "maximum number of units to consider grouping into a word", "15");
	cl.register_arg("phonedelta", "float", "maximum permissible difference between expected and observed number of units in a word", "4.0");
	cl.register_arg("indexfrac", "float", "a term occuring in a smaller fraction of a words pronunciation will not be associated with it in the reverse index", "0.1");
	cl.register_arg("fm-lmwt", "float", "LM weight in first pass", "0.2");
	cl.register_arg("dm-lmwt", "float", "LM weight in second pass", "0.02");
	cl.register_arg("lm", "string", "the language model file", "");
	cl.register_arg("maxalphas", "int", "number of distinct paths alive at any time in second pass", "20");
	cl.register_arg("lengthwt", "float", "weight associated with length portion of match score", "0.1");
	cl.register_arg("usexm", "bool", "assign 0 acoustic cost to exact matches", "false");
	cl.register_arg("del_cost", "float", "cost of deleting a unit in lattice rescoring", "3.0");
	cl.register_arg("ins_cost", "float", "used in backoff score when cosine distance is 0", "3.0");
	cl.register_arg("max-files", "int", "cease processing after this many utterances are processed", "10000000");

	cl.parse(argv, argc);
	string lm = *((string *) cl.get("lm"));
	string dict = *((string *) cl.get("dictionary"));
	string modelin = *((string *) cl.get("model-in"));
	string modelout = *((string *) cl.get("model-out"));
	string infile = *((string *) cl.get("unit-infile"));
	string lattice_outfile = *((string *) cl.get("lattice-outfile"));
	string lattice_infile = *((string *) cl.get("lattice-infile"));
	int n = *((int *) cl.get("ngram-level"));
	int maxn = *((int *) cl.get("max-matches"));
	int maxl = *((int *) cl.get("max-word-len"));
	float phonedelta = *((float *) cl.get("phonedelta"));
	float indexfrac = *((float *) cl.get("indexfrac"));
	float fmlmwt = *((float *) cl.get("fm-lmwt"));
	float dmlmwt = *((float *) cl.get("dm-lmwt"));
	int ma = *((int *) cl.get("maxalphas"));
	float lw = *((float *) cl.get("lengthwt"));
	bool check_xm = *((bool *) cl.get("usexm"));
	float del_cost = *((float *) cl.get("del_cost"));
	float ins_cost = *((float *) cl.get("ins_cost"));
	int max_files = *((int *) cl.get("max-files"));

	if (dict=="" && modelin=="") {
		cout << "Either a dictionary or an input model must be specified\n";
		exit(1);
	}

	if (dict!="" && modelin!="") {
		cout << "Both a dictionary and an input model cannot be specified at the same time\n";
		exit(1);
	}

	if (!(modelout!="" || (lattice_outfile!="" && infile!=""))) {
		cout << "Either an output model or both a lattice file and an input file must be specified\n";
		exit(1);
	}

	if (maxl <= n) {
		cout << "ngram level should be less than the maximum legnth of hypothesized words\n";
		exit(1);
	}

	fastMatch fm;
	if (dict != "") {
		fm.create_index(dict.c_str(), n, indexfrac);
	} else {
		assert(modelin != "");
		fm.read_index(modelin.c_str());
	}

	if (modelout!="") {
		fm.write_index(modelout.c_str());
	}

	if (infile!="" && lattice_infile=="") { // going to do decoding
		if (lm=="") {
			cout << "going to do decoding, but no LM is specified\n";
			exit(1);
		}
		fm.LM = new language_model(lm.c_str(), fm.idx4word, fm.word4idx);
		fm.phonedelta = phonedelta;
		fm.fm_lmwt = fmlmwt;
		fm.dm_lmwt = dmlmwt;
		fm.lengthwt = lw;
		fm.maxalphas = ma;
		fm.check_xm = check_xm;

		FILE *fin = fopen(infile.c_str(), "r");
		if (!fin) {
			cout << "unable to open " << infile << endl;
			exit(1);
		}

		const int slen = 1000;
		char outfile[slen];
		strcpy(outfile, lattice_outfile.c_str());
		char *p = strrchr(outfile, '.');
		FILE *fout;
		if (p && !strcmp(p, ".gz"))  { // deal with gzipp'd output
	        string process = "gzip - > " + string(outfile);
	        fout = POPEN(process.c_str(), "w");
			if (!fout) {cerr << "Unable to process " << process << endl; exit(1);}
		} else {
			fout = fopen(outfile, "w");
			if (!fout) { cout << "unable to open " << outfile << endl; exit(1);}
		}

		time_t start_time;
		time(&start_time);

		char buff[slen];
		vector<int> matches, units, times;
		int c=0;
		while (fgets(buff, slen, fin)) {
			fputs(buff, fout);
			string inname(buff);
			get_units(units, times, fm, fin);
			fm.decode(units, maxl, maxn);
			vector<pair<int, int> > output = fm.backtrace(times, fout);
			if (output.size()) {
				reverse(output.begin(), output.end());
				for (size_t i=0; i<output.size(); i++) 
					cout << fm.word4idx[-output[i].first] << " ";  // index uses negative numbers
				cout << "@@ ";
				cout << int((1+output.front().second)/2) << " ";
				for (size_t i=1; i<output.size(); i++) {
					// the times that are stored are end-times
					int t = int(output[i].second + output[i-1].second) / 2;
					cout << t << " ";
				}
				cout << "@@ " << inname;
			} else {
				cout << "NULL @@ " << inname;
			}
			if (++c==max_files) break;
		}
		time_t end_time;
		time(&end_time);
		double elapsed_time = difftime(end_time, start_time);
		cout << elapsed_time << " seconds in decoding\n";
	} else if (infile!="" && lattice_infile!="") { // going to annotate lattices

		// AM related parameters
		fm.lengthwt = lw;
		fm.check_xm = check_xm;

		FILE *fin = fopen(infile.c_str(), "r");
		if (!fin) {
			cout << "unable to open " << infile << endl;
			exit(1);
		}

		const int slen = 1000;
		char linfile[slen];
		strcpy(linfile, lattice_infile.c_str());
		char *p = strrchr(linfile, '.');
		FILE *lin;
		if (p && !strcmp(p, ".gz"))  { // deal with gzipp'd input
	        string process = "gzip -cd " + lattice_infile;
	        lin = POPEN(process.c_str(), "r");
			if (!lin) {cerr << "Unable to process " << process << endl; exit(1);}
		} else {
			lin = fopen(lattice_infile.c_str(), "r");
			if (!lin) { cout << "unable to open " << lattice_infile << endl; exit(1);}
		}

		char outfile[slen];
		strcpy(outfile, lattice_outfile.c_str());
		p = strrchr(outfile, '.');
		FILE *fout;
		if (p && !strcmp(p, ".gz"))  { // deal with gzipp'd output
	        string process = "gzip - > " + string(outfile);
	        fout = POPEN(process.c_str(), "w");
			if (!fout) {cerr << "Unable to process " << process << endl; exit(1);}
		} else {
			fout = fopen(outfile, "w");
			if (!fout) { cout << "unable to open " << outfile << endl; exit(1);}
		}

		time_t start_time;
		time(&start_time);

		char buff[slen];
		vector<int> units, times;
		vector<latrec> lattin;
		int c=0;
		while (fgets(buff, slen, fin)) {
			get_units(units, times, fm, fin);
			fgets(buff, slen, lin);
			fputs(buff, fout);
			get_lat(lattin, fm, lin);

			for (size_t i=0; i<lattin.size(); i++) {
				latrec &lr = lattin[i];
				vector<int>::iterator a = lower_bound(times.begin(), times.end(), lr.st);
				vector<int>::iterator b = upper_bound(times.begin(), times.end(), lr.et);
				int x = int(a-times.begin()), y = int(b-times.begin());
				vector<int> block(units.begin()+x, units.begin()+y);
				float acscore = 0;
				if (lr.word != INT_MAX) {  // a word known to the index
					if (block.size()) {
						float tf = -fm.get_tfidfscore(lr.word, block);
						assert(tf>=0.0f);
						if (tf == 0.0f) 
							acscore = -del_cost*fm.length4word[lr.word] - ins_cost*block.size();
						else
							acscore = log(tf);
					} else  {
						string wds = fm.word4idx[-lr.word];
						acscore = -del_cost * fm.length4word[lr.word];  // deleted the units
                                                if (wds=="~SIL") acscore = 0.0;
					}
					strcpy(buff, fm.word4idx[-lr.word].c_str());
				} else {  // acscore will default to 0; cant use the fm dictionary
					strcpy(buff, lr.wordval.c_str());
				}
				fprintf(fout, "%i %i %s", lr.st, lr.et, buff);
				if (lr.from != -1) {
					fprintf(fout, " %i %i", lr.from, lr.to);
				}
				if (lr.annotations != "") {
					strcpy(buff, lr.annotations.c_str());
					fprintf(fout, " %s,", buff);
					fprintf(fout, "tfidf=%f\n", acscore);
				} else 
					fprintf(fout, " tfidf=%f\n", acscore);
			}
			fprintf(fout, ".\n");
			if (++c==max_files) break;
		}
		time_t end_time;
		time(&end_time);
		double elapsed_time = difftime(end_time, start_time);
		cout << elapsed_time << " seconds annotating lattices\n";
	}
	return 0;
}

