// Geoffrey Zweig (gzweig@microsoft.com)
// Copyright (c) 2009 Microsoft Corporation. All rights reserved.

#include <iostream>
#include <assert.h>
#include <math.h>
#include "match_funcs.h"
#include "futil.h"

using namespace std;

inline float sqr(float a) {return a*a;}

const float LOG_0 = -1e30F;
const float LOG_1 = 0.0F;

void logadd(float &target, float inc)
{
        if (target < inc)
                swap(target, inc);
        target += log(1 + exp(inc-target));
}

void fastMatch::make_doc(doc &currdoc, vector<int> &phones, int count) {
	vector<int> ng(nglen);
	for (size_t i=0; i<phones.size(); i++) {
		for (size_t j=1; j<=size_t(nglen) && i+j<=phones.size(); j++) {
			ng.resize(j);
			copy(phones.begin()+i, phones.begin()+i+j, ng.begin());
			if (!idx4ngram.count(ng)) {
				idx4ngram[ng] = ngnum;
				ngram4idx[ngnum] = ng;
				ngnum++;
			}
			currdoc.count4unit[idx4ngram[ng]] += count;
			currdoc.nwords += count;
		}
	}
}

void fastMatch::parse_phones(char *input, vector<int> &output, int &inum) {
	output.clear();
	char * p;
	p = strtok(input, " \n");
	ERR_check(ERR::ERR_USERINPUT, p, "wrong format while parsing phones");
	do {
		string ps(p);
		if (!idx4unit.count(ps)) {
			idx4unit[ps] = inum;
			unit4idx[inum++] = ps;
		}
		output.push_back(idx4unit[ps]);
	} while ( (p=strtok(NULL, " \n")) != NULL );
}

void fastMatch::create_index(const char *fname, int ngl, float mf)
{	
	nglen = ngl;
	minfrac = mf;
	ngnum = 0;
	IO::auto_file_t fin(fname, "rt");

	string last_word = "";
	const int slen = 10000;
    char buff[slen], dictpron[slen], realpron[slen];
	int count;
	int inum=0, wnum=0;
	set<vector<int> > xprons;
	int noccs = 0;
	while (fgets(buff,slen,fin)) {
		// get the separate fields
		char *p = strtok(buff, "\t\n");
		string ws(p);
		if (ws == "!sent_start")  // fix HTK notation
			ws = "<s>";
		else if (ws == "!sent_end")
			ws = "</s>";
		if (!idx4word.count(ws)) {  // word indexes are negative; undo this everywhere idx4word 
			idx4word[ws] = wnum;    // is accessed
			word4idx[wnum] = ws;
			wnum--;
		}
		ERR_check(ERR::ERR_USERINPUT, p=strtok(NULL, "\t\n"), "failed to find separator in word index");
		strcpy(dictpron, p);
		ERR_check(ERR::ERR_USERINPUT, p=strtok(NULL, "\t\n"), "failed to find second separator in word index");
		count = atoi(p);
		ERR_check(ERR::ERR_USERINPUT, p=strtok(NULL, "\t\n"), "failed to find third separator in word index");
		strcpy(realpron, p);

		// see how long the dictionary pron is
		vector<int> dphones;
		parse_phones(dictpron, dphones, inum);
		sumphones4[-idx4word[ws]] += count*dphones.size();  // expected distribution

		// get the individual phones
		vector<int> phones;
		parse_phones(realpron, phones, inum);
		
		// update the pronunciation lists
		if (ws != last_word) {
			if (last_word != "") {
				xm4.push_back(xprons);
				noccs4doc.push_back(noccs);
			}
			docs.push_back(doc(-idx4word[ws]));
			xprons.clear();
			noccs = 0;
		}
		doc &currdoc = docs.back();
		noccs += count;
		make_doc(currdoc, phones, count);			
		last_word = ws;
		xprons.insert(dphones);
	}

	// handle the last word right
	xm4.push_back(xprons);
	noccs4doc.push_back(noccs);

	int tnum = ngnum;
	compute_idf();
	index.clear();
	for (size_t i=0; i<docs.size(); i++) {
		index.push_back(idx_entry(docs[i].label));
		compute_tfidf_vec(docs[i], index.back().tfidfvec); 
	}
	ERR_check(ERR::ERR_SANITY, ngnum == tnum, "ngnum wrong size after assembling tfidf vector");

	length4word.resize(idx4word.size());
	map<string, int>::iterator mis;
	for (mis=idx4word.begin(); mis!=idx4word.end(); mis++) {
		length4word[-mis->second] = sumphones4[-mis->second] / float(noccs4doc[-mis->second]);
	}
}

void fastMatch::compute_idf()
{
	// compute the idf scores for all the words in the corpus
	size_t ndocs = docs.size();
    idf.resize(ngnum);
	scratch.resize(ngnum);
	fill(scratch.begin(), scratch.end(), 0.0f);
	fill(idf.begin(), idf.end(), 0.0f);
	float docinc = float(1.0/ndocs);
	for (size_t i=0; i<ndocs; i++) {
		map<int, int>::iterator mi;
		for (mi=docs[i].count4unit.begin(); mi!=docs[i].count4unit.end(); mi++)
			idf[mi->first] += docinc;
	}

	for (size_t j=0; j<idf.size(); j++) 
		idf[j] = (float) log(1.0/idf[j]) + 1.0f;

	// make the reverse index
	docs4idx.resize(ngnum);
	size_t doc_term_pairs = 0;
	for (size_t i=0; i<ndocs; i++) {
		map<int, int>::iterator mi;
		for (mi=docs[i].count4unit.begin(); mi!=docs[i].count4unit.end(); mi++) {
			float frac = float(mi->second) / noccs4doc[i];
			if (frac >= minfrac) {
				ERR_check(ERR::ERR_SANITY, docs[i].label == int(i), "retrieved doc has wrong index");
				docs4idx[mi->first].push_back(i);
				doc_term_pairs++;
			}
		}
	}
	cout << ndocs << " words; " << doc_term_pairs << " word-term pairs" << endl;
}

void fastMatch::display_idf() {
	map<vector<int>, int>::iterator mi;
	cout << "ng size: " << idx4ngram.size() << " " << ngnum << endl;
	for (mi=idx4ngram.begin(); mi!=idx4ngram.end(); mi++) {
		const vector<int> &ngram = mi->first;
		for (size_t i=0; i<ngram.size(); i++) 
			cout << unit4idx[ngram[i]] << " ";
		cout << ": " << idf[mi->second] << endl;
	}
}

void fastMatch::compute_tfidf_vec(doc &currdoc, vector<pair<int, float> > &tfidfvec) 
{
	tfidfvec.clear();
	map<int, int>::iterator mi;
	float norm = 0.0;
	for (mi = currdoc.count4unit.begin(); mi != currdoc.count4unit.end(); mi++) {
		if (mi->first >= int(idf.size()))  // doc has a wprd/ngram not in the index
			continue;
		float tfidf = (*mi).second * idf[(*mi).first] / currdoc.nwords;
		norm += tfidf*tfidf;
		tfidfvec.push_back(pair<int, float>((*mi).first, tfidf));
	}

	if (norm != 0) {    // could be 0 if query has only words/units not in listings
		norm = (float) sqrt(norm);
		norm = (float) 1.0/norm;
		for (size_t i=0; i<tfidfvec.size(); i++) {
			tfidfvec[i].second *= norm;
		}
	}
}

float fastMatch::cosine(vector<pair<int, float> > &v1, vector<pair<int, float> > &v2) 
{
	float cosval = 0.0;
	for (size_t i=0; i<v2.size(); i++)
		scratch[v2[i].first] = v2[i].second;
	for (size_t i=0; i<v1.size(); i++)
		cosval += v1[i].second * scratch[v1[i].first];
	for (size_t i=0; i<v2.size(); i++)
		scratch[v2[i].first] = 0.0f;
	return cosval;
}

inline float fastMatch::tfidfscore(int w, vector<pair<int, float> > &tfidfvec, vector<int> &units) {
	float lp = exp(-sqr(length4word[w] - units.size())*lengthwt);
	float acscore = -1.0f;  // default if we have an exact match - best approx. match score
	if (!check_xm || !xm4[w].count(units))  // use approximate match score
		acscore = -cosine(index[w].tfidfvec, tfidfvec);
	return acscore*lp;
}

void fastMatch::match(vector<int> &units, vector<pair<int, float> > &matches, int maxn)
{
	doc currdoc(-1);
	make_doc(currdoc, units);
	vector<pair<int, float> > tfidfvec;
	compute_tfidf_vec(currdoc, tfidfvec);

	// find the word candidates
	vector<int> wordcand;
	for (size_t i=0; i<tfidfvec.size(); i++) {
		if (units.size()>1 && ngram4idx[tfidfvec[i].first].size()==1)  // don't use single phones
			continue;
		vector<int> &wc = docs4idx[tfidfvec[i].first];
		for (size_t j=0; j<wc.size(); j++) {
			if (fabs(length4word[wc[j]]-units.size()) < phonedelta) 
				wordcand.push_back(wc[j]);
		}
	}
	sort(wordcand.begin(), wordcand.end());
	vector<int>::iterator vi = unique(wordcand.begin(), wordcand.end());
	wordcand.erase(vi, wordcand.end());

	vector<pair<float, int> > values;
	for (size_t i=0; i<wordcand.size(); i++) {
		float tscore = tfidfscore(wordcand[i], tfidfvec, units);
		values.push_back(pair<float, int>(tscore*pow((float)noccs4doc[wordcand[i]], fm_lmwt), 
									  wordcand[i]));
	}

	sort(values.begin(), values.end());
	matches.clear();
	int limit = min(maxn, (int)values.size());
	for (int i=0; i<limit; i++) {
		matches.push_back(pair<int, float>(values[i].second, 
			-values[i].first / pow((float)noccs4doc[values[i].second], fm_lmwt))); // store TFIDF only
		ERR_check(ERR::ERR_SANITY, matches.back().second > 0, "number of matches must be strictly positive");
		matches.back().second = log(matches.back().second);
	}
}

float fastMatch::get_tfidfscore(int w, vector<int> &units) {
	doc currdoc(-1);
	make_doc(currdoc, units);
	vector<pair<int, float> > tfidfvec;
	compute_tfidf_vec(currdoc, tfidfvec);
	return tfidfscore(w, tfidfvec, units);
}

void fastMatch::show_units(vector<int> &units)
{
	for (size_t i=0; i<units.size(); i++) {
		ERR_check(ERR::ERR_SANITY, unit4idx.count(units[i]), "attempting to display unknown unit");
		cout << unit4idx[units[i]] << " ";
	}
	cout << endl;
}

void fastMatch::show_tfidf(vector<pair<int, float> > &vec) {
	for (size_t i=0; i<vec.size(); i++) {
		const vector<int> &ngram = ngram4idx[vec[i].first];
		for (size_t j=0; j<ngram.size(); j++) 
			cout << unit4idx[ngram[j]] << " ";
		cout << ": " << vec[i].second << endl;
	}
}

pair<lmidx, float> fastMatch::succ(lmidx state, int word) {
	if (!LM->lm_words.count(word))
		word = LM->unk_idx;
	return pair<lmidx, float>(LM->lm_next_hist(state, word), LM->lm_logp(state, word));
}

void fastMatch::decode(vector<int> &units, int maxl, int maxn) {
	alphas.clear();
	alphas.resize(units.size()+1); 
	pred.clear();
	alphas[0][lmrec(LM->initial_state, LM->bos_idx, time_lat_rec::lstart)] = LOG_1;
	map<lmrec,float>::iterator mi;
	vector<vector<pair<int, float> > > all_cands(units.size());
	beams.clear();
	beams.resize(units.size()+1);
	beams.back() = -FLT_MAX;  // dummy final state

	for (int i=0; i<int(units.size()); i++) {

		// determine the beam
		vector<float> probs;
		for (mi=alphas[i].begin(); mi!=alphas[i].end(); mi++)
			probs.push_back(mi->second);
		float limit = -FLT_MAX;
		if (probs.size() > maxalphas) {
			nth_element(probs.begin(), probs.begin()+maxalphas-1, probs.end(), greater<float>());
			limit = probs[maxalphas-1];
		}
		beams[i] = limit;

		// compute all the candidate extensions once
		for (int j=i; j<i+maxl && j<int(units.size()); j++) {
			vector<int> u2(units.begin()+i, units.begin()+j+1);
			match(u2, all_cands[j], maxn);  // get words spanning positions i through j inclusive

			vector<pair<int, float> > &candidates = all_cands[j];
			if (i==0) { // hack a forced start in <s>
				vector<pair<int, float> > newc;
				for (size_t r=0; r<candidates.size(); r++)
					if (candidates[r].first == -LM->bos_idx)
						newc.push_back(candidates[r]);
				candidates = newc;
			}
		}

		// advance the alphas
		for (mi=alphas[i].begin(); mi!=alphas[i].end(); mi++) {

			if (mi->second < limit) 
				continue;

			lmidx s = mi->first.state;
			int pw = mi->first.w;
			float av = mi->second;

			for (int j=i; j<i+maxl && j<int(units.size()); j++) {
				vector<pair<int, float> > &candidates = all_cands[j];
				int et = j+1;  // endtime - when the next word will start
				int lp=time_lat_rec::lstart, elp=time_lat_rec::lstart; // no lattice constraints
				map<lmrec,float> &a = alphas[et];
				for (size_t c=0; c<candidates.size(); c++) {
					int w = candidates[c].first;
					pair<lmidx, float> nextstate = succ(s,-w); // next LM state + logprob
															   // -w because indexes are negative
					lmidx ns = nextstate.first;
					float lmprob = nextstate.second;
					float acprob = candidates[c].second;
					if (a.find(lmrec(ns,w,elp))==a.end()) 
						a[lmrec(ns,w,elp)] = LOG_0;
					float candprob = av + acprob + dm_lmwt*lmprob;
					pred[statetime(ns,et,w,elp)].push_back(statetime_cost(s,i,pw,lp,candprob));
					if (candprob > a[lmrec(ns,w,elp)]) {
						a[lmrec(ns,w,elp)] = candprob;
						// the new way must be better; move it to position 0 for traceback
						swap(pred[statetime(ns,et,w,elp)].front(),  pred[statetime(ns,et,w,elp)].back());
					}
				}
			}
		}
	}
}

struct statetime_state_compare {
	// will be applied to find predecessors that differ only in start-time
	bool operator () (const statetime_cost &a, const statetime_cost &b) const {
		if (a.state < b.state) return true;
		else if (a.state > b.state) return false;
		else if (a.w < b.w) return true;
		else if (a.w > b.w) return false;
		else return (a.score > b.score);  // want high-scoring first
	}
};

struct statetime_state_equal {
	// will be applied to remove preds that differ only in start-time
	bool operator () (const statetime_cost &a, const statetime_cost &b) const {
		return (a.state == b.state && a.w == b.w);
	}
};

struct wordspan_start_compare {
	bool operator () (const wordspan &a, const wordspan &b) const {
		if (a.starttime == b.starttime) {
			if (a.endtime == b.endtime)
				return (a.idx < b.idx);
			else 
				return (a.endtime < b.endtime);
		}
		return (a.starttime < b.starttime);
	}
};

vector<pair<int, int> > fastMatch::backtrace(vector<int> &times, FILE *fout) {
	float dataprob = LOG_0;
	lmidx endstate = -1; // impossible: >= 0
	vector<pair<int, int> > words;
	int lastword = -1;
	int latp = -1;
	vector<set<statetime> > livestatetimes(alphas.size());
	map<lmrec,float>::iterator mi;
	for (mi=alphas.back().begin(); mi!=alphas.back().end(); mi++) {
		if (LM->final_states.count(mi->first.state)) {
			statetime s(mi->first.state, alphas.size()-1,mi->first.w,mi->first.lat_end_state);
			livestatetimes[alphas.size()-1].insert(s);
			if (mi->second > dataprob) {
				dataprob = mi->second;
				endstate = mi->first.state;
				lastword = mi->first.w;
				latp = mi->first.lat_end_state;
			}
		}
	}

	if (endstate == -1) {
		fprintf(fout, ".\n");
		return words;  // empty vector
	}

	ERR_check(ERR::ERR_SANITY, latp != -1, "endstate and latp are inconsistent");
	ERR_check(ERR::ERR_SANITY, lastword != -1, "endstate and latword are inconsistent");
	words.push_back(pair<int,int>(lastword, times.back()));  // last word ends at final time
	statetime_cost pr = pred[statetime(endstate,alphas.size()-1,lastword,latp)][0];
	int et = alphas.size()-1;
	while (pr.state != LM->initial_state) {
		words.push_back(pair<int, int>(pr.w, times[pr.t-1]));
		if (pr.score < beams[et]) {
			cerr << "Error: 1 best falls below beam: " << pr.t << " " << beams[et] << " " << pr.score << endl;
			exit(1);
		}
		et = pr.t;
		pr = pred[statetime(pr.state, pr.t, pr.w, pr.latpos)][0];
	}

	// write the lattice
	ERR_check(ERR::ERR_SANITY, alphas.size()==times.size()+1, "alphas different from lattice size");
	vector<wordspan> lattice;
	vector<float> nullvec;
	for (int t=int(alphas.size()-1); t>0; t--) {
		set<statetime>::iterator si;
		for (si=livestatetimes[t].begin(); si!=livestatetimes[t].end(); si++){
			ERR_check(ERR::ERR_SANITY, si->t == t, "live state has wrong time indx");
			int et = times[si->t - 1];  
			vector<statetime_cost> &curpreds = pred[*si];

			// remove lower scoring preds that differ only in start-time
			sort(curpreds.begin(), curpreds.end(), statetime_state_compare());
			vector<statetime_cost>::iterator vi =
				unique(curpreds.begin(), curpreds.end(), statetime_state_equal());
			curpreds.erase(vi, curpreds.end());

			for (size_t i=0; i<curpreds.size(); i++) {

				if (curpreds[i].score < beams[t]) // skip poorly scoring predecessors
					continue;

				int st = (curpreds[i].t==0)?(1):(times[curpreds[i].t-1]+1);
				lattice.push_back(wordspan(si->w,st,et,-1, nullvec));
				statetime_cost &stc = curpreds[i];
				livestatetimes[stc.t].insert(statetime(stc.state,stc.t,stc.w,stc.latpos));
			}
		}
	}

	sort(lattice.begin(), lattice.end(), wordspan_start_compare());
	// remove redundancy that is present since we ignore the structure and just write the wordspans
	vector<wordspan>::iterator vi = unique(lattice.begin(), lattice.end()); 
	lattice.erase(vi, lattice.end());
	for (size_t i=0; i<lattice.size(); i++) {
		string word = word4idx[-lattice[i].idx]; // negative indexes, so minus
		fprintf(fout, "%i %i %s\n", lattice[i].starttime, lattice[i].endtime, word.c_str());
	}
	fprintf(fout, ".\n");

	return words;
}

void fastMatch::write_index(const char *fname) {
	FILE *fout = fopen(fname, "wb");
	if (!fout) {
		cout << "Unable to open " << fname << endl;
		exit(1);
	}

	// write the easy things first
	write_string2int_map(fout, idx4word);
	write_int2string_map(fout, word4idx);
	write_string2int_map(fout, idx4unit);
	write_int2string_map(fout, unit4idx);
	write_vec_float(fout, length4word);
	write_vec_float(fout, idf);
	write_vec_int(fout, noccs4doc);

	fwrite((void *) &nglen, sizeof(int), 1, fout);
	fwrite((void *) &ngnum, sizeof(int), 1, fout);

	// now for the tough ones

	int s = idx4ngram.size();
	fwrite((void *) &s, sizeof(int), 1, fout);
	for (map<vector<int>, int>::iterator mi=idx4ngram.begin(); mi!=idx4ngram.end(); mi++) {
		vector<int> v = mi->first;
		write_vec_int(fout, v);
		fwrite((void *) &(mi->second), sizeof(int), 1, fout);
	}

	ERR_check(ERR::ERR_SANITY, (size_t)s == ngram4idx.size(), "wrote incomplete ngram4idx while writing index");
	fwrite((void *) &s, sizeof(int), 1, fout);
	for (map<int, vector<int> >::iterator mi=ngram4idx.begin(); mi!=ngram4idx.end(); mi++) {
		fwrite((void *) &(mi->first), sizeof(int), 1, fout);
		vector<int> v = mi->second;
		write_vec_int(fout, v);
	}

	s = index.size();
	fwrite((void *) &s, sizeof(int), 1, fout);
	for (int i=0; i<s; i++) {
		idx_entry &a = index[i];
		int ts = a.tfidfvec.size();
		fwrite((void *) &ts, sizeof(int), 1, fout);
		if (ts)
			fwrite((void *) &(a.tfidfvec[0]), sizeof(pair<int, float>), ts, fout);
		fwrite((void *) &a.label, sizeof(int), 1, fout);
	}

	s = docs4idx.size();
	fwrite((void *) &s, sizeof(int), 1, fout);
	for (int i=0; i<s; i++)
		write_vec_int(fout, docs4idx[i]);

	s = xm4.size();
	fwrite((void *) &s, sizeof(int), 1, fout);
	for (int i=0; i<s; i++) {
		int sx = xm4[i].size();
		fwrite((void *) &sx, sizeof(int), 1, fout);
		if (sx) {
			for (set<vector<int> >::iterator si=xm4[i].begin(); si!=xm4[i].end(); si++) {
				vector<int> temp(*si);
				write_vec_int(fout, temp);
			}
		}
	}
}


void fastMatch::read_index(const char *fname) {

	FILE *fin = fopen(fname, "rb");
	if (!fin) {
		cout << "Unable to open " << fname << endl;
		exit(1);
	}

	// read the easy things first
	read_string2int_map(fin, idx4word);
	read_int2string_map(fin, word4idx);
	read_string2int_map(fin, idx4unit);
	read_int2string_map(fin, unit4idx);
	read_vec_float(fin, length4word);
	read_vec_float(fin, idf);
	read_vec_int(fin, noccs4doc);

	fread((void *) &nglen, sizeof(int), 1, fin);
	fread((void *) &ngnum, sizeof(int), 1, fin);
	scratch.resize(ngnum);

	// now for the tough ones

	int s, w, ts; 
	vector<int> v;
	fread((void *) &s, sizeof(int), 1, fin);
	for (int i=0; i<s; i++) {
		read_vec_int(fin, v);
		fread((void *) &w, sizeof(int), 1, fin);
		idx4ngram[v] = w;
	}

	fread((void *) &s, sizeof(int), 1, fin);
	ERR_check(ERR::ERR_SANITY, (size_t)s == idx4ngram.size(), "read incomplete ngram4idx while reading index");
	for (int i=0; i<s; i++) {
		fread((void *) &w, sizeof(int), 1, fin);
		read_vec_int(fin, v);
		ngram4idx[w] = v;
	}

	fread((void *) &s, sizeof(int), 1, fin);
	index.resize(s);
	for (int i=0; i<s; i++) {
		fread((void *) &ts, sizeof(int), 1, fin);
		index[i].tfidfvec.resize(ts);
		if (ts)
			fread((void *) &(index[i].tfidfvec[0]), sizeof(pair<int, float>), ts, fin);
		fread((void *) &(index[i].label), sizeof(int), 1, fin);
	}

	fread((void *) &s, sizeof(int), 1, fin);
	docs4idx.resize(s);
	for (int i=0; i<s; i++)
		read_vec_int(fin, docs4idx[i]);

	fread((void *) &s, sizeof(int), 1, fin);
	xm4.resize(s);
	for (int i=0; i<s; i++) {
		int sx;
		fread((void *) &sx, sizeof(int), 1, fin);
		if (sx) {
			for (int j=0; j<sx; j++) {
				read_vec_int(fin, v);
				xm4[i].insert(v);
			}
		}
	}
}

	


