// Geoffrey Zweig (gzweig@microsoft.com)
// Patrick Nguyen (panguyen@microsoft.com)
// Copyright (c) 2009 Microsoft Corporation. All rights reserved.

#include <assert.h>
#include <iostream>
#include "futil.h"
#include "ngutil.h"

void write_string2int_map(FILE *fout, map<string, int> &mapping)
{
	ERR_check(ERR::ERR_CONDITION, fout, "trying to write mapping to closed file handle");
	int s = int(mapping.size());
	fwrite((void *) &s, sizeof(int), 1, fout);
	for (map<string, int>::iterator mi=mapping.begin(); mi!=mapping.end(); mi++) {
		s = int(strlen(mi->first.c_str())) + 1;  // include terminal \0
		fwrite((void *) &s, sizeof(int), 1, fout);
		fwrite((void *) mi->first.c_str(), sizeof(char), s, fout);
		s = mi->second;
		fwrite((void *) &s, sizeof(int), 1, fout);
	}
}

void write_int2string_map(FILE *fout, map<int, string> &mapping) {
	ERR_check(ERR::ERR_CONDITION, fout, "trying to write mapping to closed file handle");
	int s = int(mapping.size());
	fwrite((void *) &s, sizeof(int), 1, fout);
	for (map<int, string>::iterator mi=mapping.begin(); mi!=mapping.end(); mi++) {
		s = mi->first;
		fwrite((void *) &s, sizeof(int), 1, fout);
		s = int(strlen(mi->second.c_str())) + 1;  // include terminal \0
		fwrite((void *) &s, sizeof(int), 1, fout);
		fwrite((void *) mi->second.c_str(), sizeof(char), s, fout);
	}
}

void write_vec_float(FILE *fout, vector<float> &v)
{
	ERR_check(ERR::ERR_CONDITION, fout, "trying to write vector of floats to closed file handle");
	int s = int(v.size());
	fwrite((void *) &s, sizeof(int), 1, fout);
	if(s) 
		fwrite((void *) &v[0], sizeof(float), s, fout);
}

void write_vec_double(FILE *fout, vector<double> &v)
{
	ERR_check(ERR::ERR_CONDITION, fout, "trying to write vector to closed file handle");
	int s = int(v.size());
	fwrite((void *) &s, sizeof(int), 1, fout);
	if(s)
		fwrite((void *) &v[0], sizeof(double), s, fout);
}

void write_vec_string(FILE *fout, vector<string> &v)
{
	ERR_check(ERR::ERR_CONDITION, fout, "trying to write vector to closed file handle");
	int s = int(v.size());
	fwrite((void *) &s, sizeof(int), 1, fout);
	if(s)
		for (int i=0; i<s; i++) {
			int sz = int(strlen(v[i].c_str())) + 1;  // include terminal \0
			fwrite((void *) &sz, sizeof(int), 1, fout);
			fwrite((void *) v[i].c_str(), sizeof(char), sz, fout);
		}
}

void write_vec_int(FILE *fout, vector<int> &v)
{
	ERR_check(ERR::ERR_CONDITION, fout, "trying to write vector to closed file handle");
	int s = int(v.size());
	fwrite((void *) &s, sizeof(int), 1, fout);
	if(s) 
		fwrite((void *) &v[0], sizeof(int), s, fout);
}

void read_string2int_map(FILE *fin, map<string, int> &mapping)
{
	ERR_check(ERR::ERR_CONDITION, fin, "trying to read mapping to closed file handle");
	mapping.clear();
	int sz;
	fread((void *) &sz, sizeof(int), 1, fin);
	const int slen = 1000;
	char buff[slen];
	for (int i=0; i<sz; i++) {
		int s;
		fread((void *) &s, sizeof(int), 1, fin);
		ERR_check(ERR::ERR_SANITY, s<slen, "buffer overflow while reading string");
		fread((void *) buff, sizeof(char), s, fin); 
		fread((void *) &s, sizeof(int), 1, fin);
		mapping[string(buff)] = s;
	}
}

void read_int2string_map(FILE *fin, map<int, string> &mapping)
{
	ERR_check(ERR::ERR_CONDITION, fin, "trying to read mapping to closed file handle");
	mapping.clear();
	int sz;
	fread((void *) &sz, sizeof(int), 1, fin);
	const int slen = 1000;
	char buff[slen];
	for (int i=0; i<sz; i++) {
		int s,sl;
		fread((void *) &s, sizeof(int), 1, fin);
		fread((void *) &sl, sizeof(int), 1, fin);
		ERR_check(ERR::ERR_SANITY, sl<slen, "buffer overflow while reading string mapping");
		fread((void *) buff, sizeof(char), sl, fin);
		mapping[s] = string(buff);
	}
}

void read_vec_string(FILE *fin, vector<string> &v)
{
	ERR_check(ERR::ERR_CONDITION, fin, "trying to read vector to closed file handle");
	int s;
	const int slen = 1000;
	char buff[slen];
	fread((void *) &s, sizeof(int), 1, fin);
	v.resize(s);
	if(s)
		for (int i=0; i<s; i++) {
			int sz;
			fread((void *) &sz, sizeof(int), 1, fin);
			ERR_check(ERR::ERR_CONDITION, sz<slen, "buffer overflow while trying to read string in vector");
			fread((void *) buff, sizeof(char), sz, fin);
			v[i] = string(buff);
		}
}



void read_vec_float(FILE *fin, vector<float> &v) {
	int s;
	fread((void *) &s, sizeof(int), 1, fin);
	v.resize(s);
	if(s)
		fread((void *) &v[0], sizeof(float), s, fin);
}

void read_vec_double(FILE *fin, vector<double> &v) {
	int s;
	fread((void *) &s, sizeof(int), 1, fin);
	v.resize(s);
	if(s)
		fread((void *) &v[0], sizeof(double), s, fin);
}

void accum_vec_float(FILE *fin, vector<float> &v) {
	int s;
	fread((void *) &s, sizeof(int), 1, fin);
	if (v.size() == 0) {
		v.resize(s);
		fill(v.begin(), v.end(), 0.0f);
	} else 
		ERR_check(ERR::ERR_CONDITION, v.size() == (size_t)s, "reading float vector of inconsistent size");
	vector<float> temp(s);
	if(s)
		fread((void *) &temp[0], sizeof(float), s, fin);
	for (int i=0; i<s; i++)
		v[i] += temp[i];
}

void accum_vec_double(FILE *fin, vector<double> &v) {
	int s;
	fread((void *) &s, sizeof(int), 1, fin);
	if (v.size() == 0) {
		v.resize(s);
		fill(v.begin(), v.end(), 0.0f);
	} else 
		ERR_check(ERR::ERR_CONDITION, v.size() == (size_t)s, "reading float vector of inconsistent size");
	vector<double> temp(s);
	if(s)
		fread((void *) &temp[0], sizeof(double), s, fin);
	for (int i=0; i<s; i++)
		v[i] += temp[i];
}

void read_vec_int(FILE *fin, vector<int> &v) {
	int s = 0;
	const size_t ret = fread((void *) &s, sizeof(int), 1, fin);
	ERR_check(ERR::ERR_SANITY, ret == 1, "failed to read vector size");
	v.resize(s);
	if(s)
		fread((void *) &v[0], sizeof(int), s, fin);
}

void write_float(FILE *fout, float f) {
	fwrite((void *) &f, sizeof(float), 1, fout);
}

void read_float(FILE *fin, float &f) {
	fread((void *) &f, sizeof(float), 1, fin);
}

void write_double(FILE *fout, double f) {
	fwrite((void *) &f, sizeof(double), 1, fout);
}

void read_double(FILE *fin, double &f) {
	fread((void *) &f, sizeof(double), 1, fin);
}

/**
 * parse command-line arguments. They must be of the form --cfg val
 * @return the number of arguments that were parsed (including argv[0], ie the index of the next arg)
 */
int CFG::parseCommandLine(std::map<std::string, std::string> &args, const int argc, char **argv) {
    int k;

    for(k=1; k < argc; k+=2) {
        const int l = (int)strlen(argv[k]);
        ERR_check(ERR::ERR_USERINPUT, l >= 2, "arguments must come in pairs");
        /* terminate parsing */
        if( !strcmp(argv[k], "--") ) {
            break;
        }
        if(k+1 >= argc) {
            (void)fprintf(stderr, "%s missing argument\n", argv[k]);
            ERR_check(ERR::ERR_USERINPUT, 0, "missing argument");
        }
        if( strncmp(argv[k], "--", 2) ) {
            (void)fprintf(stderr, "Argument is bogus: %s\n", argv[k]);
			ERR_check(ERR::ERR_ALWAYS, 0, "Failed because we were expecting a --arg");
        }
        if( !strcmp(argv[k], "--include") )
            CFG::parseConfigFile(args, argv[k+1]);
        args[argv[k]+2] = argv[k+1];
    }
    return k;
}

/**
 * load association map from space-separated file
 * allow the include directive
 */
void CFG::parseConfigFile(std::map<std::string, std::string> &args, char *fname) {
	IO::auto_file_t f(fname, "rt");
	IO::line_iterator<> line(f);

	if( !f )
		(void)fprintf(stderr, "Cannot open config file %s\n", fname);

	while(line.next()) {
        char *p = strchr(line.getline(), ' ');
        if( !p ) {
            cerr << ": Parse error in " << fname << ": buf does not contain exaclty two space-separated values: " << line.getline() << endl;
            exit(1);
        }
        *p = 0;
        p++;
        if( !strcmp(line.getline(), "include") )
            CFG::parseConfigFile(args, p);
        else
            args[line.getline()] = p;
    }
    return;
}
/**
 * convert a string of characters into bool
 */
bool CFG::str2bool(const char *str) {
	if(!strcmp(str, "true"))
		return true;
	else {
		if( !strcmp(str, "false") )
			return false;
		return !!atoi(str);
	}
}
