// Geoffrey Zweig (gzweig@microsoft.com)
// Copyright (c) 2009 Microsoft Corporation. All rights reserved.

#include "lm.h"

#include <stdio.h>
#include <iostream>
#include <assert.h>
#include <float.h>
#include <math.h>
#include <string.h>
#include <algorithm>

#include "ngutil.h"

#ifndef _GCC_
# pragma warning( disable:4996 )
#endif

using namespace std;

const string silence_word = "~SIL";
const float silence_logp = -0.434f * log(10.0f);
int silence_index;
bool seen_sil = false;
const string BOS = "<s>";
const string EOS = "</s>";
const string unknown_word = "<unk>";

language_model::language_model(const char *lmfile, map<string, int> &idx4word, map<int, string> &word4idx)
{
	// all LM logps are expected to be log base 10
    IO::auto_file_t fin(lmfile, "rt");

    const int slen = 1000;
    char buff[slen];
    int ngramlen = -1, nm=-1;
	max_ngramlen = -1;
	loc4.clear();
	int ngramlistsize = 0;

	ERR_check(ERR::ERR_SANITY, word4idx.size() == idx4word.size(), "[lm] word to index forward and reverse mapping have different number of associations");
	int wnum = 0;  // words indexed negatively
	for (map<string, int>::iterator mi=idx4word.begin(); mi!=idx4word.end(); mi++)
		wnum = (wnum<mi->second)?(wnum):(mi->second);

	// put in the null history state
	map<vector<int>, lmhist *> ngram4;
	lmhist null_history;
	ngrams.push_back(null_history);
	ngramlistsize++;
	const vector<int> null_vec;
	ngram4[null_vec] = &ngrams.back();
	initial_state = ngramlistsize-1;
	ngrams.back().idx = ngramlistsize-1;
	loc4.push_back(&ngrams.back());

    while (fgets(buff, slen, fin))
    {
		if( !strncmp(buff, "\\data\\", 6) ) continue;
		if( !strncmp(buff, "\\end\\", 5) )  break;
		if (strstr(buff, "ngram ") == buff)
		{
			char *p = strchr(buff, ' ');
			ERR_check(ERR::ERR_SANITY, p, "[lm] parse error in header: no space after ngram");
			char *p2 = strchr(p, '=');
			ERR_check(ERR::ERR_SANITY, p2, "[lm] parse error in header: no '=' after ngram");
			*p2 = '\0';
			nm = atoi(++p);
			if (nm>max_ngramlen)
                max_ngramlen = nm;
            continue;
		}
        else if (buff[0] == '\\')
        {
            char *p = strchr(buff, '-');
			ERR_check(ERR::ERR_SANITY, p, "[lm] error parsing \\n-gram: line");
            *p = '\0';
            ngramlen = atoi(buff+1);
            if (ngramlen>max_ngramlen)
                max_ngramlen = ngramlen;
        }
        else if (buff[0]=='-' || buff[0]=='0')
        {
            char *p = strpbrk(buff, " \t");
			ERR_check(ERR::ERR_USERINPUT, p, "[lm] looking for TAB character in ngram entry");
            *p = '\0';
            float nglogp = (float)atof(buff);
            vector<int> ngram;
			string w;
            for (int i=0; i<ngramlen; i++)
            {
                char *p2 = strpbrk(++p, " \t\n\r");
				ERR_check(ERR::ERR_USERINPUT, p, "[lm] missing words in ngram line");
                *p2 = '\0';
				w.assign(p);
				if (w == silence_word)
					seen_sil = true;
				if (!idx4word.count(w)) {
                    idx4word[w] = --wnum; // words labeled with negative integers
					word4idx[wnum] = w;
				}
				lm_words.insert(idx4word[w]);
                ngram.push_back(idx4word[w]); 
                p = p2;
            }
			// add the suffix as a target ngram
			vector<int> tgt = (ngram.size()==(size_t)max_ngramlen)?(vector<int>(ngram.begin()+1, ngram.end())):(ngram);
			if (ngram.size() == (size_t)max_ngramlen) {
				// should have seen the ending n-1gram 
				if (!ngram4.count(tgt)) {
					cout << "Error: the ngram\n";
					for (size_t t=0; t<ngram.size(); t++)
						cout << word4idx[ngram[t]] << " ";
					cout << endl << "is present, but\n";
					for (size_t t=0; t<tgt.size(); t++)
						cout << word4idx[tgt[t]] << " ";
					cout << "\nis not\n" << flush;
					exit(1);
				}
			} else {
				ERR_check(ERR::ERR_SANITY, !ngram4.count(tgt), "[lm] ngram already inserted");
				ngrams.push_back(lmhist());
				ngramlistsize++;
				if (word4idx[ngram.back()] == EOS)
					final_states.insert(ngramlistsize-1);
				ngrams.back().lastword = ngram.back();
				ngram4[tgt] = &ngrams.back();
				ngrams.back().backoff.logp = ((float)atof(++p))*log(10.0f);
				vector<int> botgt(tgt.begin()+1, tgt.end());
				// should have the suffix as a n-1 gram
				if (!ngram4.count(botgt)) {
					cout << "Error: the ngram\n";
					for (size_t t=0; t<tgt.size(); t++)
						cout << word4idx[tgt[t]] << " ";
					cout << endl << "is present, but\n";
					for (size_t t=0; t<botgt.size(); t++)
						cout << word4idx[botgt[t]] << " ";
					cout << "\nis not\n" << flush;
					exit(1);
				}
				ngrams.back().backoff.succ = ngram4[botgt];	
				ngrams.back().idx = ngramlistsize-1;
				loc4.push_back(&ngrams.back());
			}
			
			// add the ngram as an extension to its predecessor
			int last = ngram.back();
			ngram.pop_back();
			ERR_check(ERR::ERR_SANITY, ngram4.count(ngram), "[lm] should have seen prefix as (n-1)-gram");
			ngram4[ngram]->succs.push_back(succ_rec(last, ngram4[tgt], nglogp*log(10.0f)));
        }
        else
        {
			if (!isspace(buff[0]) && buff[0]) {
				cout << "expecting space but saw:\n" << buff << endl;
				exit(1);
			}
            continue;
        }
    }
    fin.close();

	if (max_ngramlen == 1)  // a unigrma LM - no </s> state
		final_states.insert(initial_state);

	if (!seen_sil && idx4word.count(silence_word)) {  // silence in graph but not LM
		cout << "Warning: no " << silence_word << " in the LM\n";
		cout << "Adding it with logp " << silence_logp << endl;
	}

	if (!idx4word.count(silence_word)) {
		idx4word[silence_word] = --wnum;
		word4idx[wnum] = silence_word;
	}

	silence_index = idx4word[silence_word];
        sil_idx = silence_index;
	lm_words.insert(silence_index);
	word4int = &word4idx;

	if (!idx4word.count(unknown_word)) {
		cout << "LM must contain " << unknown_word << " and doesn't\n";
		exit(1);
	}
	unk_idx = idx4word[unknown_word];

	if (!idx4word.count(BOS)) {
		cout << "LM must contain " << BOS << endl;
		exit(1);
	}

	if (!idx4word.count(EOS)) {
		cout << "LM must contain " << EOS << endl;
		exit(1);
	}
	bos_idx = idx4word[BOS];
	eos_idx = idx4word[EOS];

	for (list<lmhist>::iterator li = ngrams.begin(); li!=ngrams.end(); li++) 
		sort(li->succs.begin(), li->succs.end());

	for (list<lmhist>::iterator li = ngrams.begin(); li!=ngrams.end(); li++) 
		ERR_check(ERR::ERR_SANITY, loc4[li->idx] == &(*li), "[lm] location of ngram is inconsistent");

	cout << "LM read with " << ngrams.size() << " history states\n";
}

const string &language_model::unk_wrd() {
	return unknown_word;
}

lmidx language_model::lm_next_hist(lmidx h, int l)
{
	path.clear();
	path.push_back(h);
	if (l == silence_index)  // no change by adding silence
		return h;
	lmhist *ch = loc4[h];
	while (1) {
		vector<succ_rec>::iterator vi = lower_bound(ch->succs.begin(), ch->succs.end(), succ_rec(l));
		if (vi!=ch->succs.end() && vi->wdidx == l) {
			path.push_back(vi->succ->idx);
			return vi->succ->idx;
		}
		else {
			path.push_back(ch->backoff.succ->idx);
			ch = ch->backoff.succ;
		}
	}
}

float language_model::lm_logp(lmidx h, int l)
{
	if (!seen_sil && l==silence_index) // asking about silence and its not in the LM
		return silence_logp;
	lmhist *ch = loc4[h];
	float net = 0;
	while (1) {
		vector<succ_rec>::iterator vi = lower_bound(ch->succs.begin(), ch->succs.end(), succ_rec(l));
		if (vi!=ch->succs.end() && vi->wdidx == l) 
			return net + vi->logp;
		else {
			net += ch->backoff.logp;
			ch = ch->backoff.succ;
		}
	}
}
