// Geoffrey Zweig (gzweig@microsoft.com)
// Patrick Nguyen (panguyen@microsoft.com)
// Copyright (c) 2009, 2010 Microsoft Corporation. All rights reserved.
//
#include <math.h>
#include <iostream>
#include <stdio.h>
#include <algorithm>
#include "model.h"
#include "edit_distance.h"
#include "futil.h"

#ifndef _GCC_
# define _WIN32_WINNT 0x0503
# define WINVER 0x0503
# include <windows.h>
#endif // _GCC_

#include "mpiwrap.h"

static const float LOG_0 = -1e30F;
static const float LOG_1 = 0.0F;

#ifndef _GCC_
# pragma warning( disable:4996 )
#endif

void logadd(double &target, double inc) {
	if (target < inc)
		swap(target, inc);
    target += log(1 + exp(inc-target));
}

double model::fsum(const vector<feature> &f) const {
	double tot = 0;
	for (size_t i=0; i<f.size(); i++)
		tot += lambda[f[i].idx] * f[i].val;
	return tot;
}

// @return zero if the sum of elements is zero
static bool is_zero_array(const vector<size_t> &vec) {
	return accumulate(vec.begin(), vec.end(), 0) == 0;
}

void model::compute_alphas(mode _mode) {
	alphas.clear();
	alphas.resize(uttIt->cur_maxtime()+1);  // using 0..maxtime
	pred.clear();
	alphas[0][lmrec(LM->initial_state, LM->bos_idx, time_lat_rec::lstart)] = LOG_1;
	feature_cache.clear();
	leven_cache.clear();
	exist_cache.clear();
	expect_cache.clear();
	for (int i=0; i<uttIt->cur_maxtime(); i++) { // the last wordspans will push forward to maxtime

		map<lmrec,double>::iterator mi;

		// determine the beam
		vector<double> probs;
		for (mi=alphas[i].begin(); mi!=alphas[i].end(); mi++)
			probs.push_back(mi->second);

		double limit = -FLT_MAX;
		if ((int)probs.size() > max_alphas) {
			nth_element(probs.begin(), probs.begin()+max_alphas-1, probs.end(), greater<float>());
			limit = probs[max_alphas-1];
		}

		// advance the alphas
		for (mi=alphas[i].begin(); mi!=alphas[i].end(); mi++) {
			if (mi->second < limit) 
				continue;

			const lmidx s = mi->first.state;
			const int pw = mi->first.w;
			const double av = mi->second;
			const int lp = mi->first.lat_end_state;

			const vector<wordspan> candidates = start(time_lat_rec(i+1,lp));
			for (size_t c=0; c<candidates.size(); c++) {
				const int w = candidates[c].idx;
				const int et = candidates[c].endtime;
				const int elp = candidates[c].lat_end_state;
				map<lmrec,double> &a = alphas[et];
				const pair<lmidx, float> nextstate = succ(s,w); // next LM state + logprob
				const lmidx ns = nextstate.first;
				const float lmprob = nextstate.second;
				const vector<feature> &f = features(mi->first, lmrec(ns, w, elp), i+1, et, 
					lp, elp, lmprob, w, 
					&LM->path, candidates[c].external.size() == 0 ? 0 : &candidates[c].external);
				if (a.find(lmrec(ns,w,elp))==a.end()) {
					a[lmrec(ns,w,elp)] = LOG_0;
					if (_mode == viterbi) 
						pred[statetime(ns,et,w,elp)].resize(1); // will keep exactly 1 pred
				}
				if (_mode == sum) {
					logadd(a[lmrec(ns,w,elp)], av + fsum(f));
					pred[statetime(ns,et,w,elp)].push_back(statetime(s,i,pw,lp));
				} else {
					double candprob = av + fsum(f);
					if (candprob > a[lmrec(ns,w,elp)]) {
						a[lmrec(ns,w,elp)] = candprob;
						pred[statetime(ns,et,w,elp)][0] = statetime(s,i,pw,lp);
					}
				}
			}
		}
	}
}

double model::compute_betas() {
	betas.clear();
	betas.resize(uttIt->cur_maxtime()+1);

	// initialize the betas, for reachable states
	map<lmrec,double>::iterator mi;
	for (mi=alphas[uttIt->cur_maxtime()].begin(); mi!=alphas[uttIt->cur_maxtime()].end(); mi++)
		if (LM->final_states.count(mi->first.state) && mi->first.lat_end_state==time_lat_rec::lend) 
			betas[uttIt->cur_maxtime()][lmrec(mi->first.state, mi->first.w, mi->first.lat_end_state)] 
			= LOG_1;

	for (int i=uttIt->cur_maxtime(); i>0; i--) {  // the first wordspan will push back to time 0
		for (mi=betas[i].begin(); mi!=betas[i].end(); mi++) {
			const lmidx s = mi->first.state;
			const int cw = mi->first.w;
			const int lp = mi->first.lat_end_state;
			const double bv = mi->second;
			const vector<statetime> &preds = pred[statetime(s,i,cw,lp)];
			for (size_t p=0; p<preds.size(); p++) {
				const lmidx ps = preds[p].state;
				const int st = preds[p].t;
				const int pw = preds[p].w;
				const int plp = preds[p].latpos;
				const vector<feature> &f = features(lmrec(ps, pw, plp), mi->first, st+1, i, plp, lp);
				map<lmrec,double> &b = betas[st];
				if (b.find(lmrec(ps,pw,plp))==b.end()) 
					b[lmrec(ps,pw,plp)] = LOG_0;
				logadd(b[lmrec(ps,pw,plp)], bv+fsum(f));
			}
		}
	}
	return betas[0][lmrec(LM->initial_state, LM->bos_idx, time_lat_rec::lstart)];
}

void model::compute_gradient(float sign) {
	double Z = LOG_0;
	map<lmrec,double>::iterator mi;
	for (mi=alphas[uttIt->cur_maxtime()].begin(); mi!=alphas[uttIt->cur_maxtime()].end(); mi++)
		if (LM->final_states.count(mi->first.state) && mi->first.lat_end_state==time_lat_rec::lend)
			logadd(Z, alphas[uttIt->cur_maxtime()][lmrec(mi->first.state,mi->first.w,time_lat_rec::lend)]);
    const double ZB = betas[0][lmrec(LM->initial_state,LM->bos_idx,time_lat_rec::lstart)];
	const double logratio = (Z-ZB) / (fabs(Z) > 1.0f? fabs(Z): 1.0f);
	if ( fabs(logratio) > 3e-2f ) { 
		cerr << "forward and backward computations disagree: " << Z << " " << ZB << " on file: " << uttIt->cur_constraintFN() << endl;
		exit(1);
	}

	for (int i=uttIt->cur_maxtime(); i>0; i--) {
		map<lmrec,double> &b = betas[i];
		map<lmrec,double>::iterator mi;
		for (mi=b.begin(); mi!=b.end(); mi++) {
			lmidx s = mi->first.state;
			int w = mi->first.w;
			int lp = mi->first.lat_end_state;
			vector<statetime> &preds = pred[statetime(s,i,w,lp)];
			for (size_t p=0; p<preds.size(); p++) {
				lmidx ps = preds[p].state;
				int st = preds[p].t;
				int pwi = preds[p].w;
				int plp = preds[p].latpos;
				vector<feature> &f = features(lmrec(ps, pwi, plp), mi->first, st+1, i, plp, lp);
				map<lmrec, double> &a = alphas[st];
				if (a.find(lmrec(ps,pwi,plp))==a.end()) {
					cerr << "inconsistency in gradient computation: expected alpha not found\n";
					exit(1);
				}
				double pw = exp(a[lmrec(ps,pwi,plp)] + b[lmrec(s,w,lp)] + fsum(f) - Z);  // posterior of wordspan

				for (size_t k=0; k<f.size(); k++) 
					gradient[f[k].idx] += sign * f[k].val * pw;  // gradient not log gradient
			}
		}
	}
}

vector< pair<int,int> > model::backtrace() {
	dataprob = LOG_0;
	lmidx endstate = -1; // impossible: >= 0
	vector< pair<int, int> > words;
	int lastword = LM->eos_idx;
	int latp = -1;
	map<lmrec,double>::iterator mi;
	for (mi=alphas[uttIt->cur_maxtime()].begin(); mi!=alphas[uttIt->cur_maxtime()].end(); mi++)
		if (LM->final_states.count(mi->first.state) && mi->first.lat_end_state==time_lat_rec::lend 
			&& mi->second > dataprob) {
				dataprob = mi->second;
				endstate = mi->first.state;
				lastword = mi->first.w;
				latp = mi->first.lat_end_state;
		}

	if (endstate == -1) {
		return words;  // empty vector
	}
	ERR_check(ERR::ERR_SANITY, latp != -1, "[backtrace] end state and latp inconsistent");
	statetime pr = pred[statetime(endstate,uttIt->cur_maxtime(),lastword,latp)][0];
	words.push_back( pair<int,int>(lastword, uttIt->cur_times()[pr.t-1]) ); 
	while (pr.w != LM->bos_idx) {
		const int w = pr.w;
		pr = pred[pr][0];
		words.push_back( pair<int,int>(w, uttIt->cur_times()[pr.t-1]) );
	} 
	if (words.back().first != LM->bos_idx)  // sometimes true, when <s> </s> is all 
		words.push_back(pair<int,int>(LM->bos_idx, 0));                   

	return words;
}

void model::read_dicts(vector<string> dicts, int &unitnum, bool model_trained) {
	sym4stream_idx.resize(num_streams);
	prons4stream_word.resize(num_streams);
	idx4stream_sym.resize(num_streams);
	const int slen=10000;
	char buff1[slen], buff2[slen], buff3[slen];

	// see what the minimum word index is, in case we need to add words
	int minwd = vocab_util::get_minwd(idx4word);
	for (size_t i=0; i<dicts.size(); i++) {
		IO::auto_file_t fin(dicts[i].c_str(), "rt");
		fscanf(fin, "%s%s%s\n", buff1, buff2, buff3);
		ERR_check(ERR::ERR_USERINPUT, buff1[0] == '#', "[model] ill-formed header in dictionary: does not start with #");
		ERR_check(ERR::ERR_USERINPUT, !strcmp(buff3, "dictionary"), "[model] dictionary does not have 'dictionary' name in header");
		if (idx4streamname.find(string(buff2)) == idx4streamname.end()) {
			cerr << "[model] dictionary name " << buff2 << " does not match a stream name\n";
			exit(1);
		}
		int currstream = idx4streamname[string(buff2)];
		ERR_check(ERR::ERR_SANITY, currstream < num_streams, "[model] buffer overlow in stream number.");
		int warn_count=0, warn_limit=10;
		while(fgets(buff1,slen,fin)) {
			char *p = strtok(buff1, " \n");  // get the word
			ERR_check(ERR::ERR_USERINPUT, p, "[model] missing separator in dictionary entry");
			string ps(p);
			if (idx4word.find(ps) == idx4word.end()) {  
				idx4word[ps] = --minwd;
				word4idx[minwd] = ps;
				if (warn_count < warn_limit)   // situation is reasonable when LM is small
					cout << "Warning: " << ps << " in dictionary but not seen before\n";
                                else if (warn_count == warn_limit) 
					cout << "Suppressing further dictionary warnings\n";
				warn_count++;
			}  // language model was read prior to this
			int cw = idx4word[ps];
			vector<int> pron;
			while ( (p = strtok(NULL, " \n")) != NULL) {
				string ps(p);
				if (idx4stream_sym[currstream].find(ps) == idx4stream_sym[currstream].end()) {
					if (model_trained) {
						cerr << "Cannot add new detector units to a trained model\n";
						cerr << "Found " << ps << " not in model\n";
						exit(1);
					}
					idx4stream_sym[currstream][ps] = unitnum;
					sym4stream_idx[currstream][unitnum] = ps;
					unitnum++;
				}
				pron.push_back(idx4stream_sym[currstream][ps]);
			}
			prons4stream_word[currstream][cw].push_back(pron);
		}
	}
	cout << "dictionaries read\n";
}

void model::number_levenshtein(const std::vector<std::string > &detLists) {
	// create the levenshtein features
	int sn = featurenum;
	for (int i=0; i<int(idx4stream_sym.size()); i++) {  // for each stream
		map<string, int>::iterator mi;
		for (mi=idx4stream_sym[i].begin(); mi!=idx4stream_sym[i].end(); mi++) {
			int u = mi->second;
			feature_bundle f;
			f.push_back(i);
			f.push_back(1);
			f.push_back(levenshtein);
			f.push_back(match);
			f.push_back(u);
			if (stream_feature_idx4.find(f) == stream_feature_idx4.end()) {
				stream_feature_idx4[f] = featurenum++;
			}
			f[3] = sub;
			if (stream_feature_idx4.find(f) == stream_feature_idx4.end()) {
				stream_feature_idx4[f] = featurenum++;
			}
			f[3] = del;
			if (stream_feature_idx4.find(f) == stream_feature_idx4.end()) {
				stream_feature_idx4[f] = featurenum++;
			}
			f[3] = ins;
			if (stream_feature_idx4.find(f) == stream_feature_idx4.end()) {
				stream_feature_idx4[f] = featurenum++;
			}
		}
	}

	// now read the input files
	const int slen=10000;
	char buff1[slen], buff2[slen], buff3[slen];
	// now read the input files
	size_t k;
	for(k=0; k < detLists.size(); ++k) {
		MultiFiles detFiles(detLists[k].c_str());
		while(detFiles.moveNextFile()) {
			const bool hasFirstLine = detFiles.moveNextLine();
			ERR_check(ERR::ERR_SANITY, hasFirstLine, "[fe] Detection file is unexpectedly empty");
			(void)sscanf(detFiles.curLine(), "%s%s%s", buff1, buff2, buff3);
			ERR_check(ERR::ERR_USERINPUT, buff1[0] == '#', "[fe] ill-formed header in detection stream file: must start with a '#'");
			ERR_check(ERR::ERR_USERINPUT, !strcmp(buff3, "stream"), "[fe] missing 'stream' keyword in detection header");
			ERR_check(ERR::ERR_USERINPUT, idx4streamname.find(string(buff2)) != idx4streamname.end(), "unknown stream name found in detection file");
			int currstream = idx4streamname[string(buff2)];
			ERR_check(ERR::ERR_SANITY, currstream < num_streams, "[fe] buffer overflow in stream id");
			int t;
			vector<int> units;
			while( detFiles.moveNextLine() ) {
				(void)sscanf(detFiles.curLine(), "%s%i", buff1, &t);
				string buffstr(buff1);
				ERR_check(ERR::ERR_USERINPUT, idx4stream_sym[currstream].find(buffstr) != idx4stream_sym[currstream].end(), "[fe] unknown symbol in detection stream");
				units.push_back(idx4stream_sym[currstream][buffstr]);
			}

			// in case a unit is missing from the dictionary, make insertion features for everything
			for (vector<int>::iterator u=units.begin(); u!=units.end(); u++) {
				feature_bundle f;
				f.push_back(currstream);
				f.push_back(1);
				f.push_back(levenshtein);
				f.push_back(ins);
				f.push_back(*u);
				if (stream_feature_idx4.find(f) == stream_feature_idx4.end()) 
					stream_feature_idx4[f] = featurenum++;
			}
		}
	}
	cout << (featurenum - sn) << " levenshtein features created\n";
}

void model::number_expectation(const std::vector<std::string > &detLists) {
	const int sn = featurenum;
	// go over the prons and create the expectation features
	for(int currstream=0; currstream<num_streams; currstream++) {
		map<int, vector<vector<int> > >::iterator mi;
		for (mi=prons4stream_word[currstream].begin(); mi!=prons4stream_word[currstream].end(); mi++) {
			for (size_t pnum=0; pnum<mi->second.size(); pnum++) {
				vector<int> &pron = mi->second[pnum];

				// create the expectation features
				for (int n=1; n<=ngramlevel4expectation_stream[currstream]; n++) {
					for (int s = 0; s<=int(pron.size())-n;s++) {
						feature_bundle f;
						f.push_back(currstream);
						f.push_back(n);
						f.push_back(expectation);
						f.push_back(correct_accept);
						f.insert(f.end(),pron.begin()+s,pron.begin()+s+n);
						if (stream_feature_idx4.find(f) == stream_feature_idx4.end()) 
							stream_feature_idx4[f] = featurenum++;
						f[3] = false_reject;
						if (stream_feature_idx4.find(f) == stream_feature_idx4.end()) 
							stream_feature_idx4[f] = featurenum++;
					}
				}
			}
		}
	}

	// now read the input files
	// make the false accept features by looking at the patterns that actually occur in the data
	const int slen=10000;
	char buff1[slen], buff2[slen], buff3[slen];
	size_t k;
	for(k=0; k < detLists.size(); ++k) {
		// now read the input files
		MultiFiles detFiles(detLists[k].c_str());
		while(detFiles.moveNextFile()) {
			const bool hasFirstLine = detFiles.moveNextLine();
			ERR_check(ERR::ERR_SANITY, hasFirstLine, "[fe] detection file is unexpectedly empty");
			sscanf(detFiles.curLine(), "%s%s%s", buff1, buff2, buff3);
			ERR_check(ERR::ERR_USERINPUT, buff1[0] == '#', "[fe] ill-formed header in detection file: must start with a '#'");
			ERR_check(ERR::ERR_USERINPUT, !strcmp(buff3, "stream"), "[fe] missing 'stream' keyword in detection header");
			ERR_check(ERR::ERR_USERINPUT, idx4streamname.find(string(buff2)) != idx4streamname.end(), "[fe] unknown stream name found in stream header");
			int currstream = idx4streamname[string(buff2)];
			ERR_check(ERR::ERR_SANITY, currstream < num_streams, "[fe] buffer overflow in stream id");
			int t;
			vector<int> units;
			while(detFiles.moveNextLine()) {
				(void)sscanf(detFiles.curLine(), "%s%i", buff1, &t);
				string buffstr(buff1);
				ERR_check(ERR::ERR_SANITY, idx4stream_sym[currstream].find(buffstr) != idx4stream_sym[currstream].end(), "[fe] unexpected stream name in detection file");
				units.push_back(idx4stream_sym[currstream][buffstr]);
			}

			// create the expectation features for false accepts
			// other types of expectation features already added when dictionary was read
			for (int n=1; n<=ngramlevel4expectation_stream[currstream]; n++) {
				for (int s = 0; s<=int(units.size())-n;s++) {
					feature_bundle f;
					f.push_back(currstream);
					f.push_back(n);
					f.push_back(expectation);
					f.push_back(false_accept);
					f.insert(f.end(),units.begin()+s,units.begin()+s+n);
					if (stream_feature_idx4.find(f) == stream_feature_idx4.end()) 
						stream_feature_idx4[f] = featurenum++;
				}
			}
		}
	}
	cout << (featurenum - sn) << " expectation features created\n";
}

void model::number_existence() {
	// go over the prons and create the existence features
	int sn = featurenum;
	for(int currstream=0; currstream<num_streams; currstream++) {
		map<int, vector<vector<int> > >::iterator mi;
		for (mi=prons4stream_word[currstream].begin(); mi!=prons4stream_word[currstream].end(); mi++) {
			int wnum = mi->first;
			for (size_t pnum=0; pnum<mi->second.size(); pnum++) {
				vector<int> &pron = mi->second[pnum];
				// create the existence features
				for (int n=1; n<=ngramlevel4existence_stream[currstream]; n++) {
					for (int s = 0; s<=int(pron.size())-n;s++) {
						feature_bundle f;
						f.push_back(currstream);
						f.push_back(n);
						f.push_back(existence);
						f.push_back(wnum);
						f.insert(f.end(),pron.begin()+s,pron.begin()+s+n);
						if (stream_feature_idx4.find(f) == stream_feature_idx4.end())
							stream_feature_idx4[f] = featurenum++;
					}
				}
			}
		}
	}

	(void)printf("[fe] creating existence features\n"); fflush(stdout);
	feature_bundle f;
	vector<int> units;

	int uttId = 0;

	while(uttIt->moveNext()) {
		uttIt->getNumConstraints();
		uttIt->skipDenConstraints();
		constraint_t::const_iterator latStates = uttIt->cur_words_starting().begin();

		for(; latStates != uttIt->cur_words_starting().end(); ++latStates) {
			vector<wordspan>::const_iterator links = latStates->second.begin();
			for(; links != latStates->second.end(); ++links) {
				const int starttime = links->starttime;
				const int endtime = links->endtime;
				const detection st(-1,starttime), et(-1, endtime);
				const int w = links->idx;
				for (int i=0; i<num_streams; i++) {
					vector<detection>::iterator first = 
						lower_bound(uttIt->cur_units4stream()[i].begin(), uttIt->cur_units4stream()[i].end(), st);
					vector<detection>::iterator last = 
						upper_bound(uttIt->cur_units4stream()[i].begin(), uttIt->cur_units4stream()[i].end(), et);

					vector<detection>::iterator vi;
					units.clear();
					for (vi=first; vi<last; vi++) 
						units.push_back(vi->unit);

					const int maxn = ngramlevel4existence_stream[i];
					for (int n=1; n<=maxn; n++) { 
						for (int s=0; s<=int(units.size())-n;s++) {
							// same feature can be added multiple times, but we'll unique later
							f.clear();
							f.push_back(i);
							f.push_back(n);
							f.push_back(existence);
							f.push_back(w);
							f.insert(f.end(),units.begin()+s,units.begin()+s+n);
							// new words (not used in model training) may not have features
							if (stream_feature_idx4.find(f) == stream_feature_idx4.end())
								stream_feature_idx4[f] = featurenum++;
						}
					}
				}
			}
		}
		if( !(++uttId%1000) ) {
			(void)printf("[fe] exist making finished %d files, created %d features\n", uttId, featurenum-sn);fflush(stdout);
		}
	}
	uttIt->rewind();
	cout << (featurenum-sn) << " existence features created\n";
}

void model::number_units_and_features(const vector<string> &dicts, const std::vector<std::string > &detLists) {

	// first read all the dictionary files
	int unitnum = 0;
    read_dicts(dicts, unitnum);
	ERR_check(ERR::ERR_USERINPUT, dicts.size() == (size_t)num_streams, "[fe] need exactly one dictionary per stream");

	featurenum = 0;
	// initialize the arrays that indicate what ngram level to use for the different kinds of 
	// features
	// go from the string representation set up by the command line
	// to an integer representation of the streams
	ngramlevel4existence_stream.resize(num_streams);
	ngramlevel4expectation_stream.resize(num_streams);
	for (map<string,int>::iterator mi = idx4streamname.begin(); mi!=idx4streamname.end(); mi++) {
		ngramlevel4existence_stream[mi->second] = nglev4exist_stream[mi->first];
		ngramlevel4expectation_stream[mi->second] = nglev4expect_stream[mi->first];
	}

	// number the units
	const int slen=10000;
	char buff1[slen], buff2[slen], buff3[slen];
	size_t k;
	for(k=0; k < detLists.size(); ++k) {
		MultiFiles detFiles(detLists[k].c_str());
		while(detFiles.moveNextFile()) {
			const bool hasFirstLine = detFiles.moveNextLine();
			ERR_check(ERR::ERR_USERINPUT, hasFirstLine, "[fe] detection file is unexpectedly empty");
			sscanf(detFiles.curLine(), "%s%s%s", buff1,  buff2,  buff3);
			ERR_check(ERR::ERR_USERINPUT, buff1[0] == '#', "[fe] ill-formed header in detection file: must start with a '#'");
			ERR_check(ERR::ERR_USERINPUT, !strcmp(buff3, "stream"), "[fe] missing 'stream' keyword in header");
			ERR_check(ERR::ERR_USERINPUT, idx4streamname.find(string(buff2)) != idx4streamname.end(), "[fe] unknown stream name in detection file");
			int currstream = idx4streamname[string(buff2)];
			ERR_check(ERR::ERR_SANITY, currstream < num_streams, "[fe] buffer overlow in stream id");
			int t;
			while(detFiles.moveNextLine()) {
				(void)sscanf(detFiles.curLine(), "%s%i", buff1, &t);
				const string buffstr(buff1);
				if (idx4stream_sym[currstream].find(buffstr) == idx4stream_sym[currstream].end()) {
					idx4stream_sym[currstream][buffstr] = unitnum;
					sym4stream_idx[currstream][unitnum] = buffstr;
					unitnum++;
				}
			}
		}
	}

	if (fexConfig.use_levenshtein_features.size())
		number_levenshtein(detLists);
	if ( !is_zero_array(fexConfig.existOrder) )
		number_existence();
	if ( !is_zero_array(fexConfig.expectOrder) )
		number_expectation(detLists);

	int minwd = vocab_util::get_minwd(idx4word);
	ERR_check(ERR::ERR_SANITY, -minwd==idx4word.size(), "word range expected to be -1 .. -V is inconsistent with index size");
	ERR_check(ERR::ERR_SANITY, LM->unk_idx < 0 && LM->unk_idx >= minwd, "<unk> is expected in the LM but not in the vocabulary range");
	idx4external.resize(fexConfig.external_format.size());
	if (fexConfig.external_format.size()) {// using externally defined features in annotated lattices
		for (size_t i=0; i<fexConfig.external_format.size(); i++) {
			idx4external[i] = featurenum;
			featurenum += fexConfig.lexternal[i] ? idx4word.size() + 1 : 1;
		}
		external_lexwords = (int)idx4word.size();
	}
}

bool model::in_prons(vector<int> &pat, vector<vector<int> > &prons) {
	int plen = (int) pat.size();
	for (size_t pn =0; pn<prons.size(); pn++) {
		int len = (int)prons[pn].size();
		for (int st=0; st<=len-plen; st++) {
			bool match = 1;
			for (int z=0; z<plen; z++)
				if (prons[pn][st+z] != pat[z]) {
					match = 0;
					break;
				}
			if (match)
				return true;
		}
	}
	return false;
}

void model::make_levenshtein_features(int starttime, int endtime, int w) {
	
	pair<int, int> times(starttime, endtime);
	pair<pair<int, int>, int> key(times, w);
	if (leven_cache.find(key) != leven_cache.end()) {
		vector<feature> &cf = leven_cache[key];
		feats.insert(feats.end(), cf.begin(), cf.end());
		return;
	}

	vector<feature> cf;
	ERR_check(ERR::ERR_SANITY, (size_t)num_streams == uttIt->cur_units4stream().size(), "[fe] inconsistent number of streams from current file");
	detection st(-1,starttime), et(-1, endtime);
	for (int i=0; i<num_streams; i++) {
		if (!prons4stream_word[i].count(w))
			continue;  // no pronunciation; no levenshtein features

		vector<detection>::iterator first = 
			lower_bound(uttIt->cur_units4stream()[i].begin(), uttIt->cur_units4stream()[i].end(), st);
		vector<detection>::iterator last = 
			upper_bound(uttIt->cur_units4stream()[i].begin(), uttIt->cur_units4stream()[i].end(), et);

		vector<detection>::iterator vi;
		vector<int> units;
		for (vi=first; vi<last; vi++) 
			units.push_back(vi->unit);

		vector<vector<int > > &prons = prons4stream_word[i][w];
		// now add the levenshtein features
		vector<int> u2(units.size()+1);
		u2[0] = -1;  // flag
		copy(units.begin(), units.end(), u2.begin()+1);

		// find the best pron
		int bc = INT_MAX;
		vector<alignop> bestalgn;
		vector<int> bestpron;
		for (size_t p=0; p<prons.size(); p++) {
			vector<int> pr(prons[p].size()+1);
			pr[0] = -1;  // flag
			copy(prons[p].begin(), prons[p].end(), pr.begin()+1);
			vector<alignop> alignment;
			int c = min_ed(pr, u2, alignment); 
			if (c<bc) {
				bc = c;
				bestalgn = alignment;
				bestpron = pr;
			}
		}

		map<int, int> ins4, del4, sub4, match4;
		for (size_t a=0; a<bestalgn.size(); a++) {
			if (bestalgn[a].op == ins)
				ins4[u2[bestalgn[a].hpos]]++;
			else if (bestalgn[a].op == del)
				del4[bestpron[bestalgn[a].rpos]]++;
			else if (bestalgn[a].op == sub) {
				if (u2[bestalgn[a].hpos] == bestpron[bestalgn[a].rpos])
					match4[u2[bestalgn[a].hpos]]++;
				else
					sub4[bestpron[bestalgn[a].rpos]]++;
			}
		}

		map<int, int>::iterator mi;
		for (mi=ins4.begin(); mi!=ins4.end(); mi++) {
			feature_bundle f;
			f.push_back(i);
			f.push_back(1);
			f.push_back(levenshtein);
			f.push_back(ins);
			f.push_back(mi->first);
			// all units should be present since new dictionaries are not allowed to add new units
			ERR_check(ERR::ERR_SANITY, stream_feature_idx4.find(f)!=stream_feature_idx4.end(), "[fe] unit found but not in dictionary");
			feats.push_back(feature(stream_feature_idx4[f], float(mi->second)));
			cf.push_back(feats.back());
		}
		for (mi=del4.begin(); mi!=del4.end(); mi++) {
			feature_bundle f;
			f.push_back(i);
			f.push_back(1);
			f.push_back(levenshtein);
			f.push_back(del);
			f.push_back(mi->first);
			ERR_check(ERR::ERR_SANITY, stream_feature_idx4.find(f)!=stream_feature_idx4.end(), "[fe] unit found but not in dictionary");
			feats.push_back(feature(stream_feature_idx4[f], float(mi->second)));
			cf.push_back(feats.back());
		}
		for (mi=sub4.begin(); mi!=sub4.end(); mi++) {
			feature_bundle f;
			f.push_back(i);
			f.push_back(1);
			f.push_back(levenshtein);
			f.push_back(sub);
			f.push_back(mi->first);
			ERR_check(ERR::ERR_SANITY, stream_feature_idx4.find(f)!=stream_feature_idx4.end(), "[fe] unit found but not in dictionary");
			feats.push_back(feature(stream_feature_idx4[f], float(mi->second)));
			cf.push_back(feats.back());
		}
		for (mi=match4.begin(); mi!=match4.end(); mi++) {
			feature_bundle f;
			f.push_back(i);
			f.push_back(1);
			f.push_back(levenshtein);
			f.push_back(match);
			f.push_back(mi->first);
			ERR_check(ERR::ERR_SANITY, stream_feature_idx4.find(f)!=stream_feature_idx4.end(), "[fe] unit found but not in dictionary");
			feats.push_back(feature(stream_feature_idx4[f], float(mi->second)));
			cf.push_back(feats.back());
		}
	}
	leven_cache[key] = cf;
}

void model::make_existence_features(int starttime, int endtime, int w) {

	pair<int, int> times(starttime, endtime);
	pair<pair<int, int>, int> key(times, w);
	if (exist_cache.find(key) != exist_cache.end()) {
		vector<feature> &cf = exist_cache[key];
		feats.insert(feats.end(), cf.begin(), cf.end());
		return;
	}

	vector<feature> cf;
	ERR_check(ERR::ERR_SANITY, (size_t)num_streams == uttIt->cur_units4stream().size(), "[fe] number of streams inconsistent in current file");
	detection st(-1,starttime), et(-1, endtime);
	for (int i=0; i<num_streams; i++) {
		vector<detection>::iterator first = 
			lower_bound(uttIt->cur_units4stream()[i].begin(), uttIt->cur_units4stream()[i].end(), st);
		vector<detection>::iterator last = 
			upper_bound(uttIt->cur_units4stream()[i].begin(), uttIt->cur_units4stream()[i].end(), et);

		vector<detection>::iterator vi;
		vector<int> units;
		for (vi=first; vi<last; vi++) 
			units.push_back(vi->unit);

		int maxn = ngramlevel4existence_stream[i];
		for (int n=1; n<=maxn; n++) { 
			for (int s=0; s<=int(units.size())-n;s++) {
				// same feature can be added multiple times, but we'll unique later
				feature_bundle f;
				f.push_back(i);
				f.push_back(n);
				f.push_back(existence);
				f.push_back(w);
				f.insert(f.end(),units.begin()+s,units.begin()+s+n);
				// new words (not used in model training) may not have features
				if (stream_feature_idx4.find(f) != stream_feature_idx4.end()) {
					feats.push_back(feature(stream_feature_idx4[f], 1.0));
					cf.push_back(feats.back());
				}
			}
		}
	}
	exist_cache[key] = cf;
}

void model::make_expectation_features(int starttime, int endtime, int w) {

	pair<int, int> times(starttime, endtime);
	pair<pair<int, int>, int> key(times, w);
	if (expect_cache.find(key) != expect_cache.end()) {
		vector<feature> &cf = expect_cache[key];
		feats.insert(feats.end(), cf.begin(), cf.end());
		return;
	}

	vector<feature> cf;
	ERR_check(ERR::ERR_SANITY, (size_t)num_streams == uttIt->cur_units4stream().size(), "[fe] number of streams inconsistent in current file");
	detection st(-1,starttime), et(-1, endtime);
	for (int i=0; i<num_streams; i++) {
		if (!prons4stream_word[i].count(w))
			continue;  // no pronunciation; no expectation features

		vector<detection>::iterator first = 
			lower_bound(uttIt->cur_units4stream()[i].begin(), uttIt->cur_units4stream()[i].end(), st);
		vector<detection>::iterator last = 
			upper_bound(uttIt->cur_units4stream()[i].begin(), uttIt->cur_units4stream()[i].end(), et);

		vector<detection>::iterator vi;
		vector<int> units;
		for (vi=first; vi<last; vi++) 
			units.push_back(vi->unit);

		// now the expectation features
		// a correct accept if any pron has a sequence
		// a false accept if no pron has a sequence
		set<int> ca;  // keep track of the correct accepts
		vector<vector<int > > &prons = prons4stream_word[i][w];
		int maxn = ngramlevel4expectation_stream[i];
		for (int n=1; n<=maxn; n++) {
			for (int s=0; s<=int(units.size())-n;s++) {
				vector<int> pattern(units.begin()+s,units.begin()+s+n);
				feature_bundle f;
				f.push_back(i);
				f.push_back(n);
				f.push_back(expectation);
				if (in_prons(pattern, prons)) {
					f.push_back(correct_accept);
					f.insert(f.end(),pattern.begin(), pattern.end());
					// must check because a new dictionary might have a new unit pattern
					// even if the set of units is unchanged
					if (stream_feature_idx4.find(f)!=stream_feature_idx4.end()) {
						int idx;
						feats.push_back(feature(idx=stream_feature_idx4[f], 1.0));
						cf.push_back(feats.back());
						ca.insert(idx);
					}
				} else {
					f.push_back(false_accept);
					f.insert(f.end(), pattern.begin(), pattern.end());
					if (stream_feature_idx4.find(f)!=stream_feature_idx4.end()) {
						feats.push_back(feature(stream_feature_idx4[f], 1.0));
						cf.push_back(feats.back());
					}
				}
			}
		}

		// do the false reject expectation features
		// it's a false reject if all prons have it and the hyp doesn't
		map<int, int> nrejects4;
		for (size_t p=0; p<prons.size(); p++) {
			set<int> counted;
			for (int n=1; n<=maxn; n++) {
				for (int s=0; s<=int(prons[p].size())-n;s++) {
					feature_bundle f;
					f.push_back(i);
					f.push_back(n);
					f.push_back(expectation);
					f.push_back(correct_accept);
					f.insert(f.end(),prons[p].begin()+s,prons[p].begin()+s+n);
					if (ca.count(stream_feature_idx4[f]))  // exists as a correct accept!
						continue;
					f[3] = false_reject;
					if (stream_feature_idx4.find(f)==stream_feature_idx4.end()) 
						continue;
					int idx = stream_feature_idx4[f];
					if (counted.count(idx))
						continue;  // only count once for each pron
					nrejects4[idx]++;
				}
			}
		}
		for (map<int,int>::iterator mi=nrejects4.begin(); mi!=nrejects4.end(); mi++)
			if ((size_t)mi->second == prons.size()) { // all prons had it
				feats.push_back(feature(mi->first, 1.0));
				cf.push_back(feats.back());
			}
	}
	expect_cache[key] = cf;
}


vector<feature> &model::features(lmrec from, lmrec to, int starttime, int endtime,
			    int fromlatp, int tolatp,
				float lmcost, int w, const vector<lmidx> *path, const vector<float> *external_features) {

	ERR_check(ERR::ERR_USERINPUT, w!=LM->unk_idx, "[fe] <unk> is not allowed in the lattice.");
    ERR_check(ERR::ERR_SANITY, starttime <= endtime, "[fe] arc going backwards while querying feature");

	statetime A(from.state , starttime, to.w, fromlatp), B(to.state, endtime, to.w, tolatp);
	pair<statetime, statetime> key(A, B);
	if (feature_cache.find(key) != feature_cache.end())
		return feature_cache[key];

	feats.clear();

	// make the features for each stream
	ERR_check(ERR::ERR_SANITY, (size_t)num_streams == uttIt->cur_units4stream().size(), "[fe] inconsistent number of streams in the current file");
	detection st(-1,starttime), et(-1, endtime);

	if ( !is_zero_array(fexConfig.existOrder) )
		make_existence_features(starttime, endtime, w);
	if ( !is_zero_array(fexConfig.expectOrder) )
		make_expectation_features(starttime, endtime, w);
	if ( fexConfig.use_levenshtein_features.size() )
		make_levenshtein_features(starttime, endtime, w);

	// add the LM feature
	if (fexConfig.use_global_lm) {
		feats.push_back(feature(global_lm_feature_idx, lmcost));
	}

	if( fexConfig.use_full_lm ) {
		for (size_t p=0; p<path->size()-1; p++) {
			pair<lmidx, lmidx> arc = pair<lmidx, lmidx>((*path)[p], (*path)[p+1]);
			ERR_check(ERR::ERR_SANITY, lm_feature_idx4.find(arc) != lm_feature_idx4.end(), "[fe] LM arc does not have a corresponding feature");
			feats.push_back(feature(lm_feature_idx4[arc], 1.0));
		}
		feats.push_back(feature(bo_feature_idx,log((float)path->size())));
	}

	if (fexConfig.basefiles.length() > 0) {
		vector<detection>::iterator first = lower_bound(uttIt->cur_baseline().begin(), uttIt->cur_baseline().end(), st);
		vector<detection>::iterator last = upper_bound(uttIt->cur_baseline().begin(), uttIt->cur_baseline().end(), et);

        if (first==last && w == LM->sil_idx) // floating silence
			feats.push_back(feature(baseline_feature_idx, 0.0));
		else if (first!=uttIt->cur_baseline().end() && last == first + 1 && first->unit == w)
			feats.push_back(feature(baseline_feature_idx, 1.0)); // baseline exists w/in boundary
		else 
			feats.push_back(feature(baseline_feature_idx, -1.0));
	}

	if( external_features ) {
		ERR_check(ERR::ERR_SANITY, idx4external.size() == external_features->size(), "[fe] unexpected number of external features");
		for (size_t i=0; i<idx4external.size(); i++) {
			feats.push_back(feature(idx4external[i], (*external_features)[i]));  // first note the global summary feature
			
			if (fexConfig.lexternal[i]) {  // and add the specific lexicalized feature
				const int fidx = idx4external[i] + ((-w > external_lexwords) ? (-LM->unk_idx) : (-w));
				ERR_check(ERR::ERR_SANITY, (size_t)fidx < lambda.size() && (size_t)fidx >= 0, "[fe] external feature out of bounds");
				feats.push_back(feature(fidx, (*external_features)[i]));
			}
		}
	}

	// existence and expectation features can be added multiple times
	// want just on instance
	// so uniq it
	sort(feats.begin(), feats.end());
	vector<feature>::iterator vi = unique(feats.begin(), feats.end());
	feats.erase(vi, feats.end());

	feature_cache[key] = feats;
	return feats;
}

void model::display(const feature_bundle &fb) {
	int snum = fb[0];
	int nglevel = fb[1];
	cout << "stream " << snum << " ngram level " << nglevel;
	detector_feature_type ft = (detector_feature_type) fb[2];
	if (ft == existence) {
		cout << " existence: ";
		string word = word4idx[fb[3]];
		cout << word << " + ";
		for (size_t i=4; i<fb.size(); i++)
			cout << sym4stream_idx[snum][fb[i]] << " ";
		cout << endl;
	} else if (ft == expectation) {
		cout << " expectation: ";
		if (fb[3] == correct_accept) {
			cout << "correct accept of ";
		} else if (fb[3] == false_reject) {
			cout << "false reject of ";
		} else if (fb[3] == false_accept) {
			cout << "false accept of ";
		} else {
			cout << "unknown expectation type\n";
			exit(1);
		}
		for (size_t i=4; i<fb.size(); i++)
			cout << sym4stream_idx[snum][fb[i]] << " ";
		cout << endl;
	} else if (ft == levenshtein) {
		cout << " levenshtein: ";
		if (fb[3] == ins) {
			cout << "insert ";
		} else if (fb[3] == sub) {
			cout << "substitution of ";
		} else if (fb[3] == del) {
			cout << "delete ";
		} else if (fb[3] == match) {
			cout << "match ";
		} else {
			cout << "unknown expectation type\n";
			exit(1);
		}
		for (size_t i=4; i<fb.size(); i++)
			cout << sym4stream_idx[snum][fb[i]] << " ";
		cout << endl;
	} else {
		cout << "unknown feature type encountered\n";
		exit(1);
	}
}

void model::display_detector_features() {
	map<feature_bundle, int>::iterator mi;
	for (mi = stream_feature_idx4.begin(); mi!= stream_feature_idx4.end(); mi++) {
		cout << "[" << mi->second << " @ " << lambda[mi->second] << "] ";
		display(mi->first);
	}
}

void model::number_lm_features() {
	const int sn = featurenum;

	if (fexConfig.use_global_lm)
		global_lm_feature_idx = featurenum++;

	if( fexConfig.use_full_lm ) {
		// go over all arcs in the LM and add the features
		list<lmhist>::const_iterator li;
		for (li=LM->ngrams.begin(); li!=LM->ngrams.end(); ++li) {
			const lmidx from = li->idx;
			if (li->idx != LM->initial_state) {  // otherwise backoff is undefined
				const lmidx to = li->backoff.succ->idx;
				lm_feature_idx4[pair<lmidx, lmidx>(from, to)] = featurenum++;
			}
			for (size_t i=0; i<li->succs.size(); i++) {
				const lmidx to = li->succs[i].succ->idx;
				lm_feature_idx4[pair<lmidx, lmidx>(from, to)] = featurenum++;
			}
		}
		bo_feature_idx = featurenum++;
	}
	cout << (featurenum-sn) << " language model features created" << endl;
}

void model::number_baseline_features() {
	baseline_feature_idx = featurenum++;
};

void model::display_global_lm_features() {
	if (fexConfig.use_global_lm) {
		cout << "[" << global_lm_feature_idx << " @ " << lambda[global_lm_feature_idx] 
		<< "] global_lm weight\n";
	}
}
void model::display_lm_features() {
	display_global_lm_features();
	if( fexConfig.use_full_lm ) {
		map<pair<lmidx, lmidx>, int>::iterator mi;
		for (mi=lm_feature_idx4.begin(); mi!=lm_feature_idx4.end(); mi++) {
			cout << "[" << mi->second << " @ " << lambda[mi->second] << "] LM " 
				 << mi->first.first << " -> " << mi->first.second << "\n";
		}
		display_lm_backoff_feature();
	}
}

void model::display_baseline_features() {
	cout << "[" << baseline_feature_idx << " @ " << lambda[baseline_feature_idx] 
		 << "] baseline weight\n";
}

void model::display_external_features(const bool nonlexonly) {
	ERR_check(ERR::ERR_SANITY, idx4external.size() == fexConfig.external_format.size(), "[fe] external index mapping does not match configuration");
	for (size_t i=0; i<idx4external.size(); i++) {
		// first print the global feature value
		cout << "[" << idx4external[i] << " @ " << lambda[idx4external[i]]
			     << "] global " << fexConfig.external_format[i] << " weight\n";
		if( nonlexonly && fexConfig.lexternal[i] ) continue;

		// now print the lexicalized features
		if( fexConfig.lexternal[i] ) {
			for(int wid=1; wid <= external_lexwords; ++wid) {
				cout << "[" << (idx4external[i]+wid) << " @"
				     << " " << lambda[idx4external[i]+wid]
				<< "] " << fexConfig.external_format[i] << "::" << (wid == -LM->unk_idx ? "<unk>": word4idx[-wid].c_str()) << " weight\n";
			}
		}
	}
}


void model::display() {
	if (!lambda.size()) {
		lambda.resize(featurenum);
		fill(lambda.begin(), lambda.end(), 0.0f);
	}
	display_detector_features();
	display_lm_features();
	if (fexConfig.basefiles.length() > 0)
		display_baseline_features();
	display_external_features(false);
}


int sign(double f) {
	if (f>0)
		return 1;
	else if (f<0)
		return -1;
	else
		return 0;
}

void model::gradient_step(bool set_scale) {
    if (set_scale) {  // call on 1st iteration
		for (size_t i=0; i<gradient.size(); i++) {
			step_size[i] = (float) min(step_size[i], (float) fabs(gradient[i]));
			step_size[i] = max(step_size[i], min_step);
		}
	}

	for (size_t i=0; i<gradient.size(); i++) { 
			gradient[i] += lambda[i] * training_examples * l2_regularizer;
			if( sign(lambda[i]) == 0 ) {
				const int l1_derivative = sign(gradient[i]);
				const float l1_delta = - training_examples * l1_derivative * l1_regularizer;
				gradient[i] += l1_delta;
				if( l1_derivative != sign(gradient[i]) ) gradient[i] = 0; // regularized objective is dominated by L1 around 0.
			} else {
				const int l1_derivative = sign(lambda[i]);
				gradient[i] += training_examples * l1_derivative * l1_regularizer;
			}
	}

	for (size_t i=0; i<lambda.size(); i++) {
		float delta;
		if (gradient[i]*last_gradient[i] > 0) {
			step_size[i] = min(step_size[i]*zeta_up, max_step);
			delta = - sign(gradient[i]) * step_size[i];
			last_gradient[i] = gradient[i];
		} else if (gradient[i]*last_gradient[i] < 0) {
			step_size[i] = max(step_size[i]*zeta_down, min_step);
			last_gradient[i] = 0;
			delta = 0;
		} else {
			// no change to step size
			delta = - sign(gradient[i]) * step_size[i];
			last_gradient[i] = gradient[i];
		}
		// now add delta
		if( l1_regularizer > 0.0f ) {
			const int oldsign = sign(lambda[i]);
			lambda[i] += delta;
			if( oldsign != 0 && sign(lambda[i]) != oldsign ) // no crossover
				lambda[i] = 0;
		} else
			lambda[i] += delta;
	}
}


pair<lmidx, float> model::succ(lmidx state, int word) {
	if (!LM->lm_words.count(word))
		word = LM->unk_idx;
	return pair<lmidx, float>(LM->lm_next_hist(state, word), LM->lm_logp(state, word));
}

void model::initialize_accumulators() {
	min_step = 0.000001f;
	max_step = 50.0f;
	initial_step = 0.1f;
	zeta_up = 1.2f;
	zeta_down = 0.5f;
	lambda.resize(featurenum);
	fill(lambda.begin(), lambda.end(), 0.0f);
	gradient.resize(featurenum);
	fill(gradient.begin(), gradient.end(), 0.0f);
	last_gradient.resize(featurenum);
	fill(last_gradient.begin(), last_gradient.end(), 0.0f);
	step_size.resize(featurenum);
	fill(step_size.begin(), step_size.end(), initial_step);
}

void model::reset_accumulators() {
	fill(gradient.begin(), gradient.end(), 0.0f);
}

#ifdef __USE_MPI
static bool uttTodo(const size_t uttId) {
	return uttId % MPI::numProcesses() == (size_t)MPI::GetRank();
}
#else
static bool uttTodo(const size_t) {
	return true;
}
#endif 

void model::train(const CFG::TrainConfig &traincfg) {
	initialize_accumulators();	
	for (size_t i=0; i < traincfg.num_iterations; i++) {
		double ldp = LOG_1;
		training_examples = 0;
		for (size_t f=0; uttTodo(f) ? uttIt->moveNext() : uttIt->skipNext(); f++) {
			if( uttTodo(f) ) {
				uttIt->getNumConstraints();
				compute_alphas(sum);
				const double zn = compute_betas();
				if(printScores)
					(void)printf("Numerator %s: %.4f\n", IO::basename(uttIt->cur_constraintFN()), zn);
				compute_gradient(-1.0);

				uttIt->getDenConstraints();
				compute_alphas(sum);
				const double zd = compute_betas();
				if(printScores)
					(void)printf("Denominator %s: %.4f\n", IO::basename(uttIt->cur_constraintFN()), zd);
				compute_gradient(1.0);

				ldp += zn - zd;
			} else {
				uttIt->skipNumConstraints();
				uttIt->skipDenConstraints();
			}
			training_examples++;
		}
#ifdef __USE_MPI
		MPI::reduce(ldp); MPI::bcast(ldp);
#endif
		const double dpe = ldp / training_examples;
		cout << "Log data prob per utterance after iteration " << i << ": " << dpe << " for " 
			 << training_examples << " examples" << endl;
		double ls = 0;
		size_t activeSetSize = 0;
		for (size_t k=0; k<lambda.size(); k++) {
			ls += fabs(lambda[k]);
			activeSetSize += sign(lambda[k]) != 0;
		}
		cout << "L1 lambda norm after iteration " << i << ": " << ls << endl;
		cout << "L0 lambda norm after iteration " << i << ": " << activeSetSize << endl;
		cout << "L1 was computed over " << lambda.size() << " features" << endl;

#ifdef __USE_MPI
		MPI::reduce(&gradient[0], gradient.size()); MPI::bcast(&gradient[0], gradient.size());
#endif
		gradient_step(i==0);
		if (fexConfig.use_global_lm)
			display_global_lm_features();
		if (fexConfig.use_full_lm)
			display_lm_backoff_feature();
		if (fexConfig.basefiles.length() > 0)
			display_baseline_features();
		if (idx4external.size())
			display_external_features(true);
		reset_accumulators();
		cout << "parameters updated\n";
#ifdef __USE_MPI
        if(MPI::isRoot())
#endif // fall-through
        {
			if(traincfg.saveItModel.length() > 0) {
				char fn[MAX_LINE_LENGTH];
				sprintf(fn, traincfg.saveItModel.c_str(), i);
				(void)printf("Saving model as %s\n", fn);
				save_model(fn);
			}
		}
		uttIt->rewind();
	}
}

void model::decode(const bool asConstraint) {
	(void)printf("=DECODING\n");
	for (size_t f=0; uttTodo(f) ? uttIt->moveNext() : uttIt->skipNext(); f++) {
			if( uttTodo(f) ) {
				uttIt->getDenConstraints();
				compute_alphas(viterbi);
				if( asConstraint ) {
					(void)printf("%s\n", uttIt->cur_constraintFN());
					vector< pair<int,int> > wordstime = backtrace();
					if( wordstime.size() ) {
						for (size_t i=wordstime.size(); i --> 1;)
							(void)printf("%d %d %s\n", wordstime[i].second+1, wordstime[i-1].second, word4idx[wordstime[i].first].c_str());
						(void)printf("%d %d </s>\n", wordstime[0].second+1, uttIt->cur_times().back());
						ERR_check(ERR::ERR_SANITY, wordstime[0].second < uttIt->cur_times().back(), "[dec] last word extends past end of utterance");
					}
					(void)printf(".\n");
				} else {
					vector< pair<int,int> > wordstime = backtrace();
					cout << uttIt->cur_constraintFN() << ": " << dataprob << endl;
					cout << "=>";
					if (wordstime.size()) {
						reverse(wordstime.begin(), wordstime.end());
						for (size_t i=0; i < wordstime.size(); i++) 
							cout << ' ' << word4idx[wordstime[i].first];
						cout << endl;
					} else 
						cout << " NULL" << endl;
				}
			} else
				uttIt->skipDenConstraints();
	}
}
		

void model::save_model(const char *fname) {
	IO::auto_file_t fout(fname, "wb");

	write_vec_float(fout, lambda);
	write_vec_double(fout, last_gradient);
	write_vec_float(fout, step_size);
	write_float(fout, zeta_up);
	write_float(fout, zeta_down);
	write_float(fout, max_step);
	write_float(fout, min_step);
	write_float(fout, initial_step);
	write_float(fout, l1_regularizer);
	write_float(fout, l2_regularizer);
	write_vec_int(fout, ngramlevel4existence_stream);
	write_vec_int(fout, ngramlevel4expectation_stream);
	write_string2int_map(fout, nglev4exist_stream);
	write_string2int_map(fout, nglev4expect_stream);
	ERR_check(ERR::ERR_SANITY, idx4stream_sym.size() == sym4stream_idx.size(), "[model] forward and reverse mapping from stream unit names do not match");

	int ns = (int)idx4stream_sym.size();
	fwrite((void *) &ns, sizeof(int), 1, fout);
	for (int i=0; i<ns; i++) 
		write_string2int_map(fout, idx4stream_sym[i]);
	fwrite((void *) &ns, sizeof(int), 1, fout);
	for (int i=0; i<ns; i++) 
		write_int2string_map(fout, sym4stream_idx[i]);

	ns = (int)stream_feature_idx4.size();
	fwrite((void *) &ns, sizeof(int), 1, fout);
	map<feature_bundle, int>::iterator mif;
	for(mif = stream_feature_idx4.begin(); mif!=stream_feature_idx4.end(); mif++) {
                vector<int> fb;
                for (size_t q=0; q<mif->first.size(); q++)
                    fb.push_back((int) mif->first[q]);
		write_vec_int(fout, fb);
		fwrite((void *) &mif->second, sizeof(int), 1, fout);
	}

	ns = (int)lm_feature_idx4.size();
	fwrite((void *) &ns, sizeof(int), 1, fout);
	map<pair<lmidx, lmidx>, int>::iterator mip;
	for (mip=lm_feature_idx4.begin(); mip!=lm_feature_idx4.end(); mip++) {
		fwrite((void *) &mip->first.first, sizeof(lmidx), 1, fout);
		fwrite((void *) &mip->first.second, sizeof(lmidx), 1, fout);
		fwrite((void *) &mip->second, sizeof(int), 1, fout);
	}

	// write the prons ...
	ns = (int)prons4stream_word.size();
	fwrite((void *) &ns, sizeof(int), 1, fout);
	for (int i=0; i<ns; i++) {
		int ns2 = (int)prons4stream_word[i].size();
		fwrite((void *) &ns2, sizeof(int), 1, fout);
		map<int, vector<vector<int> > >::iterator mi;
		for (mi = prons4stream_word[i].begin(); mi!= prons4stream_word[i].end(); mi++) {
			fwrite((void *) &mi->first, sizeof(int), 1, fout);
			ns2 = (int) mi->second.size();
			fwrite((void *) &ns2, sizeof(int), 1, fout);
			for (int j=0; j<ns2; j++)
				write_vec_int(fout, mi->second[j]);
		}
	}

	write_string2int_map(fout, idx4streamname);
	write_string2int_map(fout, idx4word);
	write_int2string_map(fout, word4idx);

	fwrite((void *) &num_streams, sizeof(int), 1, fout);
	fwrite((void *) &featurenum, sizeof(int), 1, fout);
	fwrite((void *) &fexConfig.use_global_lm, sizeof(bool), 1, fout);
	fwrite((void *) &fexConfig.use_full_lm, sizeof(bool), 1, fout);
	fwrite((void *) &global_lm_feature_idx, sizeof(int), 1, fout);
	fwrite((void *) &bo_feature_idx, sizeof(int), 1, fout);

	const bool use_baseline = fexConfig.basefiles.length() != 0;
	fwrite((void *) &use_baseline, sizeof(bool), 1, fout); 
	fwrite((void *) &baseline_feature_idx, sizeof(int), 1, fout);
	IO::xfwrite_scalar<unsigned>(fexConfig.use_levenshtein_features.size(), fout);
	for(size_t k=0; k < fexConfig.use_levenshtein_features.size(); ++k)
		IO::xfwrite_scalar<bool>(fexConfig.use_levenshtein_features[k], fout);

	IO::xfwrite_scalar<unsigned>(fexConfig.lexternal.size(), fout);
	for(size_t k=0; k < fexConfig.lexternal.size(); ++k)
		IO::xfwrite_scalar<bool>(fexConfig.lexternal[k], fout);
	IO::xfwrite_scalar( external_lexwords, fout );
	write_vec_int(fout, idx4external);
	write_vec_string(fout, fexConfig.external_format);
}


void model::read_model(const char *fname) {
	IO::auto_file_t fin(fname, "rb");

	read_vec_float(fin, lambda);  
	read_vec_double(fin, last_gradient);
	read_vec_float(fin, step_size);
	read_float(fin, zeta_up);
	read_float(fin, zeta_down);
	read_float(fin, max_step);
	read_float(fin, min_step);
	read_float(fin, initial_step);
	read_float(fin, l1_regularizer);
	read_float(fin, l2_regularizer);
	read_vec_int(fin, ngramlevel4existence_stream);
	read_vec_int(fin, ngramlevel4expectation_stream);
	read_string2int_map(fin, nglev4exist_stream);
	read_string2int_map(fin, nglev4expect_stream);

	int ns;
	fread((void *) &ns, sizeof(int), 1, fin);
	idx4stream_sym.resize(ns);
	for (int i=0; i<ns; i++) 
		read_string2int_map(fin, idx4stream_sym[i]);

	fread((void *) &ns, sizeof(int), 1, fin);
	sym4stream_idx.resize(ns);
	for (int i=0; i<ns; i++) 
		read_int2string_map(fin, sym4stream_idx[i]);

	fread((void *) &ns, sizeof(int), 1, fin);
	for (int i=0; i<ns; i++) {
		vector<int> fb;
		read_vec_int(fin, fb);
		int n;
		fread((void *) &n, sizeof(int), 1, fin);
		stream_feature_idx4[(feature_bundle) fb] = n;
	}

	fread((void *) &ns, sizeof(int), 1, fin);
	for (int i=0; i<ns; i++) {
		lmidx a, b;
		int idx;
		fread((void *) &a, sizeof(lmidx), 1, fin);
		fread((void *) &b, sizeof(lmidx), 1, fin);
		fread((void *) &idx, sizeof(int), 1, fin);
		lm_feature_idx4[pair<lmidx, lmidx>(a,b)] = idx;
	}

	// read the prons ...
	fread((void *) &ns, sizeof(int), 1, fin);
	prons4stream_word.resize(ns);
	for (int i=0; i<ns; i++) {
		int nwords;
		fread((void *) &nwords, sizeof(int), 1, fin);
		for (int j=0; j<nwords; j++) {
			int w;
			fread((void *) &w, sizeof(int), 1, fin);
			int nprons;
			fread((void *) &nprons, sizeof(int), 1, fin);
			vector<vector<int> > prons;
			prons.resize(nprons);
			for (int k=0; k<nprons; k++)
				read_vec_int(fin, prons[k]);
			prons4stream_word[i].operator [](w) = prons;
		}
	}

	read_string2int_map(fin, idx4streamname);
	read_string2int_map(fin, idx4word);
	read_int2string_map(fin, word4idx);

	fread((void *) &num_streams, sizeof(int), 1, fin);
	fread((void *) &featurenum, sizeof(int), 1, fin);
	fread((void *) &fexConfig.use_global_lm, sizeof(bool), 1, fin);
	fread((void *) &fexConfig.use_full_lm, sizeof(bool), 1, fin);
	fread((void *) &global_lm_feature_idx, sizeof(int), 1, fin);
	fread((void *) &bo_feature_idx, sizeof(int), 1, fin);

	bool use_baseline; 
	fread((void *) &use_baseline, sizeof(bool), 1, fin); 
	ERR_check(ERR::ERR_SANITY, (use_baseline == !!fexConfig.basefiles.length()), "[model] Model specifies baseline but no basefiles!");
	fread((void *) &baseline_feature_idx, sizeof(int), 1, fin);
	unsigned len;
	IO::xfread_scalar(len, fin);
	fexConfig.use_levenshtein_features.resize(len);
	for(size_t k = 0; k < len; ++k) {
		bool tmp;
		IO::xfread_scalar(tmp, fin);
		fexConfig.use_levenshtein_features[k] = tmp;
	}

	IO::xfread_scalar(len, fin);
	fexConfig.lexternal.resize(len);
	for(size_t k = 0; k < len; ++k) {
		bool tmp;
		IO::xfread_scalar(tmp, fin);
		fexConfig.lexternal[k] = tmp;
	}
	IO::xfread_scalar( external_lexwords, fin );
	read_vec_int(fin, idx4external);
	read_vec_string(fin, fexConfig.external_format);

	fin.close();

	// propagate back to stream
	fexConfig.stream_names.resize(num_streams);
	fexConfig.existOrder.resize(num_streams, 0);
	fexConfig.expectOrder.resize(num_streams, 0);
	map<string, int>::const_iterator strNit = idx4streamname.begin();
	for(; strNit != idx4streamname.end(); ++strNit) {
		int streamIdx = strNit->second;
		ERR_check(ERR::ERR_SANITY,  streamIdx>=0 && streamIdx < num_streams, "[model] unexpected stream index");
		fexConfig.stream_names[streamIdx] = strNit->first;
		fexConfig.existOrder[streamIdx] = nglev4exist_stream[strNit->first];
		fexConfig.expectOrder[streamIdx] = nglev4expect_stream[strNit->first];
	}
	ERR_check(ERR::ERR_SANITY, idx4streamname.size() == (size_t)num_streams, "[model] inconsistent number of streams in the model");
}

