// Geoffrey Zweig (gzweig@microsoft.com)
// Patrick Nguyen (panguyen@microsoft.com)
// Copyright (c) 2009 Microsoft Corporation. All rights reserved.

#ifndef MODEL_H
#define MODEL_H

#include "ngutil.h"
#include <math.h>
#include <float.h>
#include "lm.h"
#include "scarftypes.h"
#include "scarfread.h"

struct model {

	CFG::FEXConfig fexConfig;

	bool printScores; // during training

	UttIterator *uttIt;

	vector<float> lambda, step_size;
	vector<double> gradient, last_gradient;

	float zeta_up, zeta_down, max_step, min_step, initial_step, l1_regularizer, l2_regularizer;
	vector<map<lmrec,double> > alphas, betas;
	int max_alphas;  // the pruning parameter
	map<statetime, vector<statetime> > pred;
	vector<int> ngramlevel4existence_stream;   // expand existence features to this ngram level
	vector<int> ngramlevel4expectation_stream; // expand expectation features to this ngram level
	map<string, int> nglev4exist_stream;       // same but for string input
	map<string, int> nglev4expect_stream;
	vector<map<string, int> > idx4stream_sym;  // vector indexes the stream
	vector<map<int, string> > sym4stream_idx;  // vector indexes the stream
	map<feature_bundle, int> stream_feature_idx4;     // for detector features
	map<pair<lmidx, lmidx>, int> lm_feature_idx4;      // for language model features
	vector<map<int, vector<vector<int> > > > prons4stream_word;
	map<string, int> idx4streamname, idx4word;
	map<int, string> word4idx;
	map<pair<statetime, statetime>, vector<feature> > feature_cache;
	map<pair<pair<int, int>, int>, vector<feature> > exist_cache, expect_cache, leven_cache;
		// these caches map from the compination of start-time, end-time, word index to feature
	int num_streams, training_examples;
	int external_lexwords;
	language_model *LM;
	double dataprob;
	int featurenum;  // global counter
	int global_lm_feature_idx, baseline_feature_idx, bo_feature_idx;
	vector<string> dictionaries;

	vector<feature> feats;

	// data for using externally created features in annotated lattices
	vector<int> idx4external;  // the feature number for external feature 1, 2, ...

	void compute_alphas(mode _mode);
	double compute_betas();
	void compute_gradient(float sign);  // sign +-1 depending on numerator/denominator
	vector< pair<int,int> > backtrace();
	void decode(const bool asConstraint);  // all the denominator constraint files
	void gradient_step(bool set_scale = false);
	double fsum(const vector<feature> &f) const;
	void number_levenshtein(const std::vector<std::string > &detLists);
	void number_expectation(const std::vector<std::string > &detLists);
	void number_existence();
	void number_units_and_features(const vector<string> &dicts, const std::vector<std::string > &detLists);

	void number_lm_features();
	void number_baseline_features();
	void initialize_accumulators();
	void reset_accumulators();
	void train(const CFG::TrainConfig &traincfg);
	void read_dicts(vector<string> dicts, int &unitnum, bool model_trained=false);

	inline const vector<wordspan> start(time_lat_rec startpoint) const {
		constraint_t::const_iterator next = uttIt->cur_words_starting().find(startpoint);
		if( next != uttIt->cur_words_starting().end() ) return next->second;
		else return vector<wordspan>();
	}

	pair<lmidx , float> succ(lmidx state, int word); // returns next state and logprob

	bool in_prons(vector<int> &pat, vector<vector<int> > &prons);

	// from and w imply to and lmcost, but these are sometimes know at the time of the function call,
	// so pass them in as parameters to save any recomputation
	// this function computes and caches the values in a map indexed on the four parameters
	// path points to the sequence of states visited in the LM
	vector<feature> &features(lmrec from, lmrec to, int starttime, int endtime,
						 int fromlatp, int tolatp,
						 float lmcost=FLT_MIN, int w=INT_MIN, const vector<lmidx> *path=NULL,
						 const vector<float> *external_features=NULL);
	void make_levenshtein_features(int starttime, int endtime, int w);
	void make_existence_features(int starttime, int endtime, int w);
	void make_expectation_features(int starttime, int endtime, int w);

	void display(const feature_bundle &fb);
	void display_detector_features();
	void display_lm_features();
	void display_global_lm_features();
	void display_baseline_features();
	void display_external_features(const bool nonlexonly);
	inline void display_lm_backoff_feature() const
	{cout << "[" << bo_feature_idx << " @ " << lambda[bo_feature_idx] << "] LM backoff feature\n";}
	void display();
	void renumber_times(); // 1..n where n distinct times exist 

	void save_model(const char *fname);
	void read_model(const char *fname);

	model(const CFG::FEXConfig &fexcfg) : fexConfig(fexcfg), external_lexwords(-1) {
		uttIt = new UttIterator(fexConfig, fexConfig.external_format, idx4word, word4idx, idx4stream_sym);
	}
	inline ~model() { }
};

void logadd(double &target, double inc);

#endif
