/*
 * Patrick Nguyen (panguyen@microsoft.com)
 * Copyright (c) 2009 Microsoft Corporation.  All rights reserved.
 */

#include "ngutil.h"

#include "mpiwrap.h"

#include <string>

#ifdef __USE_MPI
#define OMPI_SKIP_MPICXX 1
#include <mpi.h>

int MPI::GetRank() {
	return rankId_;
}

size_t MPI::numProcesses() {
	return (size_t)numProcess_;
}

const std::string &MPI::hostname() { return hostname_; }

// return true if I am a master
static bool elect_nodemaster() {
	bool ret = false;
	// first, collection of all nodes
	if( MPI::isRoot() ) {
		std::map< std::string, int > h2m; // host to master
		int c;
		// read hostnames from everyone, proceed backwards so that root is not included
		for(c=(int)MPI::numProcesses(); c --> 0;) {
			if( c == MPI::GetRank() ) continue; // root cannot be node master
			MPI_Status status;
			int strl;
			const int ret1 = MPI_Recv((void *)&strl, 1, MPI_INT, c, 0x11a, MPI_COMM_WORLD, &status);
			ERR_check(ERR::ERR_ALWAYS, ret1 == MPI_SUCCESS, "[mpi] could not get strlen from child");
			std::vector<char> buf((size_t)strl+1);
			ERR_check(ERR::ERR_SANITY, strl != 0, "[mpi] hostname has zero len");
			const int ret2 = MPI_Recv((void *)&buf[0], strl, MPI_CHAR, c, 0x11b, MPI_COMM_WORLD, &status);
			ERR_check(ERR::ERR_ALWAYS, ret2 == MPI_SUCCESS, "[mpi] could not get host name");
			buf[strl] = 0;
			std::string childHost(&buf[0]);
			if( h2m.find(childHost) == h2m.end() )
				h2m[childHost] = c;
			MPI::send((size_t)h2m[childHost], c);
		}
		if( MPI::numProcesses() == 1 ) ret = true; // root is node master if I have only one process
	} else {
		// send my host name
		int strl = (int)MPI::hostname().length();
		const int ret1 = MPI_Send((void *)&strl, 1, MPI_INT, MPI::ROOT_RANK, 0x11a, MPI_COMM_WORLD);
		ERR_check(ERR::ERR_ALWAYS, ret1 == MPI_SUCCESS, "[mpi] could not send strlen to root");
		const int ret2 = MPI_Send((void *)MPI::hostname().c_str(), strl, MPI_CHAR, MPI::ROOT_RANK, 0x11b, MPI_COMM_WORLD);
		ERR_check(ERR::ERR_ALWAYS, ret2 == MPI_SUCCESS, "[mpi] could not send host name to root");
		size_t myMaster;
		MPI::recv(myMaster);
		ret = (size_t)MPI::GetRank() == myMaster;
	}
	return ret;
}

void MPI::Initialize(int *argc, char ***argv) {
	//MPI_Init(argc, argv);
	int provided;
	MPI_Init_thread(argc, argv, MPI_THREAD_FUNNELED, &provided);

	MPI_Comm_size(MPI_COMM_WORLD, &numProcess_);
	MPI_Comm_rank(MPI_COMM_WORLD, &rankId_);
	char hostname[MAX_LINE_LENGTH];
	(void)gethostname(hostname, sizeof(hostname)-1);
	hostname_.assign(hostname);

	isNodeMaster_ = elect_nodemaster();

	//MPI::SyncPoint();
	//(void)printf("[mpi] init: rank = %d, host = %s, nodeMaster = %s, root = %s\n", 
	//	MPI::GetRank(), MPI::hostname().c_str(), MPI::isNodeMaster() ? "yes" : "no", MPI::isRoot() ? "yes" : "no"
	//	);
	//fflush(stdout);fflush(stderr);
}

void MPI::Uninitialize() {
	MPI_Finalize();
}

bool MPI::isRoot() {
	return rankId_ == ROOT_RANK;
}

bool MPI::isNodeMaster() {
	return isNodeMaster_;
}

static const MPI_Datatype MPI_SIZE_T = 
#ifdef __USE_64BIT
	MPI_LONG_LONG
#else
	MPI_UNSIGNED
#endif // __USE_64BIT
	;

template<typename scalar_t> MPI_Datatype mpiType();

template<> MPI_Datatype mpiType<char>() { return MPI_CHAR; }
template<> MPI_Datatype mpiType<float>() { return MPI_FLOAT; }
template<> MPI_Datatype mpiType<int>() { return MPI_INT; }
template<> MPI_Datatype mpiType<uint32>() { return MPI_UNSIGNED; }
template<> MPI_Datatype mpiType<double>() { return MPI_DOUBLE; }
#ifdef __USE_64BIT
template<> MPI_Datatype mpiType<size_t>() { return MPI_SIZE_T; }
#endif

template<typename scalar_t>
void reduce_data(scalar_t *data, const size_t n) {
	if( MPI::isRoot() ) {
		// set up the receiving buffer
		std::vector<scalar_t> recv(n);
		const int errcode = MPI_Reduce( (void*)data, (void*)&recv[0], n, mpiType<scalar_t>(), MPI_SUM, MPI::ROOT_RANK, MPI_COMM_WORLD);
		ERR_check(ERR::ERR_ALWAYS, errcode == MPI_SUCCESS, "[mpi] failed reduce from root");
		(void)memcpy(data, &recv[0], n * sizeof(scalar_t));
	} else {
		// child node
		const int errcode = MPI_Reduce( (void*)data, NULL, n, mpiType<scalar_t>(), MPI_SUM, MPI::ROOT_RANK, MPI_COMM_WORLD);
		ERR_check(ERR::ERR_ALWAYS, errcode == MPI_SUCCESS, "[mpi] failed reduce from child");
	}
}

template<typename scalar_t>
void reduce_scalar(scalar_t &s) {
	scalar_t outs = s;
	MPI_Reduce( (void *)&s, &outs, 1, mpiType<scalar_t>(), MPI_SUM, MPI::ROOT_RANK, MPI_COMM_WORLD );
	s = outs;
}

template<typename scalar_t>
void bcast_data(scalar_t *data, const size_t n) {
	const int errcode = MPI_Bcast( (void*)data, n, mpiType<scalar_t>(), MPI::ROOT_RANK, MPI_COMM_WORLD );
	ERR_check(ERR::ERR_ALWAYS, errcode == MPI_SUCCESS, "[mpi] failed bcast");
}
template<typename scalar_t>
void bcast_scalar(scalar_t &s) { bcast_data<scalar_t>( &s, 1 ); }

void MPI::reduce(char &data) { reduce_scalar<char>(data); }
void MPI::reduce(float &data) { reduce_scalar<float>(data); }
void MPI::reduce(uint32 &data) { reduce_scalar<uint32>(data); }
void MPI::reduce(double &data) { reduce_scalar<double>(data); }
void MPI::reduce(char *data, const size_t n) { reduce_data<char>(data, n); }
void MPI::reduce(float *data, const size_t n) { reduce_data<float>(data, n); }
void MPI::reduce(uint32 *data, const size_t n) { reduce_data<uint32>(data, n); }
void MPI::reduce(double *data, const size_t n) { reduce_data<double>(data, n); }

void MPI::bcast(char &s) { bcast_scalar<char>(s); }
void MPI::bcast(float &s) { bcast_scalar<float>(s); }
void MPI::bcast(uint32 &s) { bcast_scalar<uint32>(s); }
void MPI::bcast(double &s) { bcast_scalar<double>(s); }
void MPI::bcast(char *data, const size_t n) { bcast_data<char>(data, n); }
void MPI::bcast(float *data, const size_t n) { bcast_data<float>(data, n); }
void MPI::bcast(uint32 *data, const size_t n) { bcast_data<uint32>(data, n); }
void MPI::bcast(double *data, const size_t n) { bcast_data<double>(data, n); }

#ifdef __USE_64BIT
void MPI::reduce(size_t &data) { reduce_scalar<size_t>(data); }
void MPI::reduce(size_t *data, const size_t n) { reduce_data<size_t>(data, n); }
void MPI::bcast(size_t &s) { bcast_scalar<size_t>(s); }
void MPI::bcast(size_t *data, const size_t n) { bcast_data<size_t>(data, n); }
#endif

// send a string from current process (must be root) to process rankTo
void MPI::send_s(const char *str, int rankTo) {
	ERR_check(ERR::ERR_ALWAYS, isRoot(), "[mpi] must send from root");
	const int32 strl = (int32)strlen(str);
	const int ret1 = MPI_Send( (void *)&strl, 1, MPI_INT, rankTo, 0x01, MPI_COMM_WORLD );
	ERR_check(ERR::ERR_ALWAYS, ret1 == MPI_SUCCESS, "[mpi] could not send string len");
	if( strl == 0 ) return;
	const int ret2 = MPI_Send( (void*)str, strl, MPI_CHAR, rankTo, 0x01, MPI_COMM_WORLD );
	ERR_check(ERR::ERR_ALWAYS, ret2 == MPI_SUCCESS, "[mpi] could not send string");
	return;
}

void MPI::recv_s(char *buf, const size_t maxlen) {
	ERR_check(ERR::ERR_ALWAYS, !isRoot(), "[mpi] attempting to receive from root");
	MPI_Status status;
	int32 strl;
	const int ret1 = MPI_Recv((void *)&strl, 1, MPI_INT, ROOT_RANK, 0x01, MPI_COMM_WORLD, &status);
	ERR_check(ERR::ERR_ALWAYS, ret1 == MPI_SUCCESS, "[mpi] could not get string len from root");
	ERR_check(ERR::ERR_ALWAYS, strl < (int)maxlen, "[mpi] attempting to transmit very long string ");
	buf[strl] = 0; // yes, null-terminate before we fill the content
	if( strl == 0 ) return;
	const int ret2 = MPI_Recv((void *)buf, strl, MPI_CHAR, ROOT_RANK, 0x01, MPI_COMM_WORLD, &status);
	ERR_check(ERR::ERR_ALWAYS, ret2 == MPI_SUCCESS, "[mpi] could not get string from root");
}

template<typename scalar_t>
void send_scalar(const scalar_t msg, int rankTo) {
	const int ret1 = MPI_Send( (void *)&msg, 1, mpiType<scalar_t>(), rankTo, 0x023, MPI_COMM_WORLD );
	ERR_check(ERR::ERR_ALWAYS, ret1 == MPI_SUCCESS, "[mpi] could not send data");
}
template<typename scalar_t>
void recv_scalar(scalar_t &msg) {
	MPI_Status status;
	const int ret1 = MPI_Recv( (void *)&msg, 1, mpiType<scalar_t>(), MPI::ROOT_RANK, 0x023, MPI_COMM_WORLD, &status );
	ERR_check(ERR::ERR_ALWAYS, ret1 == MPI_SUCCESS, "[mpi] could not receive data");
}

template<typename scalar_t>
void send_vector(const std::vector<scalar_t> &data, int rankTo) {
	//(void)printf("[mpi] raw send %d bytes\n", data.size());
	MPI::send(data.size(), rankTo);
	if( data.size() ) {
		const int ret2 = MPI_Send( (void*)&data[0], data.size(), mpiType<scalar_t>(), rankTo, 0xc94, MPI_COMM_WORLD );
		ERR_check(ERR::ERR_ALWAYS, ret2 == MPI_SUCCESS, "[mpi] could not send data across");
	}
}
template<typename scalar_t>
void recv_vector(std::vector<scalar_t> &data) {
	data.clear();
	size_t bsiz;
	MPI::recv(bsiz);
	//(void)printf("[mpi] raw recv %d bytes\n", bsiz);
	if( !bsiz ) return;
	data.resize(bsiz);
	MPI_Status status;
	const int ret2 = MPI_Recv((void *)&data[0], bsiz, mpiType<scalar_t>(), MPI::ROOT_RANK, 0xc94, MPI_COMM_WORLD, &status);
	ERR_check(ERR::ERR_ALWAYS, ret2 == MPI_SUCCESS, "[mpi] could not get string from root");
}

void MPI::send(const size_t msg, int rankTo) { send_scalar<size_t>(msg, rankTo); }

void MPI::recv(size_t &msg) { recv_scalar<size_t>(msg); }
void MPI::send(const std::vector<char> &data, int rankTo) { send_vector(data, rankTo); }
void MPI::recv(std::vector<char> &data) { recv_vector<char>(data); }

void MPI::send(const std::vector<std::string> &list, int to) {
	std::vector<char> serbuf;
	size_t k;
	for(k=0; k < list.size(); ++k) {
		const size_t now = serbuf.size();
		serbuf.resize(now + list[k].length() + 1);
		memcpy(&serbuf[now], list[k].c_str(), list[k].length()+1);
	}
	MPI::send(serbuf, to);
}

void MPI::recv(std::vector<std::string> &list) {
	std::vector<char> serbuf;
	MPI::recv(serbuf);
	//(void)printf("[mpi] list recv got the serialized buf\n");
	list.clear();
	if( !serbuf.size() ) return;
	char *p = &serbuf[0];
	for(; (size_t)(p - &serbuf[0]) < serbuf.size();) {
		list.push_back(p);
		p += list.back().length()+1;
	}
	ERR_check(ERR::ERR_SANITY, p == &serbuf[0] + serbuf.size(), "[mpi] serialized buf parsing error for list of strings");
}

void MPI::RedirectOut(const char *tmpl)  {
	if(!tmpl || !tmpl[0]) return;
	char bufout[MAX_LINE_LENGTH];
	(void)sprintf_s(bufout, tmpl, "out", MPI::GetRank());
	char buferr[MAX_LINE_LENGTH];
	(void)sprintf_s(buferr, tmpl, "err", MPI::GetRank());
	if(isRoot()) {
		(void)printf("Redirecting (out,err) to %s\n", tmpl);
	}
	// annoying file sharing
	//FILE *f;
	//errno_t err;
	// Reassign "stderr" to "freopen.out": 
	//err = freopen_s( &f, bufout, "wt", stdout );
	//ERR_check(ERR::ERR_ALWAYS, err == 0 && "could not redirect stdout");
#ifndef _GCC_
# pragma warning(push)
# pragma warning(disable: 4996)
#endif
	(void)freopen(bufout, "wt", stdout);


	if( !strcmp(bufout, buferr) ) {
#ifdef _GCC_
		dup2(0, 1); // dup stdout into stderr
#else
		_dup2(0, 1); // dup stdout into stderr
#endif
	} else {
		// annoying sharing
		(void)freopen(buferr, "wt", stderr);
		//err = freopen_s( &f, buferr, "wt", stderr );
		//ERR_check(ERR::ERR_ALWAYS, err == 0 && "could not redirect stderr");
	}
#ifndef _GCC_
# pragma warning(pop)
#endif
	return;
}

void MPI::LocalizePath(char *dest, const size_t maxdestlen, const char *localDir, const char *origFileName) {
	ERR_check(ERR::ERR_USERINPUT, maxdestlen > strlen(origFileName) + strlen(localDir) + 1, "[mpi] buffer overflow while localizing path");
	(void)sprintf(dest, "%s/%s", localDir, IO::basename(origFileName));
}

void MPI::GetLocalDir(const CFG::MPIConfig *mpicfg, char *dest, const size_t maxdestlen) {
	ERR_check(ERR::ERR_USERINPUT, maxdestlen > mpicfg->localProcDir.length() + 3, "[mpi] buffer overflow while constructing local directory");
	(void)sprintf(dest, mpicfg->localProcDir.c_str(), MPI::GetRank());
}

void MPI::MakeLocalDir(const CFG::MPIConfig *mpicfg) {
	char local_proc_dir[MAX_LINE_LENGTH];
	MPI::GetLocalDir(mpicfg, local_proc_dir, MAX_LINE_LENGTH);

	IO::mkdirhier(local_proc_dir);
}

std::string MPI::GetLocalFileName(const CFG::MPIConfig *mpicfg, const char *fn) {
	char locdir[MAX_LINE_LENGTH];
	GetLocalDir(mpicfg, locdir, sizeof locdir);
	char localfn[MAX_LINE_LENGTH];
	LocalizePath(localfn, sizeof localfn, locdir, fn);
	return std::string(localfn);
}

// wait for all processes to reach here
void MPI::SyncPoint() {
	MPI_Barrier(MPI_COMM_WORLD);
}

// wait for an ack signal from the child
void MPI::childSync(const int childno) {
	if(childno == ROOT_RANK) return;
	const int ackID = ~ 0xdead; // alive is...
	int ack;
	if( isRoot() ) {
		MPI_Status status;
		const int ret1 = MPI_Recv((void *)&ack, 1, MPI_INT, childno, 0xbeef, MPI_COMM_WORLD, &status);
		ERR_check(ERR::ERR_ALWAYS, ret1 == MPI_SUCCESS, "[mpi] could not get ack from child");
		ERR_check(ERR::ERR_SANITY, ack == ackID, "[mpi] wrong cookie in MPI sync");
	} else {
		ack = ackID;
		ERR_check(ERR::ERR_SANITY, childno == GetRank(), "[mpi] sync with wrong ID");
		// child sending to root
		const int ret1 = MPI_Send( (void *)&ack, 1, MPI_INT, ROOT_RANK, 0xbeef, MPI_COMM_WORLD );
		ERR_check(ERR::ERR_ALWAYS, ret1 == MPI_SUCCESS, "[mpi] could not send ack to root");
	}
}

static inline std::string localfn(const char *local_proc_dir, const char *fn) {
	char buf[MAX_LINE_LENGTH];
	MPI::LocalizePath(buf, MAX_LINE_LENGTH, local_proc_dir, fn);
	return std::string(buf);
}

void MPI::BCastFile(const CFG::MPIConfig *mpicfg, const char *fn) {
	size_t fsiz;
	if( isRoot() ) {
		fsiz = IO::xfsize(fn);
	}
	MPI::bcast(fsiz);
	const size_t MAX_BUFSIZ = 256 * 1024 * 1024;
	size_t offset = 0;
	IO::auto_file_t *infile = isRoot() ? new IO::auto_file_t(fn, "rb", false) : 0;
	IO::auto_file_t outfile(GetLocalFileName(mpicfg, fn).c_str(), "wb", false);
	std::vector<char> buf(std_min(MAX_BUFSIZ, fsiz));
	do {
		const size_t toread = std_min(MAX_BUFSIZ, fsiz - offset);
		if( isRoot() ) 
			IO::xfread_many(&buf[0], toread, *infile);
		// efficient multicast
		MPI::bcast(&buf[0], toread);
		IO::xfwrite_many(&buf[0], toread, outfile);
		offset += toread;
	} while(offset < fsiz);
	ERR_check(ERR::ERR_SANITY, offset == fsiz, "[mpi] bcast finished with inconsistent length");
	if( infile ) delete infile;
}

void MPI::CopyFileLocally(const CFG::MPIConfig *mpicfg, const char *fn) {
	std::string buf( GetLocalFileName(mpicfg, fn) );
	(void)printf("[mpi cache] copying file %s locally, to %s\n", fn, buf.c_str());
	std::vector<char> content(IO::xfsize(fn));
	IO::auto_file_t in(fn, "rb", false);
	IO::xfread_many(&content[0], content.size(), in);
	in.close();
	IO::auto_file_t out(buf.c_str(), "wb", false);
	IO::xfwrite_many(&content[0], content.size(), out);
	out.close();
}

void MPI::CopyFromCache(const CFG::MPIConfig *mpicfg, const char *fn) {
	std::string buf( GetLocalFileName(mpicfg, fn) );
	(void)printf("[mpi uncache] copying file from local copy %s to %s\n", buf.c_str(), fn);
	std::vector<char> content(IO::xfsize(buf.c_str()));
	IO::auto_file_t in(buf.c_str(), "rb", false);
	IO::xfread_many(&content[0], content.size(), in);
	in.close();
	IO::auto_file_t out(fn, "wb", false);
	IO::xfwrite_many(&content[0], content.size(), out);
	out.close();
}

void MPI::ProcessMarkups(const CFG::MPIConfig *mpicfg, clparser &cl) {
	{
		char local_proc_dir[MAX_LINE_LENGTH];
		MPI::GetLocalDir(mpicfg, local_proc_dir, MAX_LINE_LENGTH);
		IO::mkdirhier(local_proc_dir);
	}

	bool done = false;
	while(!done) {
		done = true;
		map<string, arg>::const_iterator argit = cl.arginfo4.begin();
		for(; argit != cl.arginfo4.end(); ++argit) {
			// broadcast file
			if( !strncmp("mpi:bc:", argit->second.val.c_str(), 7) ) {
				MPI::BCastFile(mpicfg, argit->second.val.c_str() + 7);
				arg a = argit->second;
				if( !MPI::isRoot() ) 
					a.val = MPI::GetLocalFileName(mpicfg, argit->second.val.c_str() + 7);
				cl.arginfo4[argit->first] = a;
				done = false;
				break;
			}
			// file has been broadcasted previously, just work off the cached file
			if( !strncmp("mpi:ln:", argit->second.val.c_str(), 7) ) {
				arg a = argit->second;
				a.val = MPI::GetLocalFileName(mpicfg, argit->second.val.c_str() + 7);
				cl.arginfo4[argit->first] = a;
				done = false;
				break;
			}
		}
	}
}

#ifndef _GCC_
#include "perf.h"

class HeartBeat {
	FILE *f_;
	SysMon *mon_;

	static const size_t NUM_COUNTERS = 3 * 2; // each cpu, net, disk get 2

	// @return the size of the mon size
	// @see Sysmon::run
	static size_t compute_mon_size(size_t &headerLen, size_t &machNameOffset, size_t &dataOffset) {
		std::vector<uint32> machLen(MPI::numProcesses()+1, 0);
		if( MPI::isNodeMaster() ) {
			machLen[MPI::GetRank()] = MPI::hostname().length()+1;
			machLen.back() = 1;
		}
		MPI::reduce(&machLen[0], machLen.size());
		MPI::bcast(&machLen[0], machLen.size());

		size_t ret = 0;

		size_t numNodes = machLen.back();
		machLen.back() = 0;

		const size_t machNameLen = std::accumulate( machLen.begin(), machLen.end(), 0 ); // machine names 
		const size_t headerOffset = sizeof(uint32);
		headerLen = machNameLen + headerOffset;
		ret = headerLen; 
		//                          alive/done        clock           counters
		const size_t payload = ( sizeof(uint32) + sizeof(float) + sizeof(float) * NUM_COUNTERS); // that's where Sysmon::run matters
		ret += numNodes * payload;

		if( MPI::isNodeMaster() ) {
			machNameOffset = std::accumulate( machLen.begin(), machLen.begin() + MPI::GetRank(), 0 );
			machNameOffset += headerOffset;
			dataOffset = headerLen;
			size_t k;
			for(k=0; k < (size_t)MPI::GetRank(); ++k)
				dataOffset += payload * !!machLen[k];
		} else {
			machNameOffset = size_t(-1);
			dataOffset = size_t(-1);
		}
		(void)printf("[mpi:heartbeat] %s: mon = %d, machOff = %d, dataOff = %d, H=%d, n=%d\n", MPI::hostname().c_str(), ret, machNameOffset, dataOffset, headerLen, numNodes);
		fflush(stdout);fflush(stderr);
		return ret;
	}
public:
	HeartBeat(const char *outfn, const size_t durms) : mon_(0), f_(0) {
		size_t machNameOffset, dataOffset, fullHeaderLen;
		const size_t monSiz = compute_mon_size(fullHeaderLen, machNameOffset, dataOffset);
		char *dataPtr = 0;
		if( MPI::isRoot() ) {
			f_ = _fsopen(outfn, "wb", _SH_DENYNO);
			ERR_check(ERR::ERR_SANITY, f_, "[mon] failed to create heartbeat file from root");
			std::vector<char> buf(monSiz, 0);
			IO::xfwrite_many(&buf[0], buf.size(), f_);
			// initialize with the right data size
			fseek(f_, 0, SEEK_SET);
			const uint32 u32fullHeadLen = (uint32)fullHeaderLen;
			IO::xfwrite_scalar(u32fullHeadLen, f_);
			fflush(f_);
		}
		MPI::SyncPoint();
		if( !MPI::isRoot() && MPI::isNodeMaster() ) {
			f_ = _fsopen(outfn, "r+b", _SH_DENYNO);
			ERR_check(ERR::ERR_SANITY, f_, "[mon] failed to open heartbeat file from child");
		}

		if( MPI::isNodeMaster() ) {
			fseek(f_, machNameOffset, SEEK_SET);
			const size_t len = MPI::hostname().length() + 1;
			IO::xfwrite_many( MPI::hostname().c_str(), len, f_ );
			mon_ = new SysMon(f_, dataOffset, durms);
			fflush(f_);
		}
		MPI::SyncPoint();
		(void)printf("[mpi:heartbeat] machines written\n"); fflush(stdout);fflush(stderr);
		Sleep(1000);
		return;
	}
	~HeartBeat() {
		if(mon_) delete mon_;
		if(f_) fclose(f_);
	}
};

void MPI::InitHeartBeat(const CFG::MPIConfig *mpicfg) {
	if( !mpicfg->heartbeat.monfile.length() )
		return;
	heartBeat_ = new HeartBeat(mpicfg->heartbeat.monfile.c_str(), mpicfg->heartbeat.durms);
}

void MPI::UnInitHeartBeat() {
	if( heartBeat_ ) delete heartBeat_;
	heartBeat_ = 0;
}
#else // _GCC_ has no heartbeat (not implemented)
void MPI::InitHeartBeat(const CFG::MPIConfig *) {
   //(void)printf("[mpi:hb] I am %d / %d [%s]\n", MPI::GetRank(), MPI::numProcesses(), MPI::hostname().c_str());fflush(stdout);
}

void MPI::UnInitHeartBeat() {}
#endif

#else
int MPI::GetRank() {
	return 0;
}

size_t MPI::numProcesses() {
	return 1;
}

void MPI::SyncPoint() {}

void MPI::reduce(char &) {}
void MPI::reduce(float &) {}
void MPI::reduce(uint32 &) {}
void MPI::reduce(double &) {}
void MPI::reduce(char *, const size_t) {}
void MPI::reduce(float *, const size_t) {}
void MPI::reduce(uint32 *, const size_t) {}
void MPI::reduce(double *, const size_t) {}
void MPI::bcast(char &) {}
void MPI::bcast(float &) {}
void MPI::bcast(uint32 &) {}
void MPI::bcast(double &) {}
void MPI::bcast(char *, const size_t) {}
void MPI::bcast(float *, const size_t) {}
void MPI::bcast(uint32 *, const size_t) {}
void MPI::bcast(double *, const size_t) {}


#ifdef __USE_64BIT
void MPI::reduce(size_t &) {}
void MPI::reduce(size_t *, const size_t) {}
void MPI::bcast(size_t &) {}
void MPI::bcast(size_t *, const size_t) {}
#endif

void MPI::send_s(const char *, int) {}
void MPI::recv_s(char *, size_t) {}

void MPI::send(const std::vector<char> &, int) {}
void MPI::recv(std::vector<char> &) {}

void MPI::send(const size_t, int) {}
void MPI::recv(size_t &) {}

void MPI::send(const std::vector<std::string> &, int) {}
void MPI::recv(std::vector<std::string> &) {}

void MPI::childSync(int) {}

const std::string &MPI::hostname() { hostname_.assign("localhost"); return hostname_; }

bool MPI::isRoot() { return true; }
bool isNodeMaster() { return true; }

void MPI::RedirectOut(const char *) {}

void MPI::LocalizePath(char *dest, const size_t, const char *, const char *origFileName) {
	(void)strcpy(dest, origFileName);
}
void MPI::GetLocalDir(const CFG::MPIConfig *, char *dest, const size_t destlen) {
	if( destlen > 0 ) *dest = 0;
	ERR_check(ERR::ERR_NOTIMPL, 0, "not impl for non MPI");
}
void MPI::MakeLocalDir(const CFG::MPIConfig *) {}

void MPI::CopyFileLocally(const CFG::MPIConfig *, const char *) {}
void MPI::CopyFromCache(const CFG::MPIConfig *, const char *) {}

void MPI::BCastFile(const CFG::MPIConfig *, const char *) {}
std::string MPI::GetLocalFileName(const CFG::MPIConfig *, const char *fn) { return std::string(fn); }

#endif // __USE_MPI

int MPI::rankId_;
int MPI::numProcess_;
std::string MPI::hostname_;
bool MPI::isNodeMaster_ = false;
HeartBeat *MPI::heartBeat_ = 0;
