/*
 * Patrick Nguyen (panguyen@microsoft.com)
 * Copyright (c) 2005 Microsoft Corporation.  All rights reserved.
 *
 * Error subsystem
 * 
 * See ERR_check(.) in the rest of the code for examples.
 */

// this is the only file that does not include ngtypes.h - should not be included separately
#ifndef __NGTYPES_H
# error Do not include this file separately. This is a relay of ngtypes.h
#endif // __NGTYPES_H



#ifndef __NGERR_H
#define __NGERR_H 1

#include <exception>

// what's the minimum error level that we check for
#define __CURRENT_ERROR_LEVEL__ ERR::ERR_DBGONLY

/**
 * location of the code - 
 */
// #define CODEINFO     ERR::codeloc_t(__FILE__, __LINE__, __FUNCTION__, __FUNCSIG__)
// __FUNCTION__ and __FUNCSIG__ are non-standard
#define CODEINFO     ERR::codeloc_t(__FILE__, __LINE__, "", "")

/**
 * Check for a condition with a given "gravity"
 */
#define ERR_check(errlev, Expr, Explain) do { \
	      if(ERR::ERR_level_high(errlev) && !(Expr)) { \
		     ERR::ngerror_t(CODEINFO, errlev, #Expr, Explain).report();  \
		  } \
        } while(false)

/**
 * Same as ERR_check but do not throw
 */
#define WARN_check(errlev, Expr, Explain) do{ if(!(Expr)) { ERR::ngerror_t(CODEINFO, errlev, #Expr, Explain, false).print(); } }while(false)

namespace ERR {
	// this should be generated by cfgen?
	/**
	 * The error level
	 */
	typedef enum {
		ERR_DBGONLY    =  0x0, // expensive sanity checks
		ERR_SANITY     =  0x1, // sanity checks - internal consistency
		ERR_CONDITION  =  0x2, // ill-conditioning of the algorithms
		ERR_USERINPUT  =  0x3, // obvious user error
		ERR_NOTIMPL    =  0x4, // not implemented
		ERR_ALWAYS     =  0x5, // always checked in all configurations
	} errlevel_t;

	/**
	 * is the error level high enough?
	 */
	inline bool ERR_level_high(const errlevel_t errlev) {
		return errlev >= __CURRENT_ERROR_LEVEL__;
	}

	/**
	 * location in the code
	 */
	struct codeloc_t {
		const char *source_file;
		const int lineno;
		const char *function;
		const char *pretty_function; // pretty function name
		inline codeloc_t(const codeloc_t &c) : source_file(c.source_file), lineno(c.lineno), function(c.function), pretty_function(c.pretty_function) {}
		inline codeloc_t(const char *srcf, const int line, const char *func, const char *pfunc) : 
		   source_file(srcf), lineno(line), function(func), pretty_function(pfunc) {}
		 inline void print(FILE *f) const {
			 (void)fprintf(f, "%s (%s:%d)", pretty_function, source_file, lineno);
			 fflush(f); // use setvbuf instead
		 }
	};

	class ngerror_t : public std::exception {
	protected:
		codeloc_t  codeinfo_;
		errlevel_t errlev_;
		const bool isError_;
		const char *codecheck_;
		const char *explainText_;
	public:
		inline ngerror_t(const codeloc_t codeinfo, const errlevel_t err, const char *codecheck, const char *explainText, const bool isError = true) 
			: codeinfo_(codeinfo), errlev_(err), isError_(isError), codecheck_(codecheck), explainText_(explainText)  {}
		inline void print(FILE *f=stderr) const {
			(void)fprintf(f, "%s (%d): %s\n", isError_ ? "Error" : "Warning", errlev_, explainText_);
			if( isError_ ) {
				(void)fprintf(f, "==\nCondition violated: %s  ", codecheck_);
				fflush(f);
				codeinfo_.print(f);
				(void)fprintf(f, "\n");
			}
			fflush(f);
		}
		inline void report() const {
			print();
			exit(1);
//			throw *this;
		}
	};

} //ERR namespace
#endif /* __NGERR_H */

