/*
 * Patrick Nguyen (panguyen@microsoft.com)
 * Copyright (c) 2005, 2008 Microsoft Corporation.  All rights reserved.
 *
 * Next Generation generic utilities
 *  Includes parsing, IO, etc.
 */

#ifndef __NGUTIL_H
#define __NGUTIL_H 1

#include "ngtypes.h"
#include "ngerr.h"
#include "compiler_specific.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <wchar.h>
#include <stdio.h>

#ifdef _GCC_
# define strncpy_s strncpy
# define strcpy_s strcpy
# define sprintf_s sprintf
#else
# include <io.h>
#endif /* _GCC_ */

/**** CONFIG IO UTILITIES ****/
namespace CFG {
    int parseCommandLine(std::map<std::string, std::string> &args, const int argc, char **argv);
    void parseConfigFile(std::map<std::string, std::string> &args, char *fname);
	bool str2bool(const char *str);
}
	
/**
 * parsing utilities
 */
namespace PRS {
	/**
	 * remove end-of-line character
	 * Also strips trailing carriage returns (\r)
	 * @param stripspaces is true if you want to strip characters at the end as well
	 * @return the modified string itself
	 */
	inline char *xchomp(char *buf, const bool stripspaces = false) {
		char *p=buf+strlen(buf)-1; 
		while( p>=buf && ((*p == '\n' || *p == '\r') || (stripspaces && (*p==' ' || *p=='\t'))) ) 
		{*p=0;p--;}
		return buf;
	}
	inline wchar_t *xwchomp(wchar_t *buf, const bool stripspaces = false) {
		wchar_t *p=buf+wcslen(buf)-1; 
		while( p>=buf && ((*p == L'\n' || *p == L'\r') || (stripspaces && (*p==L' ' || *p==L'\t'))) ) 
		{*p=0;p--;}
		return buf;
	}
	/**
	 * Tokenize -- will destroy contents of the string
	 * @param sep are the characters used to separate fields
	 */
	inline std::vector<std::string> xtokenize(char *buf, const char *sep = "\n\r\t ") {
		std::vector<std::string> ret;
		char *p;
		char *start;
		for(start=p=buf; *p; p++) {
			int k;
			for(k=0; sep[k]; k++)
				if( p[0] == sep[k] ) {
					// found one
					*p = 0;
					if( start != p ) ret.push_back(start); // add if previous char was not a sep
					start = p+1;
					break;
				}
		}
		if( p != start )
			ret.push_back(start);
		return ret;
	}

	inline std::vector<std::string> xtokenize(const std::string str, const char *sep = "\n\r\t ") {
		return xtokenize((char*)str.c_str(), sep);
	}

	// same as xtokenize, but don't skip empty fields
	inline std::vector<std::string> xtokenize2(char *buf, const char *sep = "\n\r\t ") {
		std::vector<std::string> ret;
		char *p;
		char *start;
		for(start=p=buf; *p; p++) {
			int k;
			for(k=0; sep[k]; k++)
				if( p[0] == sep[k] ) {
					// found one
					*p = 0;
					ret.push_back(start); // add if previous char was not a sep
					start = p+1;
					break;
				}
		}
		if( p != start )
			ret.push_back(start);
		return ret;
	}

	/**
	 * like perl join
	 */
	inline std::string xjoin(const char *sep, const std::vector<std::string> &tok) {
		if( tok.size() == 0 )
			return std::string();
		std::string ssep(sep);
		std::string ret;
		std::vector<std::string>::const_iterator tokit, last;
		last = tok.end();
		--last;
		for(tokit = tok.begin(); tokit != last; ++tokit) {
			ret.append(*tokit);
			ret.append(ssep);
		}
		ret.append(tok.back());
		return ret;
	}

	/**
	 * test if it's a number, in the American English locale. Supports plain numbers only (eg not 0x...)
	 * pngT: replace me with appropriate functions. (isdigit?)
	 */
	inline bool isNumber(const char *buf, const bool floatOK = true, const bool negOK = true) {
		const char *p;
		for(p=buf; *p; p++) {
			if( p[0] >= '0' && p[0] <= '0' + 9 )
				continue;
			if( floatOK && p[0] == '.' )
				continue;
			if( negOK && p[0] == '-' )
				continue;
			return false;
		}
		return true;
	}
	/**
	 * @return true if the string is quoted
	 */
	inline bool isQuoted(const char *buf, const char quoteopen = '\'', const char quoteend = '\'') {
		return buf[strlen(buf)-1] == quoteend && buf[0] == quoteopen;
	}
	/**
	 * remove first and last characters
	 */
	inline std::string removeQuotes(const char *buf) {
		return std::string(buf+1, strlen(buf)-2);
	}
}

/**
 * placeholder for generic input output functions
 */
namespace IO {
	#define WIN32_MAX_CHUNK_SIZE (1024*1024)

	/**
	 * @return the size of the file or 0 if it doesn't exist
	 */
	inline size_t xfsize(const char *fname) {
#ifdef _GCC_
		struct stat finfo;
		const int ret = stat(fname, &finfo);
#else
		struct __stat64  finfo;
		const int ret = _stat64(fname, &finfo);
#endif
		if( ret < 0 )
			return 0;
		return (size_t)finfo.st_size;
	}

	static inline bool xdirexists(const char *dirname) {
#ifdef _GCC_
# if !defined(_IFDIR) && defined(S_IFDIR)
#  define _IFDIR S_IFDIR
# endif
		struct stat dstat;
		const int ret = stat(dirname, &dstat);
		if( ret != 0 ) return false;
		return !!(dstat.st_mode & _IFDIR);
#else
		struct _stat dstat;
		const int ret = _stat(dirname, &dstat);
		if( ret != 0 ) return false;
		return !!(dstat.st_mode & _S_IFDIR);
#endif
	}

#ifdef _GCC_
# define _mkdir(x) mkdir((x), 0755)
#else
	extern "C" int _mkdir(const char *);
#endif // _GCC_
	// work backwards, then go forward -- think //speechstore.
	// windows version of mkdirhier
	static inline void mkdirhier(const char *dirname) {
		char buf[MAX_LINE_LENGTH];
		(void)strcpy_s(buf, dirname);
		char *end = buf + strlen(buf);
		while(!IO::xdirexists(buf)) {
			char *p = strrchr(buf, '/');
			if( p ) {
				*p = 0;
				_mkdir(buf); // try making it here
			} else {
				_mkdir(buf);
				break; // don't check again
			}
		}
		// now I've made a directory, make it upwords
		char *q = buf + strlen(buf);;
		while( q < end ) {
			*q = '/';
			_mkdir(buf);
			q = buf + strlen(buf);
		}
		ERR_check(ERR::ERR_SANITY, xdirexists(dirname), "mkdirhier failed");
		return;
	}

	/**
	 * exception throwing IO
	 */
	inline size_t xfread(void *str, const size_t siz, const size_t count, FILE *f) {
		//windows 
		//const size_t ret = fread( str, siz, count, f );
		//ERR_check(ERR::ERR_ALWAYS, ret == count && "I/O error reading from file");
		size_t remBytes = siz * count;
		char *pcstr = (char *)str;
		while( remBytes > WIN32_MAX_CHUNK_SIZE ) {
			const size_t ret = fread( pcstr, 1, WIN32_MAX_CHUNK_SIZE, f );
			ERR_check(ERR::ERR_ALWAYS, ret == WIN32_MAX_CHUNK_SIZE, "I/O error reading from file");
			pcstr += WIN32_MAX_CHUNK_SIZE;
			remBytes -= WIN32_MAX_CHUNK_SIZE;
		}
		if( remBytes ) {
			const size_t ret = fread( pcstr, 1, remBytes, f );
			ERR_check(ERR::ERR_ALWAYS, ret == remBytes, "I/O error reading last chunk from file");
		}
		return siz * count;
	}

	/**
	 * exception throwing IO
	 */
	inline size_t xfwrite(const void *str, const size_t siz, const size_t count, FILE *f) {
		// windows is bogus.
		//const size_t ret = fwrite( str, siz, count, f );
		//ERR_check(ERR::ERR_ALWAYS, ret == count && "I/O error writing into file");
		size_t remBytes = siz * count;
		const char *pcstr = (const char *)str;
		while( remBytes > WIN32_MAX_CHUNK_SIZE ) {
			const size_t ret = fwrite( pcstr, 1, WIN32_MAX_CHUNK_SIZE, f );
			ERR_check(ERR::ERR_ALWAYS, ret == WIN32_MAX_CHUNK_SIZE, "I/O error writing into file");
			pcstr += WIN32_MAX_CHUNK_SIZE;
			remBytes -= WIN32_MAX_CHUNK_SIZE;
		}
		if( remBytes ) {
			const size_t ret = fwrite( pcstr, 1, remBytes, f );
			ERR_check(ERR::ERR_ALWAYS, ret == remBytes, "I/O error writing last chunk into file");
		}
		return siz * count;
	}
	template<typename scalar_t>
	inline void xfwrite_scalar(const scalar_t s, FILE *f) {
		(void)xfwrite(&s, sizeof(scalar_t), 1, f);
		return;
	}
	template<typename scalar_t>
	inline size_t xfwrite_many(const scalar_t *str, const size_t count, FILE *f) {
		return xfwrite((const void *)str, sizeof(*str), count, f);
	}

	/**
	 * tentative - like CRT
	 * @return number of bytes read
	 */
	inline size_t xfread_tentative(void *str, const size_t siz, const size_t count, FILE *f) {
		// windows is bogus
		//const size_t ret = fread( str, siz, count, f );
		//ERR_check(ERR::ERR_ALWAYS, ret == count && "I/O error reading from file");
		size_t remBytes = siz * count;
		char *pcstr = (char *)str;
		while( remBytes > WIN32_MAX_CHUNK_SIZE ) {
			const size_t ret = fread( pcstr, 1, WIN32_MAX_CHUNK_SIZE, f );
			if( ret != WIN32_MAX_CHUNK_SIZE )
				return count * siz + ret - remBytes;
			pcstr += WIN32_MAX_CHUNK_SIZE;
			remBytes -= WIN32_MAX_CHUNK_SIZE;
		}
		if( remBytes ) {
			const size_t ret = fread( pcstr, 1, remBytes, f );
			if(ret != remBytes)
				return count * siz + ret - remBytes; 
		}
		return siz * count;
	}
	template<typename scalar_t>
	inline void xfread_scalar(scalar_t &s, FILE *f) {
		(void)xfread(&s, sizeof(s), 1, f);
		return;
	}
	template<typename scalar_t>
	inline size_t xfread_many(scalar_t *str, const size_t count, FILE *f) {
		return xfread((void *)str, sizeof(*str), count, f);
	}
	template<typename scalar_t>
	inline size_t xfread_vec(std::vector<scalar_t> &vec, FILE *f) {
		unsigned len;
		xfread_scalar<unsigned>(len, f);
		vec.resize(len);
		if(vec.size())
			return sizeof(unsigned) + xfread_many<scalar_t>(&vec[0], vec.size(), f);
		return sizeof(unsigned);
	}
	template<typename scalar_t>
	inline size_t xfread_tentative_many(scalar_t *str, const size_t count, FILE *f) {
		return xfread_tentative((void *)str, sizeof(*str), count, f);
	}
	/**
	 * portable write - size_t can change size
	 * will die if s is too large...
	 */

	template<typename scalar_t>
	inline size_t xfread_tentative_scalar(scalar_t &s, FILE *f) {
		return xfread_tentative(&s, sizeof(s), 1, f);
	}

	// specialized read/writes for strings
	inline void xfwrite_string(const std::string &s, FILE *f) {
		const uint32 slen = s.length();
		IO::xfwrite_scalar(slen, f);
		const char *buf = s.c_str();
		IO::xfwrite_many(buf, s.length(), f);
		return;
	}
	inline void xfread_string(std::string &s, FILE *f) {
		char buf[MAX_LINE_LENGTH];
		uint32 slen;
		IO::xfread_scalar(slen, f);
		ERR_check(ERR::ERR_SANITY, slen < (uint32)MAX_LINE_LENGTH, "string too long");
		IO::xfread_many(buf, slen, f);
		buf[slen] = 0;
		s.assign(buf);
	}

	/**
	 * compute the basename
	 * @return a pointer within the string corresponding to the start of the filename
	 */
	inline const char *basename(const char *basestr) {
		const char *p1 = strrchr(basestr, '\\');
		const char *p2 = strrchr(basestr, '/');
		// std::max if well defined
		if( p1 > p2 )  return p1+1;
		if( p2 > p1 )  return p2+1;
		if( p2 ) return p2+1;
		// p1 = p2 = 0
		return basestr;
	}
	/**
	 * compute the extension 
	 * @return a pointer within the string. or NULL if not found
	 */
	inline const char *fname_extension(const char *basestr) {
		const char *p = strrchr(basestr, '.');
		if( p )
			return p+1;
		return NULL;
	}
	/**
	 * compute the extension 
	 * @return a pointer within the string. or NULL if not found
	 */
	inline char *fname_extension(char *basestr) {
		char *p = strrchr(basestr, '.');
		if( p )
			return p+1;
		return NULL;
	}
	/**
	 * stem (remove extension) (inplace)
	 */
	inline void strip_ext(char buf[]) {
		char *p = fname_extension(buf);
		if(p) *(--p) = 0; // remove dot
		return;
	}

	/**
	 * parse range
	 * Ranges are of the form x.mfcc:9-10
	 * Ranges can be of the form: x.mfcc:9 (meaning 9-9), in that case, we return to = -1
	 * Ranges are typically processed as inclusive
	 * When no range is given, from  = to = -1
	 */
	inline void ParseRanges(char *buf, int &from, int &to) {
		char *p;
		from = -1;
		to = -1;
		if( (p = strrchr(buf, ':')) != NULL ) {
			char *q = p+1;
			for(; *q; ++q) {
				// pngT: use isdigit but include ctypes.h
				if( *q != '-' && ((*q > '9') || (*q < '0')) )
					break;
			}
			if( !*q ) { // a range
				for(q = p+1; *q; ++q) {
					if(*q == '-') {
						*q = 0;
						to = (int)atoi(q+1);
					}
				}
				from = (int)atoi(p+1);
				*p = 0;
			}
		}
	}
	/**
	 * open a file - please use auto_file_t
	 */
	inline FILE *file_open(const char *fname, const char *mode) {
		int err;
		FILE *f = NULL;
		if( !strcmp(fname, "-") ) {
			switch(*mode) {
				case 'w':
					f = stdout;
					break;
				case 'r':
					f = stdin;
					break;
				default:
					ERR_check(ERR::ERR_ALWAYS, 0, "file mode unsupported");
			}
			return f;
		}

#ifdef _GCC_
		f = fopen(fname, mode);
		err = !f;
#else
		int _save_fmode;
		_save_fmode = _fmode;
		_fmode &= ~0x8000;//_O_BINARY
		_fmode |= 0x4000;//_O_TEXT;
		f = fopen(fname, mode);
		err = !f;
		_fmode = _save_fmode; 
#endif
		if( err != 0 )
			(void)fprintf(stderr, "Unable to open '%s' in %s mode\n", fname, mode);
		ERR_check(ERR::ERR_USERINPUT, err==0, "file operation failed");
		return f;
	}
	/**
	 * close a file - please use auto_file_t
	 */
	inline void file_close(FILE *f) { if(f) fclose(f); }

	// already hit eof
	inline bool xfeof(FILE *f) { return feof(f)!=0; }

#ifdef _GCC_
# define EXE_EXTENSION
#else
# define EXE_EXTENSION ".exe"
#endif
	/**
	 * stack object that auto-closes files
	 */
	struct auto_file_t {
	private: auto_file_t(const auto_file_t &__forbidden_copy__); // forbidden
	protected:
		FILE *f;
		std::string pipecmd_;
		std::string pipemode_;
	public:
		inline auto_file_t(const char *fname, const char *mode, const bool autogzip = true, const bool isPipe = false) : f(0) {
			open(fname, mode, autogzip, isPipe);
		}
		// close current file and open a new one
		inline void open(const char *fname, const char *mode, const bool autogzip = true, const bool isPipe = false) {
			close();
			if( !fname ) return;
			if( isPipe ) {
				pipecmd_ = std::string(fname);
				pipemode_ = std::string(mode);
#ifdef _GCC_
 				ERR_check(ERR::ERR_USERINPUT, mode[0] == 'r' || mode[0] == 'w', "popen mode unsupported");
 				f = POPEN(fname, mode[0] == 'r' ? "r":"w");
#else
				f = POPEN(fname, mode);
#endif
				if( !f ) { (void)fprintf(stderr, "failed to open pipe %s %s\n", mode[0] == 'r' ? "from" : "to", pipecmd_.c_str()); perror("open"); }
				ERR_check(ERR::ERR_SANITY, f, "Cannot open pipe");
			} else {
				pipecmd_.clear();
				if( autogzip && IO::fname_extension(fname) && !strcmp(IO::fname_extension(fname), "gz") ) {
					char buf[MAX_LINE_LENGTH];
					if( *mode == 'r' )
						(void)sprintf(buf, "gzip" EXE_EXTENSION " -dc %s", fname);
					else
						(void)sprintf(buf, "gzip" EXE_EXTENSION " -c -7 > %s", fname);
					
					open( buf, mode, false, true );
				} else {
					if( autogzip && IO::fname_extension(fname) && !strcmp(IO::fname_extension(fname), "bz2") ) {
						char buf[MAX_LINE_LENGTH];
						if( *mode == 'r' )
							(void)sprintf(buf, "bzip2" EXE_EXTENSION " -dc %s", fname);
						else
							(void)sprintf(buf, "bzip2" EXE_EXTENSION " -c -7 > %s", fname);
						open( buf, mode, false, true );
					} else {
						if( autogzip && IO::fname_extension(fname) && !strcmp(IO::fname_extension(fname), "pz") ) {
							char buf[MAX_LINE_LENGTH];
							if( *mode == 'r' )
								(void)sprintf(buf, "punzip" EXE_EXTENSION " < %s", fname);
							else
								(void)sprintf(buf, "pzip" EXE_EXTENSION " > %s", fname);
							open( buf, mode, false, true );
						} else {
							ERR_check(ERR::ERR_SANITY, f == NULL, "internal state unclean");
							f = fname ? file_open(fname, mode) : NULL;
						}
					}
				}
			}
		}
		inline void close() { 
			if( !f ) return;
			if( pipecmd_.length() ) {
				if(f) PCLOSE(f);
				pipecmd_.clear();
			} else {
				if( f != stdin && f != stdout )
					file_close(f); 
			}
			f = NULL;
		}
		inline ~auto_file_t() { close(); }
		inline bool feof() const { return xfeof(static_cast<auto_file_t const *>(this)->f); }
		inline void rewind() {
			if( pipecmd_.length() )  // reopen
				open( std::string(pipecmd_).c_str(), pipemode_.c_str(), false, true );
			else
				::rewind(f); 
		}
		inline operator FILE*&() { return f; }
		inline bool isOpen() const { return !!f; }
		// only works on a rewindable file, at the beginning
		// @return true if it's a unicode file, and advance to readable bytes
		bool check_unicode() {
			uint16 signature;
			IO::xfread_scalar(signature, f);
			if( signature == 0xfeff || signature == 0xfffe ) {
				return true;
			}
			rewind(); // go back to beginning
			return false;
		}
	};
	/**
	 * iterate line by line using fgets
	 */
	template<typename char_t=char, int MAX_LINE_LEN=MAX_LINE_LENGTH>
	class line_iterator {
	private:
		char_t buf_[MAX_LINE_LEN];
		auto_file_t &file_;
		bool dochomp_;
		bool done_;
		size_t linecount_;
		bool fromUnicode_;
		std::vector<wchar_t> wbuf_;

		// return true if a line was read
		bool read_next_line() {
			if( fromUnicode_ ) {
#ifdef _GCC_
				printf("%s", "fgetws call in ngutil removed for g++ compatibility\nunicode not supported\n");
				exit(1);
#else
				buf_[0] = 0; // in case we fail
				const bool ret = fgetws(&wbuf_[0], wbuf_.size(), file_) != NULL;
				if( !ret )
					return ret; // false
				const size_t conv = wcstombs(buf_, &wbuf_[0], MAX_LINE_LEN);
				if( conv != (size_t)-1) buf_[conv] = 0; // nullterminate
				else buf_[0] = 0; // failed conversion entirely
				buf_[MAX_LINE_LEN-1] = 0;
				return true;
#endif /* _GCC_ */
			} else {
				const bool ret = fgets(buf_, MAX_LINE_LEN, file_) != NULL;
				buf_[MAX_LINE_LEN-1] = 0;
				return ret;
			}
		}
	public:
		line_iterator(auto_file_t &f, const bool dochomp = true) : file_(f), dochomp_(dochomp), done_(false), linecount_(0), fromUnicode_(false) {
		}
		void checkUnicode() { 
			fromUnicode_ = file_.check_unicode();  // file is in binary, o/w automatically translated
			if( fromUnicode_ )
				wbuf_.resize(MAX_LINE_LEN);
		}
		// is underlying file unicode
		bool isUnicode() const { return fromUnicode_; }
		/**
		 * read the next line and place it into internal buffer.
		 * @return true if not at end
		 */
		inline bool next() { 
			if( done_ ) return false;
			done_ = !read_next_line();
			if( done_ )
				return false;
			if( dochomp_ )
				PRS::xchomp(buf_);
			linecount_++;
			return !done_;
		}
		inline void reset() { 
			done_ = false; linecount_ = 0; 
			if( fromUnicode_ )
				(void)file_.check_unicode(); // skip header
		}
		// read internal buffer for current read line
		inline char *getline() { return buf_; }
		size_t lineCount() const { return linecount_; }
	};


	/**
	 * load a list of strings from a file, each string comes in a line.
	 */
	inline std::vector<std::string> LoadList(const char *fname) {
		std::vector<std::string> ret;

		auto_file_t f(fname, "rt");
		line_iterator<> line(f);
		while(line.next())
			ret.push_back(std::string(line.getline()));
		return ret;
	}
	/**
	 * Convert an element to a list
	 * Load elements from a file if listname starts with a '@' otherwise return a singleton
	 * The list can be sliced using the range @from-to@filename, using zero-best indices from and to, inclusive.
	 * If to is omitted, then to = from.
	 * Otherwise, space-separated items can also form a list.
	 */
	inline std::vector<std::string> GetList(const char *listname, const char csep = ' ') {
		std::vector<std::string> ret;
		if(!listname[0])
			return ret; // empty
		if(listname[0] == '@') {
			const char *p;
			if( (p = strchr(listname+1, '@')) != NULL ) {
				// for instance, @3-5@
				ret = LoadList(p+1);
				char buf[MAX_LINE_LENGTH];
				(void)strncpy_s(buf, listname+1, p - listname - 1);
				char *q = strchr(buf, '-');
				int from, to;
				if( q ) {
					*q = 0;
					from = atoi(buf);
					to = atoi(q+1);
				} else 
					to = from = atoi(buf);
				ERR_check(ERR::ERR_USERINPUT, from <= to, "range is invalid in @ notation");
				if( from > 0 )
					ret.erase(ret.begin(), ret.begin() + from);
				ret.resize(to-from+1);
			} else
				ret = LoadList(listname+1);
			return ret;
		}
		if( strchr(listname, csep) ) {
			const size_t len = strlen(listname);
			char ssep[2]; ssep[0] = csep; ssep[1] = 0;
			if( len > MAX_LINE_LENGTH-1 ) {
				char *b = new char[len+2];
				(void)strcpy(b, listname);
				ret = PRS::xtokenize(b, ssep);
				delete [] b;
			} else {
				char buf[MAX_LINE_LENGTH];
				(void)strcpy_s(buf, listname);
				ret = PRS::xtokenize(buf, ssep);
			}
		} else 
			ret.push_back(std::string(listname));
		return ret;
	}
}
template<typename T> T std_min(const T &t1, const T &t2) { return t1 < t2 ? t1: t2; }
template<typename T> T std_max(const T &t1, const T &t2) { return t1 > t2 ? t1: t2; }

#endif /* __NGUTIL_H */
