/**
 * Patrick Nguyen (panguyen@microsoft.com)
 * Copyright (c) 2009 Microsoft Corporation. All rights reserved.
 * 
 * Watch simple perf counters
 */

#ifndef __PERF_H
#define __PERF_H 1

#include "ngutil.h"

#include <pdh.h>
#include <pdhmsg.h>
#include <time.h>

#ifndef PDH_MAX_COUNTER_NAME
# define PDH_MAX_COUNTER_NAME 1024
# define PDH_MAX_COUNTER_PATH 2048
#endif


#define CPU_OBJ_INDEX        238
#define CPU_PROC_TIME       L"% User Time"
#define CPU_KERNEL_TIME     L"% Privileged Time"

#define NET_OBJ_INDEX        510
#define NET_BYTES_RECV      L"Bytes Received/sec"
#define NET_BYTES_SENT      L"Bytes Sent/sec"

#define DISK_OBJ_INDEX       236 // Physical Disk
#define DISK_BYTES_READ     L"Disk Read Bytes/sec"
#define DISK_BYTES_WRITE    L"Disk Write Bytes/sec"

class  Counter {
	PDH_HQUERY pdhQueryHandle_;

	struct SubCounter {
		std::vector<PDH_HCOUNTER> c1_counter_, c2_counter_;
		SubCounter(const Counter *parent, const DWORD perfidx, const wchar_t *c1_name, const wchar_t *c2_name, const wchar_t *hint, const bool doSum = false) {
			wchar_t		obj_name[PDH_MAX_COUNTER_NAME];
			wchar_t *	obj_list = NULL;
			DWORD		obj_list_size = 0;
			wchar_t *	inst_list = NULL;
			DWORD		inst_list_size = 0;
			wchar_t		counter_path[PDH_MAX_COUNTER_PATH];
			wchar_t *	inst = NULL;
			DWORD       namelen;
			PDH_STATUS  st;

			namelen = PDH_MAX_COUNTER_NAME;
			//st = PdhLookupPerfNameByIndexW(NULL, NET_OBJ_INDEX, obj_name, &namelen);
			st = PdhLookupPerfNameByIndexW(NULL, perfidx, obj_name, &namelen);

			ERR_check(ERR::ERR_SANITY, st == ERROR_SUCCESS, "[mon] failed to read perf metric");
			//namelen = PDH_MAX_COUNTER_NAME;
			//st = PdhLookupPerfNameByIndexW(NULL, NET_BYTES_RECV, c1_name, &namelen);

			st = PdhEnumObjectItemsW(NULL, NULL, obj_name,
				NULL, &obj_list_size,
				NULL, &inst_list_size,
				PERF_DETAIL_WIZARD, 0);

			// st should be more data

			// Allocate buffers to hold object and instance names
			obj_list = new wchar_t[obj_list_size];
			inst_list = new wchar_t[inst_list_size];

			st = PdhEnumObjectItemsW(NULL, NULL, obj_name,
				obj_list, &obj_list_size,
				inst_list, &inst_list_size,
				PERF_DETAIL_WIZARD, 0);

			ERR_check(ERR::ERR_SANITY, st == ERROR_SUCCESS, "[mon] failed to enumerate perf metric");

			PDH_HCOUNTER c1c, c2c;
			if( !doSum ) {
				for (inst = inst_list; *inst != 0; inst += wcslen(inst) + 1) {
					if( wcsstr(inst, hint) )
						break;
				}
				if( *inst == 0 ) {
					// didn't get the hint, use the first one
					inst = inst_list;
				}
				_snwprintf_s(counter_path, PDH_MAX_COUNTER_PATH, L"\\%ls(%ls)\\%ls",
					obj_name, inst, c1_name);

				st = PdhAddCounterW(parent->pdhQueryHandle_, counter_path, 0, &c1c);
				ERR_check(ERR::ERR_SANITY, st == ERROR_SUCCESS, "[mon] failed to register counter");
				c1_counter_.push_back(c1c);

				_snwprintf_s(counter_path, PDH_MAX_COUNTER_PATH, L"\\%ls(%ls)\\%ls",
					obj_name, inst, c2_name);

				st = PdhAddCounterW(parent->pdhQueryHandle_, counter_path, 0, &c2c);
				ERR_check(ERR::ERR_SANITY, st == ERROR_SUCCESS, "[mon] failed to register counter");
				c2_counter_.push_back(c2c);
			} else {
				// doSum
				ERR_check(ERR::ERR_SANITY, !hint, "[mon] cannot specify both doSum and hint");
				for (inst = inst_list; *inst != 0; inst += wcslen(inst) + 1) {
					if( wcsstr(inst, L"_Total") )
						continue;
					_snwprintf_s(counter_path, PDH_MAX_COUNTER_PATH, L"\\%ls(%ls)\\%ls",
						obj_name, inst, c1_name);

					st = PdhAddCounterW(parent->pdhQueryHandle_, counter_path, 0, &c1c);
					ERR_check(ERR::ERR_SANITY, st == ERROR_SUCCESS, "[mon] failed to register counter");
					c1_counter_.push_back(c1c);

					_snwprintf_s(counter_path, PDH_MAX_COUNTER_PATH, L"\\%ls(%ls)\\%ls",
						obj_name, inst, c2_name);

					st = PdhAddCounterW(parent->pdhQueryHandle_, counter_path, 0, &c2c);
					ERR_check(ERR::ERR_SANITY, st == ERROR_SUCCESS, "[mon] failed to register counter");
					c2_counter_.push_back(c2c);
				}
			}
			delete [] obj_list;
			delete [] inst_list;
		}
		void run(float &c1val, float &c2val) {
			PDH_STATUS st;
			PDH_FMT_COUNTERVALUE c1, c2;

			c1val = c2val = 0.0f;
			size_t k;
			for(k=0; k < c1_counter_.size(); ++k) {
				st = PdhGetFormattedCounterValue(c1_counter_[k], PDH_FMT_DOUBLE|PDH_FMT_NOCAP100, NULL, &c1);
				st = PdhGetFormattedCounterValue(c2_counter_[k], PDH_FMT_DOUBLE|PDH_FMT_NOCAP100, NULL, &c2);
				c1val += (float)c1.doubleValue;
				c2val += (float)c2.doubleValue;
			}
		}
	} *cpu_, *net_, *disk_;
	public:
		~Counter() { 
			delete cpu_;
			delete net_;
			delete disk_;
			PdhCloseQuery(pdhQueryHandle_); 
		}
		Counter() {
			PDH_STATUS st = PdhOpenQuery(NULL, 0, &pdhQueryHandle_);
			ERR_check(ERR::ERR_SANITY, st == ERROR_SUCCESS, "[mon] failed to open initialize API");

			cpu_ = new SubCounter(this, CPU_OBJ_INDEX, CPU_PROC_TIME, CPU_KERNEL_TIME, L"_Total", false);
			//net_ = new SubCounter(this, NET_OBJ_INDEX, NET_BYTES_RECV, NET_BYTES_SENT, L"Enterprise");
			// (net doesn't have a _Total, but on MSR SI HPC, Enterprise is the gigabit network which links nodes together)
			// do sum myself:
			net_ = new SubCounter(this, NET_OBJ_INDEX, NET_BYTES_RECV, NET_BYTES_SENT, 0, true);
			//
			disk_ = new SubCounter(this, DISK_OBJ_INDEX, DISK_BYTES_READ, DISK_BYTES_WRITE, 0, true);
			//disk_ = new SubCounter(this, DISK_OBJ_INDEX, DISK_BYTES_READ, DISK_BYTES_WRITE, L"_Total");

			// initialize counters
			st = PdhCollectQueryData(pdhQueryHandle_);
			ERR_check(ERR::ERR_SANITY, st == ERROR_SUCCESS, "[mon] failed to collect data from disk");
		}

		void run(float *cpuval, float *netval, float *diskval) {
			PDH_STATUS st;

			st = PdhCollectQueryData(pdhQueryHandle_);
			ERR_check(ERR::ERR_SANITY, st == ERROR_SUCCESS, "[mon] failed to collect data from CPU");

			//float cpuval[2], netval[2], diskval[2];

			cpu_->run(cpuval[0], cpuval[1]);
			net_->run(netval[0], netval[1]);
			disk_->run(diskval[0], diskval[1]);
			//(void)printf("C: %.2f  N: %.2f  D: %.2f\n", 
			//	cpuval[0] + cpuval[1], (netval[0] + netval[1])/1024.0f/1024.0f, (diskval[0] + diskval[1])/1024.0f/1024.0f);
		}
};

class SysMon {
	static DWORD __stdcall startThread(LPVOID param) {
		SysMon *me = (SysMon *)param;
		me->run();
		return 0;
	}
	Counter cnt_;
	bool cont_;
	HANDLE hthread_;
	const uint32 waitTime_;
	clock_t startTime_; // almost equal to zero

	FILE *f_;
	const size_t foffset_;

	void run() {
		startTime_ = clock();
		const uint32 alive = 1;
		const uint32 done = 0;
		while(cont_) {
			// each must have two values allocated
			float outbuf[6];
			cnt_.run(outbuf, outbuf+2, outbuf+4);
			(void)fseek(f_, foffset_, SEEK_SET);
			// format is: status, clock, meters
			IO::xfwrite_scalar(alive, f_);
			const float totElapsed = (float)(clock() - startTime_) / CLOCKS_PER_SEC;
			IO::xfwrite_scalar(totElapsed, f_);
			IO::xfwrite_many(outbuf, 6, f_);
			fflush(f_);
			Sleep(waitTime_);
		}
		(void)fseek(f_, foffset_, SEEK_SET);
		IO::xfwrite_scalar(done, f_);
		fflush(f_);
	}
public:
	// outbuf must have 
	// @see Counter::run for which values are filled in
	// @see SysMon::run for how outbuf is used
	SysMon(FILE *f, const size_t foffset, const uint32 waitTime) : f_(f), foffset_(foffset), cont_(true), hthread_(0), waitTime_(waitTime) {
		hthread_ = CreateThread(
			NULL, // sec attrib
			0, // stacksize
			&startThread,
			(void*)this,
			0, // creation flags
			NULL // out threaded id
			);
		//SetThreadPriority(hthread_, THREAD_PRIORITY_ABOVE_NORMAL);
		SetThreadPriority(hthread_, THREAD_PRIORITY_TIME_CRITICAL);
	}
	~SysMon() {
		cont_ = false;
		(void)printf("[sysmon] waiting for thread to finish...");fflush(stdout);
		DWORD ret;
		do {
			ret = WaitForSingleObject(hthread_, 1000);
			(void)printf(".");fflush(stdout);
		} while( ret != WAIT_OBJECT_0 );
		(void)printf(" ok(%d)\n", ret);
	}
};


#endif // __PERF_H
