/**
 * Patrick Nguyen (panguyen@microsoft.com)
 * Geoff Zweig (gzweig@microsoft.com)
 *
 * Copyright (c) 2010 Microsoft Corporation. All rights reserved.
 *
 * Command-line configuration for SCARF
 */

#ifndef __SCARF_CONFIG_H
#define __SCARF_CONFIG_H 1

#include "ngutil.h"
#include "commandline.h"

namespace CFG {
struct ScarfConfig {
	virtual void register_args(clparser &cl) const = 0;
	virtual void load_args(const clparser &cl) = 0;
};

/**
* Feature extraction config
*/
struct FEXConfig : public ScarfConfig {
	std::string basefiles;
	std::vector<bool> use_levenshtein_features;
	bool use_full_lm; 
	bool use_global_lm; 
	std::vector<std::string> stream;
	std::vector<size_t> expectOrder;
	std::vector<size_t> existOrder;
	std::vector<bool> lexternal;
	std::vector<std::string> stream_names;
	std::vector<std::string> external_format;

	void register_args(clparser &cl) const {
		cl.register_arg("-basefiles", "string", "baseline files", "", false);
		cl.register_arg("-use_full_lm", "bool", "use full LM features", "false", false);
		cl.register_arg("-use_global_lm", "bool", "use global LM weight (as opposed to link-based)", "true", false);
		cl.register_arg("-stream", "string vector", "feature stream files", "", false);
		cl.register_arg("-expectOrder", "size_t vector", "order of existence features", "", false);
		cl.register_arg("-existOrder", "size_t vector", "order of expectation features. Requires a dictionary", "", false);
		cl.register_arg("-use_levenshtein_features", "bool vector", "Add edit distance features. Requires a dictionary", "", false);
		cl.register_arg("-lexternal", "bool vector", "lexicalize external features", "", false);
		cl.register_arg("-stream_names", "string vector", "vector of stream names", "", false);
		cl.register_arg("-external_format", "string vector", "names of externally defined features in the lattice (comma-separated)", "", false);
	}
	void load_args(const clparser &cl) {
		basefiles = *((std::string *)cl.get("-basefiles"));
		use_levenshtein_features = *((std::vector<bool> *)cl.get("-use_levenshtein_features"));
		use_full_lm = *((bool *)cl.get("-use_full_lm"));
		stream = *((std::vector<std::string> *)cl.get("-stream"));
		use_global_lm = *((bool *)cl.get("-use_global_lm"));
		expectOrder = *((std::vector<size_t> *)cl.get("-expectOrder"));
		existOrder = *((std::vector<size_t> *)cl.get("-existOrder"));
		lexternal = *((std::vector<bool> *)cl.get("-lexternal"));
		stream_names = *((std::vector<std::string> *)cl.get("-stream_names"));
		external_format = *((std::vector<std::string> *)cl.get("-external_format"));
	}
};

/**
* MPI config -- for parallel processing
*/
struct MPIConfig : public ScarfConfig {
	std::string redirect_log;
	bool enable; 
	struct HeartBeat {
		size_t durms;
		std::string monfile;
	} heartbeat;
	size_t numreaders;
	std::string localProcDir;

	void register_args(clparser &cl) const {
		cl.register_arg("-mpi.redirect_log", "string", "redirect std out and err","",  false);
		cl.register_arg("-mpi.enable", "bool", "Enable MPI", "false", false);
		cl.register_arg("-mpi.heartbeat.durms", "size_t", "duration between heart beats", "1000", false);
		cl.register_arg("-mpi.heartbeat.monfile", "string", "where to put the dumps", "", false);
		cl.register_arg("-mpi.numreaders", "size_t", "number of readers in parallel when caching", "1", false);
		cl.register_arg("-mpi.localProcDir", "string", "local directory for provisioning process data", "", false);
	}
	void load_args(const clparser &cl) {
		redirect_log = *((std::string *)cl.get("-mpi.redirect_log"));
		enable = *((bool *)cl.get("-mpi.enable"));
		heartbeat.durms = *((size_t *)cl.get("-mpi.heartbeat.durms"));
		heartbeat.monfile = *((std::string *)cl.get("-mpi.heartbeat.monfile"));
		numreaders = *((size_t *)cl.get("-mpi.numreaders"));
		localProcDir = *((std::string *)cl.get("-mpi.localProcDir"));
	}
};

struct GlobalConfig : public ScarfConfig {
	bool binstdio; 
	bool bufstdio;
#ifndef _GCC_
	bool crtdbgbrk;
#endif 
#ifdef __USE_MPI
	struct MPIConfig mpi;
#endif // __USE_MPI

	void register_args(clparser &cl) const {
		cl.register_arg("-binstdio", "bool", "Binary stdin,err,out", "false", false);
		cl.register_arg("-bufstdio", "bool", "Buffer stderr,out", "false", false);
#ifndef _GCC_
		cl.register_arg("-crtdbgbrk", "bool", "Enter debugger (dbg)", "false", false);
#endif
#ifdef __USE_MPI
		mpi.register_args(cl);
#endif // __USE_MPI
	}
	void load_args(const clparser &cl) {
#ifdef __USE_MPI
		mpi.load_args(cl);
#endif
		binstdio = *((bool *)cl.get("-binstdio"));
		bufstdio = *((bool *)cl.get("-bufstdio"));
#ifndef _GCC_
		crtdbgbrk = *((bool *)cl.get("-crtdbgbrk"));
#endif
	}
};

struct TrainConfig : public ScarfConfig {
	std::string saveItModel;
	size_t num_iterations;
	float l1_reg; 
	float l2_reg;
	std::string num_constraints;
	std::string den_constraints;
	bool printScores; 
	bool printModel; 

	void register_args(clparser &cl) const {
		cl.register_arg("-saveItModel", "string", "save model name after each iteration. Model name may contain %d.", "", false);
		cl.register_arg("-num_iterations", "size_t", "number of RProp batch updates", "30", false);
		cl.register_arg("-l1_reg", "float", "weight for L1-norm prior", "0.0001", false);
		cl.register_arg("-l2_reg", "float", "weight for L2-norm prior", "0.000007", false);
		cl.register_arg("-num_constraints", "string", "numerator constraint files", "", false);
		cl.register_arg("-den_constraints", "string", "denominator constraint files", "", false);
		cl.register_arg("-printScores", "bool", "print scores for each file", "true", false);
		cl.register_arg("-printModel", "bool", "display model after training is done", "true", false);
	}
	void load_args(const clparser &cl) {
		saveItModel = *((std::string *)cl.get("-saveItModel"));
		num_iterations = *((size_t *)cl.get("-num_iterations"));
		l1_reg = *((float *)cl.get("-l1_reg"));
		l2_reg = *((float *)cl.get("-l2_reg"));
		num_constraints = *((std::string *)cl.get("-num_constraints"));
		den_constraints = *((std::string *)cl.get("-den_constraints"));
		printScores = *((bool *)cl.get("-printScores"));
		printModel = *((bool *)cl.get("-printModel"));
	}
};

struct DecodeConfig : public ScarfConfig {
	std::string den_constraints;
	std::string clamp;
	bool printModel;
	bool asConstraint;

	void register_args(clparser &cl) const {
		cl.register_arg("-den_constraints", "string", "denominator constraint files", "", false);
		cl.register_arg("-clamp", "string", "clamp features to some value: comma-separated name@value", "", false);
		cl.register_arg("-printModel", "bool", "show model BEFORE decoding", "false", false);
		cl.register_arg("-asConstraint", "bool", "write as a constraint file", "false", false);
	}
	void load_args(const clparser &cl) {
		den_constraints = *((std::string *)cl.get("-den_constraints"));
		clamp = *((std::string *)cl.get("-clamp"));
		printModel = *((bool *)cl.get("-printModel"));
		asConstraint = *((bool *)cl.get("-asConstraint"));
	}
};

struct AppConfig : public ScarfConfig {
	FEXConfig fex;
	std::string lm;
	size_t max_alphas; 
	std::string model;
	std::vector<std::string> dict;
	inline AppConfig() {  }

	void register_args(clparser &cl) const {
		fex.register_args(cl);
		cl.register_arg("-lm", "string", "ARPA language model file", "", true);
		cl.register_arg("-model", "string", "model file", "", true);
		cl.register_arg("-max_alphas", "size_t", "maximum number of alphas alive", "2147483647", false);
		cl.register_arg("-dict", "string vector", "list of dictionary files", "", false);
	}
	void load_args(const clparser &cl) {
		fex.load_args(cl);
		lm = *((std::string *)cl.get("-lm"));
		model = *((std::string *)cl.get("-model"));
		max_alphas = *((size_t *)cl.get("-max_alphas"));
		dict = *((std::vector<string> *)cl.get("-dict"));
	}
};

} // namespace CFG
#endif // __SCARF_CONFIG_H
