// Geoffrey Zweig (gzweig@microsoft.com)
// Patrick Nguyen (panguyen@microsoft.com)
// Copyright (c) 2009 Microsoft Corporation. All rights reserved.

#include "lm.h"
#include <iostream>

#include "scarfread.h"

void UttIterator::renumber_times() {
	sort(times_.begin(), times_.end());
	vector<int>::iterator vi = unique(times_.begin(), times_.end());
	times_.erase(vi, times_.end());
	maxtime_ = times_.back();

	if (times_.front() < minconstrainttime_ || times_.back() > maxconstrainttime_) {
		cout << "Error: constraints do not span the observations\n";
		cout << "Constraints: " << minconstrainttime_ << " -> " << maxconstrainttime_ << endl;
		cout << "All times: " << times_.front() << " -> " << times_.back() << endl;
		exit(1);
	}

	map<int, int> newtime4;
	for (int i=0; i<(int)times_.size(); i++)
		newtime4[times_[i]] = i+1;
	maxtime_ = newtime4[maxtime_];
	constraint_t::iterator mi;
	constraint_t nm;
	for (mi=words_starting_.begin(); mi != words_starting_.end(); mi++) {
		int nt = newtime4[mi->first.t];
		nm[time_lat_rec(nt,mi->first.from)] = mi->second;
		vector<wordspan> &wsv = nm[time_lat_rec(nt,mi->first.from)];
		for (size_t j=0; j<wsv.size(); j++) {
			wsv[j].starttime = newtime4[wsv[j].starttime];
			wsv[j].endtime = newtime4[wsv[j].endtime];
		}
	}
	words_starting_ = nm;

	for (size_t i=0; i < units4stream_.size(); i++)
		for (size_t j=0; j < units4stream_[i].size(); j++)
			units4stream_[i][j].time = newtime4[units4stream_[i][j].time];

	for (size_t i=0; i < baseline_.size(); i++)
		baseline_[i].time = newtime4[baseline_[i].time];
}

void UttIterator::save_numbered() {
		dtimes_ = times_;
		hold_detects_ = units4stream_;
		hold_base_ = baseline_;
}
void UttIterator::restore_numbered() {
	// restore original numbering
	times_ = dtimes_;
	units4stream_ = hold_detects_;
	baseline_ = hold_base_;
}

//--region: iteration
bool UttIterator::moveNext() {
	times_.clear();
	maxtime_ = -1;

	const bool hasNext = read_detections_and_baseline(); // set times and units4stream_, and read the baseline
	if( !hasNext ) return false;

	// get the numbering of the detectors
	save_numbered();
	return true;
}

bool UttIterator::skipNext(const bool skipDetect, const bool skipBase) {
	times_.clear();
	maxtime_ = -1;

	const bool hasNext = skip_detections_and_baseline(skipDetect, skipBase);  // skip over detections and baseline
	if( !hasNext ) return false;

	return true;
}

void UttIterator::getNumConstraints() {
	read_num_constraints();
	renumber_times();  
}

void UttIterator::getDenConstraints() {
	restore_numbered();
	read_den_constraints();
	renumber_times();
}

void UttIterator::skipNumConstraints() {
	const bool hasNum = numConstraints_->moveNext(true);
	ERR_check(ERR::ERR_SANITY, hasNum, "trying to skip numerator constraints while at end");
	curfn_ = numConstraints_->currentName();
}

void UttIterator::skipDenConstraints() {
	const bool hasDen = denConstraints_->moveNext(true);
	ERR_check(ERR::ERR_SANITY, hasDen, "trying to skip denominator constraints while at end");
	curfn_ = denConstraints_->currentName();
}


//--endregion

bool UttIterator::read_detections() {
	units4stream_.resize(uttDetectors_.size());
	size_t streamIdx;
	for(streamIdx = 0; streamIdx < uttDetectors_.size(); ++streamIdx){
		const bool hasDetectors = uttDetectors_[streamIdx]->moveNext();
		if( !hasDetectors ) return false;
		const RawDetectorReader *dt = uttDetectors_[streamIdx];
		times_.insert(times_.end(), dt->currentTimes().begin(), dt->currentTimes().end());
		units4stream_[streamIdx] = dt->current();
	}
	return true;
}

bool UttIterator::skip_detections() {
	size_t streamIdx;
	for(streamIdx = 0; streamIdx < uttDetectors_.size(); ++streamIdx){
		const bool hasDetectors = uttDetectors_[streamIdx]->skipNext();
		if( !hasDetectors ) return false;
	}
	return true;
}

bool UttIterator::read_baseline() {
	if (fexConfig_.basefiles.length() > 0) {
		ERR_check(ERR::ERR_SANITY, baselineReader_, "need to initialize raw detectors");
		const bool hasBaseline = baselineReader_->moveNext();
		if(!hasBaseline) return false;
		times_.insert( times_.end(), baselineReader_->currentTimes().begin(), baselineReader_->currentTimes().end() );
		baseline_ = baselineReader_->current();
	}
	return true;
}

bool UttIterator::skip_baseline() {
	if (fexConfig_.basefiles.length() > 0) {
		const bool hasBaseline = baselineReader_->skipNext();
		if(!hasBaseline) return false;
	}
	return true;
}

bool UttIterator::read_detections_and_baseline() {
	const bool hasDetections = read_detections();
	if(!hasDetections) return false;
	const bool hasBaseline = read_baseline();
	if(uttDetectors_.size()) {
		ERR_check(ERR::ERR_SANITY, hasBaseline, "streams has detections but no baseline read");
	}
	return hasBaseline && denConstraints_->hasNext();
}

bool UttIterator::skip_detections_and_baseline(const bool skipDetections, const bool skipBase) {
	const bool hasDetections = skipDetections? skip_detections() : read_detections();
	if(!hasDetections) return false;
	const bool hasBaseline = skipBase ? skip_baseline() : read_baseline();
	if(uttDetectors_.size()) {
		ERR_check(ERR::ERR_SANITY, hasBaseline, "streams has detections but no baseline read");
	}
	return hasBaseline && denConstraints_->hasNext();
}

void UttIterator::read_den_constraints() {
	const bool hasDen = denConstraints_->moveNext();
	ERR_check(ERR::ERR_SANITY, hasDen, "trying to read denominator constraints past the end");
	curfn_ = denConstraints_->currentName();
	times_.insert(times_.end(), denConstraints_->currentTimes().begin(), denConstraints_->currentTimes().end());
	words_starting_ = denConstraints_->current();
	minconstrainttime_ = denConstraints_->get_min_time_constraint();
	maxconstrainttime_ = denConstraints_->get_max_time_constraint();
}

void UttIterator::read_num_constraints() {
	const bool hasNum = numConstraints_->moveNext();
	ERR_check(ERR::ERR_SANITY, hasNum, "trying to read numerator constraints past the end");
	curfn_ = numConstraints_->currentName();
	// do the numerator computation
	times_.insert(times_.end(), numConstraints_->currentTimes().begin(), numConstraints_->currentTimes().end());
	words_starting_ = numConstraints_->current();
	minconstrainttime_ = numConstraints_->get_min_time_constraint();
	maxconstrainttime_ = numConstraints_->get_max_time_constraint();
}

UttIterator::~UttIterator() {
	if( numConstraints_ ) delete numConstraints_;
	if( denConstraints_ ) delete denConstraints_;
	if( baselineReader_ ) delete baselineReader_;
	size_t k;
	for(k=0; k < uttDetectors_.size(); ++k) delete uttDetectors_[k];
}

void UttIterator::init_constraint_readers(const char *trainFiles, const char *denFiles) {
	numConstraints_ = new ConstraintsReader(trainFiles, idx4word_, word4idx_, extern_name_, fexConfig_);
	denConstraints_ = new ConstraintsReader(denFiles, idx4word_, word4idx_, extern_name_, fexConfig_);
}

void UttIterator::init_raw_detectors() {
	size_t streamIdx = 0;
	ERR_check(ERR::ERR_SANITY, idx4stream_sym_.size() == fexConfig_.stream.size(), "more unit mappings (dictionaries) then number of streams");
	for(; streamIdx < fexConfig_.stream.size(); ++streamIdx) {
		RawDetectorReader *dt = new RawDetectorReader(fexConfig_.stream_names[streamIdx], fexConfig_.stream[streamIdx].c_str(), idx4stream_sym_[streamIdx]);
		(void)printf("initialized a raw detector: %s\n", dt->name().c_str());
		uttDetectors_.push_back(dt);
	}
	if( fexConfig_.basefiles.length() > 0 ) {
		const std::string baseline_name("baseline");
		baselineReader_ = new BaselineReader(baseline_name, fexConfig_.basefiles.c_str(), idx4word_, word4idx_, false);
		(void)printf("Added baseline detector\n");
	}
}

void UttIterator::rewind() {
	ERR_check(ERR::ERR_USERINPUT, !numConstraints_->hasNext(), "Too many numerator constraint utterances");
	numConstraints_->rewind();
	ERR_check(ERR::ERR_USERINPUT, !denConstraints_->hasNext(), "Too many denominator constraint utterances");
	denConstraints_->rewind();
	if ( fexConfig_.basefiles.length() > 0 ) {
		const bool moreBase = baselineReader_->moveNext();
		ERR_check(ERR::ERR_USERINPUT, !moreBase, "Too many baseline utterances");
		baselineReader_->rewind();
	}
	size_t streamIdx;
	for(streamIdx = 0; streamIdx < uttDetectors_.size(); ++streamIdx) {
		const bool hasMoreStream = uttDetectors_[streamIdx]->moveNext();
		ERR_check(ERR::ERR_USERINPUT, !hasMoreStream, "Too many stream utterances");
		uttDetectors_[streamIdx]->rewind();
	}
}

