// Geoffrey Zweig (gzweig@microsoft.com)
// Patrick Nguyen (panguyen@microsoft.com)
// Copyright (c) 2009 Microsoft Corporation. All rights reserved.

#ifndef SCARF_READ_H
#define SCARF_READ_H 1

#include "ngutil.h"
#include "scarf_config.h"
#include "scarftypes.h"

class MultiFiles {
	std::vector<std::string> files_;
	const bool isMultiFile_;
	std::string curfn_, nextfn_;
	IO::auto_file_t *curf_;
	IO::line_iterator<> *curlines_;
	size_t cur_;

	static bool detect_multifile(const std::vector<std::string> &files) {
		if( files.size() == 0 ) return false;
		if( strstr(files[0].c_str(), ".B") ) return true;
		return false;
	}
public:
	MultiFiles(const char *list) 
		: files_(IO::GetList(list))
		, isMultiFile_(detect_multifile(files_))
		, curf_(0), curlines_(0)
		, cur_(size_t(-1))
	{
		if( isMultiFile_ ) {
			ERR_check(ERR::ERR_NOTIMPL, files_.size() == 1, "cannot have multiple B files");
			curf_ = new IO::auto_file_t(files_[0].c_str(), "rt");
			curlines_ = new IO::line_iterator<>(*curf_);
			const bool hasOne = curlines_->next();
			if( hasOne ) nextfn_.assign(curlines_->getline());
		}
	}

	bool hasNextFile() const {
		if( isMultiFile_ ) return nextfn_.length() != 0;
		return cur_+1 < files_.size();
	}

	bool moveNextFile() {
		if( !hasNextFile() ) return false;
		if( isMultiFile_ ) {
			nextfn_.swap(curfn_);
			return true;
		}
		cur_++;
		if( curf_ ) delete curf_; curf_ = 0;
		if( curlines_ ) delete curlines_; curlines_ = 0;
		ERR_check(ERR::ERR_SANITY, cur_ < files_.size(), "invalid file idx while moving to next file");
		curf_ = new IO::auto_file_t(files_[cur_].c_str(), "rt");
		curlines_ = new IO::line_iterator<>(*curf_);
		return true;
	}
	// @return true if we moved to the next line successfully, false if hit EOF on the current file
	bool moveNextLine() {
		ERR_check(ERR::ERR_SANITY, curlines_, "trying to move to next line but already at end");
		const bool hasLine = curlines_->next();
		if( isMultiFile_ ) {
			ERR_check(ERR::ERR_SANITY, hasLine, "premature EOF before a '.' in B file");
			if(!strcmp(curlines_->getline(), ".")) {
				const bool hasNextFile = curlines_->next();
				if( hasNextFile ) nextfn_.assign(curlines_->getline());
				else nextfn_.clear();
				return false;
			}
		}
		return hasLine;
	}
	const char *curFileName() const {
		if( isMultiFile_ ) return curfn_.c_str();
		else return files_[cur_].c_str();
	}
	char *curLine() {
		return curlines_->getline();
	}
	void rewind() {
		if( isMultiFile_ ) {
			curf_->rewind();
			curlines_->reset();
			const bool hasOne = curlines_->next();
			if( hasOne ) nextfn_.assign(curlines_->getline());
			return;
		}
		if( curf_ ) delete curf_; curf_ = 0;
		if( curlines_ ) delete curlines_; curlines_ = 0;
		cur_ = size_t(-1);
	}
	~MultiFiles() { if( curf_ ) delete curf_; if( curlines_ ) delete curlines_; }
};

inline float parse_extern_feat(char *p, const string &expected_name, const vector<string> &format, char *linebuff) {
	ERR_check(ERR::ERR_SANITY, p, "wrong format while parsing external feature");
	char *p2 = strchr(p, '=');
	ERR_check(ERR::ERR_SANITY, p2, "wrong format while parsing external feature (missing '=')");
	float v((float)atof(p2+1));
	string f(p, p2);
	if (f != expected_name) {
		cout << "External features " << PRS::xjoin(",", format) << " expected, but\n";
		cout << linebuff << "\nseen\n";
		exit(1);
	}
	return v;
}

class ConstraintsReader {
	vocab_t &idx4word_;
	rvocab_t &word4idx_;
	MultiFiles files_;
	size_t next_;
	constraint_t words_starting_;
	vector<int> times_;
	const vector<string> &extern_name_;
	const CFG::FEXConfig &fexConfig_;

	void get_constraints() {
		words_starting_.clear();
		times_.clear();
		int maxtime = -1;  // used to decide which lattice arcs are terminal
		bool lattice = false;
		vector<vector<int> > wordrecs;

		// expect a file with format:
		// (word starttime endtime)+
		char buff1[MAX_LINE_LENGTH];
		char ename[MAX_LINE_LENGTH], temp[MAX_LINE_LENGTH];
		while( files_.moveNextLine() ) {
			int st, et;
			std::vector<std::string> fields = PRS::xtokenize(files_.curLine());
			ERR_check(ERR::ERR_SANITY, fields.size() >= 3 || fields.size() <= 6, "wrong number of fields in constraint file");
			st = (int)atoi(fields[0].c_str());
			et = (int)atoi(fields[1].c_str());
			(void)strcpy(buff1, fields[2].c_str());
			maxtime = (et>maxtime)?(et):(maxtime);

			int lfrom=time_lat_rec::lstart, lto=time_lat_rec::lstart;  // initialize to lstart as default
			if (fields.size() >= 5) {  // we have lattice constraints
				lfrom = (int)atoi(fields[3].c_str());
				lto = (int)atoi(fields[4].c_str());
				lattice = true;
			}

			vector<float> efvals(extern_name_.size());
			int nef = 0;
			if (fields.size()==4 || fields.size()==6) { // st et word [sp ep] annotation
				if (fields.size()==4)
					strcpy(ename, fields[3].c_str());
				else
					strcpy(ename, fields[5].c_str());
				strcpy(temp, ename);
				char *p = strtok(ename, ",");
				do {
					ERR_check(ERR::ERR_SANITY,  (size_t)nef < extern_name_.size(), "too many external features in the lattice");
					efvals[nef] = parse_extern_feat(p, extern_name_[nef], extern_name_, temp);
					nef++;
				} while ( (p=strtok(NULL, ",")) != NULL );
			}		

			if ((size_t)nef != extern_name_.size()) {
				cout << "Error: " << nef << " external features read from lattice, but "
					 << extern_name_.size() << " expected\n";
				exit(1);
			}

			// add it
			if (idx4word_.find(string(buff1)) == idx4word_.end()) {
				cout <<  buff1 << " in constraint file " << files_.curFileName() << " but not in model\n";
				int minwd = vocab_util::get_minwd(idx4word_);
				idx4word_[buff1] = --minwd;
				word4idx_[minwd] = buff1;
			}

			if (st > et) {
				cout << "constraint file with end time before start time\n";
				cout << files_.curFileName() << endl;
				exit(1);
			}

			int w;
			words_starting_[time_lat_rec(st,lfrom)].push_back(
				wordspan((w=idx4word_[string(buff1)]), st, et, lto, efvals));
			times_.push_back(st);
			times_.push_back(et);

			vector<int> wr(5);
			wr[0] = st; wr[1] = et; wr[2] = lfrom; wr[3] = lto; wr[4] = w;
			wordrecs.push_back(wr);
		}

		if (!wordrecs.size()) {
			cout << "Error: empty contraint file " << currentName() << endl;
			exit(1);
		}

		// mark the final states
		for (constraint_t::iterator ci=words_starting_.begin(); ci!=words_starting_.end(); ci++) { 
			vector<wordspan> &vw = ci->second;
			for (size_t i=0; i<vw.size(); i++) 
				if (vw[i].endtime == maxtime) 
					vw[i].lat_end_state = time_lat_rec::lend;
		}

		// validate the constraints
		sort(wordrecs.begin(), wordrecs.end());
		for (size_t i=0; i<wordrecs.size()-1; i++) {
			if (wordrecs[i] == wordrecs[i+1]) {
				cout << "Error: inconsistent contraints file - the same word is listed twice ";
				cout << "with the same start-time/end-time and lattice positions:\n";
				cout << wordrecs[i][0] << " " << wordrecs[i][1] << " " << word4idx_[wordrecs[i][4]];
				if (lattice)
					cout << " " << wordrecs[i][2] << " " << wordrecs[i][3];
				cout << endl;
				exit(1);
			}
		}
	}
	void skip_constraints() {
		words_starting_.clear();
		times_.clear();
		while( files_.moveNextLine() ) {}
	}
public:
	ConstraintsReader(const char *fn, vocab_t &voc, rvocab_t &rvoc, const vector<string> &extern_name, const CFG::FEXConfig &fexConfig) 
		: idx4word_(voc), word4idx_(rvoc), files_(fn), extern_name_(extern_name), fexConfig_(fexConfig)
	{}

	bool hasNext() const { return files_.hasNextFile(); }
	bool moveNext(bool skip = false) {
		const bool hasMore = files_.moveNextFile();
		if( !hasMore ) return false;
		if(skip)
			skip_constraints();
		else
			get_constraints();
		return true;
	}
	void rewind() {
		files_.rewind();
	}
	
	const constraint_t &current() const { return words_starting_; }
	const std::vector<int> &currentTimes() const { return times_; }

	int get_min_time_constraint() const {
		return *(std::min_element(times_.begin(), times_.end()));
	}
	int get_max_time_constraint() const {
		return *(std::max_element(times_.begin(), times_.end()));
	}

	const char *currentName() const { 
		return files_.curFileName();
	}
};

class TranscriptReader {
	std::vector<std::string> files_;
	size_t next_;

	std::vector<std::string> cur_;
public:
	TranscriptReader(const char *fn) : files_(IO::GetList(fn)), next_(0) { }
	bool moveNext() {
		if( next_ == files_.size() ) return false;
		cur_.clear();
		IO::auto_file_t in(files_[next_].c_str(), "rt");
		IO::line_iterator<> lines(in);
		while(lines.next())
			cur_.push_back(lines.getline());
	}
	const char *currentFN() const { 
		ERR_check(ERR::ERR_CONDITION, next_ != 0, "need to moveNext before asking for current filename");
		return files_[next_-1].c_str(); 
	}
	const std::vector<std::string> &current() const { return cur_; }
};

class RawDetectorReader {
	std::string name_;
	MultiFiles files_;

	const vocab_t &symname2idx_;
	vector<detection> cur_;
	std::vector<int> times_;

	const bool hasStreamHeader_;

	int getUnitID(const char *unitName) const {
		const vocab_t::const_iterator it = symname2idx_.find(unitName);
		ERR_check(ERR::ERR_SANITY, it != symname2idx_.end(), "unit name not found in the dictionary, cannot dynamically expand");
		return it->second;
	}
public:
	RawDetectorReader(const std::string &name, const char *fn, vocab_t &unitvoc, const bool hasStreamHeader = true)
		: name_(name), files_(fn), symname2idx_(unitvoc), hasStreamHeader_(hasStreamHeader) { }
	bool moveNext() {
		const bool hasMore = files_.moveNextFile();
		if( !hasMore ) return false;
		cur_.clear();
		times_.clear();
		char buff1[MAX_LINE_LENGTH], buff2[MAX_LINE_LENGTH], buff3[MAX_LINE_LENGTH];
		const bool hasFirst = files_.moveNextLine();
		ERR_check(ERR::ERR_SANITY, hasFirst, "detector must have a header");
		int nv = hasStreamHeader_ ? 
			sscanf(files_.curLine(), "%s%s%s", buff1, buff2, buff3) :
			sscanf(files_.curLine(), "%s%s", buff1, buff2);
		ERR_check(ERR::ERR_USERINPUT, nv==2 + hasStreamHeader_, "invalid number of fields in header of detector");
		ERR_check(ERR::ERR_USERINPUT, buff1[0] == '#', "ill-formed header in detection file: missing '#'");
		if( hasStreamHeader_ ) {
			ERR_check(ERR::ERR_USERINPUT, !strcmp(buff3, "stream"), "missing 'stream' keyword in header");
		}
		ERR_check(ERR::ERR_SANITY, name_.compare(buff2) == 0, "inconsistent stream specififed");

		int t;
		while(files_.moveNextLine()) {
			nv = sscanf(files_.curLine(), "%s%i", buff1, &t);
			ERR_check(ERR::ERR_USERINPUT, nv==2, "detection invalid: must have symbol midpoint"); 
			if (!symname2idx_.count(buff1)) {
				cout << "unknown unit " << buff1 << " encountered in detector file\n";
				exit(1);
			}
			cur_.push_back(detection(getUnitID(buff1), t));
			times_.push_back(t);
		}
		sort(cur_.begin(), cur_.end());
		return true;
	}
	bool skipNext() {
		const bool hasMore = files_.moveNextFile();
		if( !hasMore ) return false;
		while(files_.moveNextLine()) {}
		return true;
	}
	inline const std::vector<detection> &current() const { return cur_; }
	inline const std::vector<int> &currentTimes() const { return times_; }
	inline void rewind() { files_.rewind(); }
	inline const std::string &name() { return name_; }
};

class BaselineReader {
	std::string name_;
	MultiFiles files_;

	vocab_t &symname2idx_;
	rvocab_t &idx2word_;
	vector<detection> cur_;
	std::vector<int> times_;
	const bool hasStreamHeader_;

	int getUnitID(const char *unitName) {
		const vocab_t::const_iterator it = symname2idx_.find(unitName);
                if (it == symname2idx_.end()) {
		    cout <<  unitName << " in baseline file " << files_.curFileName() << " but not in model; adding name\n";
		    int minwd = vocab_util::get_minwd(symname2idx_);
		    symname2idx_[unitName] = --minwd;
    		    idx2word_[minwd] = unitName;
                    return minwd;
                }
	        return it->second;
	}
public:
	BaselineReader(const std::string &name, const char *fn, vocab_t &unitvoc, rvocab_t &idx2word, const bool hasStreamHeader = true)
		: name_(name), files_(fn), symname2idx_(unitvoc), idx2word_(idx2word), hasStreamHeader_(hasStreamHeader) { }
	bool moveNext() {
		const bool hasMore = files_.moveNextFile();
		if( !hasMore ) return false;
		cur_.clear();
		times_.clear();
		char buff1[MAX_LINE_LENGTH], buff2[MAX_LINE_LENGTH], buff3[MAX_LINE_LENGTH];
		const bool hasFirst = files_.moveNextLine();
		ERR_check(ERR::ERR_SANITY, hasFirst, "base file empty");
		int nv = hasStreamHeader_ ? 
			sscanf(files_.curLine(), "%s%s%s", buff1, buff2, buff3) :
			sscanf(files_.curLine(), "%s%s", buff1, buff2);
		ERR_check(ERR::ERR_SANITY, nv==2 + hasStreamHeader_, "wrong number of fields in baseline file");
		ERR_check(ERR::ERR_USERINPUT, buff1[0] == '#', "missing '#' in header in baseline file");
		if( hasStreamHeader_ ) {
			ERR_check(ERR::ERR_USERINPUT, !strcmp(buff3, "stream"), "missing 'stream' keyword");
		}
		ERR_check(ERR::ERR_SANITY, name_.compare(buff2) == 0, "unexpected name in baseline file");

		int t;
		while(files_.moveNextLine()) {
			nv = sscanf(files_.curLine(), "%s%i", buff1, &t);
			ERR_check(ERR::ERR_USERINPUT, nv==2, "baseline format violated: must be symbol midpoint"); 
			cur_.push_back(detection(getUnitID(buff1), t));
			times_.push_back(t);
		}
		sort(cur_.begin(), cur_.end());
		return true;
	}
	// skip over next to file
	bool skipNext() {
		const bool hasMore = files_.moveNextFile();
		if( !hasMore ) return false;
		while(files_.moveNextLine()) {}
		return true;
	}
	inline const std::vector<detection> &current() const { return cur_; }
	inline const std::vector<int> &currentTimes() const { return times_; }
	inline void rewind() { files_.rewind(); }
	inline const std::string &name() { return name_; }
};

class UttIterator {
protected:
	//--region: 
	vector<vector<detection> > units4stream_;
	constraint_t words_starting_;
	vector<int> times_;  // of detections and word boundaries
	vector<detection> baseline_;
	int maxtime_;  // of any observation or constraint
	int maxconstrainttime_, minconstrainttime_;
	//--endregion
	//--region:tmp fields, used for saving and restoring.
	// store the original times and detections before they get renumbered
	vector<int> dtimes_;
	vector<vector<detection> > hold_detects_;
	vector<detection> hold_base_;
	//--endregion

	const CFG::FEXConfig &fexConfig_;
	const vector<string> &extern_name_;

	const char *curfn_;

	//--region: string to idx mappings
	vocab_t  &idx4word_;
	rvocab_t &word4idx_;
	vector<vocab_t> &idx4stream_sym_;  // vector indexes the stream, ie the unit vocab
	//--endregion

	//--region: iterators
	ConstraintsReader *numConstraints_, *denConstraints_;
	std::vector<RawDetectorReader*> uttDetectors_;
	BaselineReader *baselineReader_;
	//--endregion

	void renumber_times();
	void save_numbered();
	void restore_numbered();
	bool read_detections_and_baseline();
	bool skip_detections_and_baseline(const bool skipDetections, const bool skipBase);
	void read_num_constraints();
	void read_den_constraints();
	bool read_detections();
	bool skip_detections();
	bool read_baseline();
	bool skip_baseline();

public:

	UttIterator(const CFG::FEXConfig &fexConfig, const vector<string> &extern_name, vocab_t  &idx4word, rvocab_t &word4idx, vector<vocab_t> &idx4stream_sym) 
		: fexConfig_(fexConfig), extern_name_(extern_name), idx4word_(idx4word), word4idx_(word4idx), idx4stream_sym_(idx4stream_sym),
		numConstraints_(0), denConstraints_(0), baselineReader_(0)
	{}
	~UttIterator();

	void init_constraint_readers(const char *trainFiles, const char *denFiles);
	void init_raw_detectors();
	void init_has_features(const char *hasFiles) {
		if( !hasFiles ) return;
	}

	//--region: iteration
	bool moveNext();
	bool skipNext(const bool skipDetect = true, const bool skipBase = true); // move to next without reading symbols
	void getNumConstraints();
	void getDenConstraints();
	void skipNumConstraints();
	void skipDenConstraints();

	void rewind();

	//--region: current fields accessors
	inline const vector<vector<detection> >  &cur_units4stream()      const { return units4stream_; }
	inline const constraint_t                &cur_words_starting()    const { return words_starting_; }
	inline const vector<int>                 &cur_times()             const { return times_; }
	inline const vector<detection>           &cur_baseline()          const { return baseline_; }
	inline       int                          cur_maxtime()           const { return maxtime_; }

	inline       vector<vector<detection> >  &cur_units4stream()            { return units4stream_; }
	inline       constraint_t                &cur_words_starting()          { return words_starting_; }
	inline       vector<int>                 &cur_times()                   { return times_; }
	inline       vector<detection>           &cur_baseline()                { return baseline_; }
	inline       const char *                 cur_constraintFN()      const { return curfn_; }
	//--endregion
};

#endif


