// Geoffrey Zweig (gzweig@microsoft.com)
// Patrick Nguyen (panguyen@microsoft.com)
// Copyright (c) 2009 Microsoft Corporation. All rights reserved.

#ifndef SCARF_TYPES_H
#define SCARF_TYPES_H 1

struct wordspan {
	int idx, starttime, endtime, lat_end_state;
	vector<float> external;  // to hold externally defined feature values - from annotated lattices
	wordspan(int i, int s, int e, int le, vector<float> &vf) 
		{idx=i; starttime=s; endtime=e; lat_end_state=le; external = vf;}
	bool operator == (const wordspan &other)  const
	{return idx==other.idx && starttime==other.starttime && endtime==other.endtime && lat_end_state==other.lat_end_state;}
};

struct time_lat_rec { 
	// holds the start and end states of a lattice arc and the start time of its successors (1 + its end time)
	// used to index the successors of a lattice arc
	// a dummy start state of lstart is used when lattice constraints are not present
	// the initial states must be numbered lstart (0)!
	int t, from;
	static const int lstart=0, lend=INT_MAX;  // start and end state numbers
	time_lat_rec(int time, int f) {t=time; from=f;}
	bool operator < (const time_lat_rec &other) const // sort on time and start state
		{ if (t==other.t) return (from<other.from); else return (t<other.t); }
};

struct detection {
	int unit, time;
	detection(int u, int t) {unit=u; time=t;}
	bool operator < (const detection &other) const { return time < other.time; }
	bool operator == (const detection &other) const { return time == other.time && unit == other.unit; }
};

typedef map<time_lat_rec, vector<wordspan> > constraint_t;
typedef map<string, int> vocab_t;
typedef map<int, string> rvocab_t;

namespace vocab_util {
	inline int get_minwd(const map<string, int> &idx4word) {
		int minwd=0;
		for (map<string,int>::const_iterator mi = idx4word.begin(); mi!=idx4word.end(); mi++)
			minwd = (minwd<mi->second)?(minwd):(mi->second);
		return minwd;
	}
};

struct statetime {
	lmidx state;
	int t, w; // w is necessary to keep track of value of words that are unk to the lm
	int latpos; // position in lattice if lattice constraints are used
	statetime(lmidx s, int time, int word, int lp) {state=s; t=time; w=word; latpos=lp;}
	statetime() {state=0;t=-1;w=INT_MAX;latpos=time_lat_rec::lstart;}
	bool operator < (const statetime &other) const 
		{if (state < other.state) return true;
		else if (state > other.state) return false;
		else if (t < other.t) return true;
		else if (t > other.t) return false;
		else if (latpos < other.latpos) return true;
		else if (latpos > other.latpos) return false;
		else return (w < other.w);}
};

struct lmrec {  // in conjunction with time, indexes alphas, betas
	lmidx state;
	int w;  // state is not enough to determine the word when the state ends in <unk>
	int lat_end_state;  // if using lattice constraints
	lmrec(lmidx s, int word, int l) {state=s; w=word; lat_end_state=l;}
	bool operator < (const lmrec &other) const 
	{if (state < other.state) return true;
	else if (state > other.state) return false;
	else if (w < other.w) return true;
	else if (w > other.w) return false;
	else return (lat_end_state < other.lat_end_state);}
};

struct feature {
	int idx;
	float val;
	feature(int i, float v) : idx(i), val(v) {}
	inline bool operator < (const feature &other) const {return (idx < other.idx);} // don't need val
	inline bool operator == (const feature &other) const {return (idx==other.idx);}
};

enum mode {viterbi, sum};

enum detector_feature_type {existence, expectation, levenshtein};

enum expectation_type{correct_accept, false_reject, false_accept};  // for expectation features

typedef vector<int> feature_bundle;
// first is stream number
// then ngram level
// then feature type
// then indexes: word, unit+ for existence
//               expectation-type, unit+ for expectation
//               op-type, unit for levenshtein
// (language model indexing done separately)



#endif


