// Geoffrey Zweig (gzweig@microsoft.com)
// Patrick Nguyen (panguyen@microsoft.com)
// Copyright (c) 2009, 2010 Microsoft Corporation. All rights reserved.
//

#ifndef _GCC_
# define _WIN32_WINNT 0x0503
# define WINVER 0x0503
# include <windows.h>
#endif 

#include "ngutil.h"
#include "scarf_config.h"
#include "commandline.h"
#include "main_template.h"

#ifdef __USE_MPI
# include "mpiwrap.h"
#endif // __USE_MPI

#include <iostream>
#include <assert.h>
#include "model.h"

int main(int argc, char *argv[]) {

	cout << "\nSCARF Trainer v1.2\n";
	cout << "gzweig@microsoft.com\n\n";

#ifdef __USE_MPI
	MPI::Initialize(&argc, &argv);
#endif // __USE_MPI

	clparser cl;

	CFG::GlobalConfig globConfig;
	CFG::AppConfig appConfig;
	CFG::TrainConfig trainConfig;

	trainConfig.register_args(cl);
	appConfig.register_args(cl);
	globConfig.register_args(cl);

	cl.parse(argv, argc);

	globConfig.load_args(cl);
	initGlobal(cl, globConfig);
	appConfig.load_args(cl);
	trainConfig.load_args(cl);

	ERR_check(ERR::ERR_USERINPUT, appConfig.lm.length() != 0, "Please specify a language model");
	ERR_check(ERR::ERR_USERINPUT, (appConfig.dict.size() == appConfig.fex.stream.size()), "Please specify a dictionary for each stream");
	ERR_check(ERR::ERR_USERINPUT, (appConfig.fex.existOrder.size() == appConfig.fex.stream.size()), "Please specify an existence order for each stream");
	ERR_check(ERR::ERR_USERINPUT, (appConfig.fex.expectOrder.size() == appConfig.fex.stream.size()), "Please specify an expectation order for each stream");
	WARN_check(ERR::ERR_USERINPUT, appConfig.dict.size() != 0, "Not using a dictionary");
	ERR_check(ERR::ERR_USERINPUT, trainConfig.num_constraints.length() != 0, "Please specify a list of numerator constraints");
	ERR_check(ERR::ERR_USERINPUT, trainConfig.den_constraints.length() != 0, "Please specify a list of denominator constraints");

	WARN_check(ERR::ERR_USERINPUT, appConfig.fex.stream.size() != 0, "No detector streams specified");

	ERR_check(ERR::ERR_USERINPUT, appConfig.model.length() != 0, "Please specify output model file");

	model scarf(appConfig.fex);

	scarf.printScores = trainConfig.printScores;
	scarf.max_alphas = appConfig.max_alphas;

	scarf.dictionaries = appConfig.dict;

	scarf.uttIt->init_constraint_readers(trainConfig.num_constraints.c_str(), trainConfig.den_constraints.c_str());

	size_t streamIdx = 0;
	scarf.num_streams = (int)appConfig.fex.stream.size();

	(void)printf("Got %d streams\n", scarf.num_streams);

	(void)printf("reading language model\n");
	language_model LM(appConfig.lm.c_str(), scarf.idx4word, scarf.word4idx);
	scarf.LM = &LM;
	(void)printf("Done reading language model\n");

	if( scarf.fexConfig.lexternal.size() == 0 )
		scarf.fexConfig.lexternal.resize(scarf.fexConfig.external_format.size(), false);
	else if( scarf.fexConfig.lexternal.size() == 1 )
		scarf.fexConfig.lexternal.resize(scarf.fexConfig.external_format.size(), scarf.fexConfig.lexternal[0]);
	ERR_check(ERR::ERR_SANITY, scarf.fexConfig.external_format.size() == scarf.fexConfig.lexternal.size(), "--lexternal must be of the same size of --external_format");

	ERR_check(ERR::ERR_SANITY, appConfig.fex.stream_names.size() == appConfig.fex.expectOrder.size(), "must have the same number of streams and expectOrder");
	ERR_check(ERR::ERR_SANITY, appConfig.fex.stream_names.size() == appConfig.fex.existOrder.size(), "must have the same number of streams and existOrder");
	if( scarf.fexConfig.use_levenshtein_features.size() == 0 )
		scarf.fexConfig.use_levenshtein_features.resize(scarf.fexConfig.stream.size(), true);
	else if( scarf.fexConfig.use_levenshtein_features.size() == 1 )
		scarf.fexConfig.use_levenshtein_features.resize(scarf.fexConfig.stream.size(), scarf.fexConfig.use_levenshtein_features[0]);
	ERR_check(ERR::ERR_SANITY, scarf.fexConfig.use_levenshtein_features.size() == scarf.fexConfig.stream.size(), "--use_levenshtein_features must be of the same size of --stream");

	for(streamIdx = 0; streamIdx < appConfig.fex.stream_names.size(); ++streamIdx) {
		const std::string &name = appConfig.fex.stream_names[streamIdx];
		scarf.idx4streamname[name] = streamIdx;
		scarf.nglev4exist_stream[name] = appConfig.fex.existOrder[streamIdx];
		scarf.nglev4expect_stream[name] = appConfig.fex.expectOrder[streamIdx];
	}
	scarf.idx4stream_sym.resize(appConfig.fex.stream_names.size());

	// symbol table pointers have been initialized
	scarf.uttIt->init_raw_detectors();

	scarf.number_units_and_features(scarf.dictionaries, appConfig.fex.stream);
	scarf.number_lm_features();
	cout << "LM features created\n";

	if (appConfig.fex.basefiles.length() > 0) {
		scarf.number_baseline_features();
		cout << "Baseline feature created\n";
	}

	scarf.l1_regularizer = trainConfig.l1_reg;
	scarf.l2_regularizer = trainConfig.l2_reg;

	scarf.train(trainConfig);
	
#ifdef __USE_MPI
	if(MPI::isRoot())
#endif // fallthrough:
	{
		if( trainConfig.printModel )
			scarf.display();

		scarf.save_model(appConfig.model.c_str());
	}

#ifdef __USE_MPI
	MPI::UnInitHeartBeat();
	MPI::Uninitialize();
#endif

	return 0;
}
